"""
Module permettant de gérer la génération de commentaires.
C'est dans ce module qu'on va décider vers quel module
de génération de texte on va orienter.
"""
import mflog
import xarray as xr

from ftap_autom.promethee.src.conf.app_configuration import MFLOG
from ftap_autom.promethee.src.lib.localisation.localisation_manager import Localisation
from ftap_autom.promethee.src.lib.localisation.core_spatial import LocalisationError
from ftap_autom.promethee.src.lib.interpretor import AleaInterp
from ftap_autom.promethee.src.lib.text.comment.component import (
    ComponentHandlerLocalisation,
)
from ftap_autom.promethee.src.lib.text.comment.multizone import MultiZoneAdapter
import ftap_autom.promethee.src.lib.comment_generator.text_generator as text_generator


LOGGER = mflog.getLogger("comment_manager.mod")
mflog.set_config(**MFLOG)
LOGGER = LOGGER.bind(module="comment_manager")
xr.set_options(keep_attrs=True)


class CommentManager:
    def __init__(self, component, component_config):
        """Init the comment Manager

        Args:
            component ([type]): Composant.
            component_config (dict): Configuration pour le composant.
        """
        self.component = component
        self.component_config = component_config
        self.alea_conf = AleaInterp(component_config)

    def get_comment(self, geo_id: str) -> str:
        """
        Permet de récupérer le commentaire pour la zone identifiée
        Attention cette fonction a besoin que decision_tree ai été déclenché auparavant.

        Args:
            geo_id (str) : L'identifiant de la zone

        Returns:
            str: Le commentaire
        """
        self.decision_tree(geo_id)
        return self.text_generator_handler.identification_cas(geo_id)

    def produce_monozone(self, geo_id):
        self.text_generator_handler = text_generator.GenerationText(
            self.component, self.component_config,
        )
        return self.text_generator_handler.identification_cas(geo_id)

    def decision_tree(self, geo_id):
        """
        Permet de décider quel module de génération de commentaire est à utiliser.

        Args:
            geo_id (str): Id de la zone.
        """
        module = "unizone"
        maxi_risk = self.component.get_final_risk_max_level(geo_id)

        if maxi_risk >= 1:
            # Mettre d'autres conditions pour ne pas passer par la localisation.
            #   - Type de risque (upstream/downstream)

            # On recupere la définition du risque
            risk_list = self.alea_conf.select_risk_level(level=maxi_risk)
            agg = risk_list[0].generic_get("aggregationType")
            if agg == "downStream":
                hourly_maxi_risk = self.component.get_final_risk_level().sel(id=geo_id)
                # On recupere la periode où le risque le plus grand existe
                risk_period = hourly_maxi_risk.sel(
                    valid_time=(hourly_maxi_risk == maxi_risk)
                ).valid_time
                try:
                    loca_handler = Localisation(
                        self.component_config,
                        risk_level=maxi_risk,
                        geo_id=geo_id,
                        period=set(risk_period.values),
                    )
                    unique_table, _ = loca_handler.get_summarized_info()
                    # On va regarder s'il est nécessaire de faire de
                    # la localisation spatiale sur le risque le plus eleve.
                    if unique_table.id.size > 1:
                        module = "multizone"
                        LOGGER.debug("Going to multiZone commentary type")
                except LocalisationError as e:
                    # Si c'est une erreur 'normale' du module de localisaion
                    # (pas de zones descriptive, pas le bon type de risque, etc... ).
                    # On passe alors au module monozone.
                    LOGGER.error(repr(e))
                    pass

        if module == "unizone":
            self.text_generator_handler = text_generator.GenerationText(
                self.component, self.component_config,
            )
        else:
            compo_handler = ComponentHandlerLocalisation(loca_handler)
            self.text_generator_handler = MultiZoneAdapter(
                compo_handler, monozone_access=self.produce_monozone
            )
