"""
@package lib.config_processor

Config processor module
"""

# Standard packages
import os
import json
import tarfile
from datetime import datetime, timedelta
from copy import deepcopy
from functools import partial
from multiprocessing import Pool

# Third parties packages
import pandas as pd
import numpy as np
import shapely.geometry as sh
import mflog

# Own package
from ftap_autom.promethee.src.utils.hash import MD5
from ftap_autom.promethee.src.conf.app_configuration import (
    ALTERNATE_MAX,
    LOCAL,
    RESOURCE_COLUMNS,
    TEXT_ALGO,
    WORKING_DIR,
    DATETIME_FMT,
    RULES_XLS,
    JSON_KWARGS,
    MFLOG,
)
from ftap_autom.promethee.src.utils.formatter import (
    Formatter,
    TagFormatter,
    format_datestr,
)
from ftap_autom.promethee.src.utils.exception import handle_error, ConfigurationError

# Logging
mflog.set_config(**MFLOG)
LOGGER = mflog.getLogger("config_processor.mod")
LOGGER = LOGGER.bind(module="config_processor")

# Main class : ConfigProcessor


class ConfigProcessor:
    """ConfigProcessor : Class which parses a global configuration file
    (tar archive or json) and produces three configurations out of it :

    * a mask configuration (self.mask_config): config following the geojson
    standard which describes all the geographical geos used in production.
    This file aims to be used by the mask creation module lately.

    * a data configuration (self.data_config) : config which describes all
    the raw data files (using the Vortex Standard) necessary for production
    and all the pre-processings that must be done before production (parameter
    extraction, accumulations, combinations, etc.)
    This file aims to be used by the data preprocessing module lately.

    * a production configuration (self.prod_config) : config which describes
    all the files necessary for production and all the processing to be done
    in order to produce Promethee data (risks and texts).
    This file aims to be used by the core module lately.
    """

    def __init__(
        self,
        global_config_filename,
        rules,
        drafting_datetime,
        production_delta=1,
        working_dir="./",
        vapp="promethee",
        vconf="msb",
        experiment="TEST",
    ):
        """__init__

        Args:
            global_config_filename (str): path to the global configuration file
            rules (str): Name of the rules convention used for files selection. This
                argument must be a key of the app_configuration.RULES_XLS dictionary.
            drafting_datetime (datetime.datetime): Promethee's drafting datetime
            production_delta (int, optionnal) : Delta in hour between the reference
                and the production datetimes. Defaults to 1.
            working_dir (str, optionnal) : Path to the working directory.
                Defaults to "./".
        """
        global LOGGER

        # Opening the global conf file
        LOGGER.info("Creating ConfigProcessor ...", func="__init__")
        self.global_config = []
        try:
            if tarfile.is_tarfile(global_config_filename):
                LOGGER.info("Global Config is a tarfile", func="__init__")
                with tarfile.open(global_config_filename) as global_tar:
                    members = global_tar.getmembers()
                    for i, global_config_file in enumerate(members):
                        global_config = json.load(
                            global_tar.extractfile(global_config_file)
                        )
                        LOGGER.info(
                            f"File {i+1}/{len(members)} loaded", func="__init__"
                        )
                        if isinstance(global_config, list):
                            self.global_config += global_config
                        elif isinstance(global_config, dict):
                            self.global_config += [global_config]
                        else:
                            raise ConfigurationError(
                                f"Global configuration is a {type(global_config)},"
                                " while list or dict expected."
                            )
            elif global_config_filename.endswith(".json"):
                LOGGER.info("Global Config is a JSON file", func="__init__")
                with open(global_config_filename, "r") as global_config_file:
                    self.global_config += json.load(global_config_file)
                LOGGER.info("File loaded", func="__init__")
        except TypeError:
            # Cas où l'on passe un TextIO.
            self.global_config = json.load(global_config_filename)

        # drafting_datetime
        self.drafting_datetime = drafting_datetime
        self.production_timedelta = timedelta(hours=production_delta)
        LOGGER.debug(f"Drafting datetime : {self.drafting_datetime}", func="__init__")
        LOGGER.debug(
            f"Production datetime : {self.production_datetime}", func="__init__"
        )

        # Checking rules convention
        LOGGER.info("Checking rules convention", func="__init__")
        if rules not in RULES_XLS:
            LOGGER.error(
                f"Given rules convention {rules} not allowed. Changed to 'default'.",
                func="__init__",
            )
            rules = "default"

        rules_xls = RULES_XLS[rules]
        LOGGER.debug(f"Rules excel file chosen : {rules_xls}", func="__init__")

        # Opening rules xls file
        LOGGER.info("Opening rules_xls...", func="__init__")
        self.agg_param_df = pd.read_excel(
            rules_xls, sheet_name="agg_param", index_col=0
        )
        self.param_links = pd.read_excel(
            rules_xls, sheet_name="param_links", index_col=0
        )
        self.geometries_df = pd.read_excel(
            rules_xls, sheet_name="geometries", index_col=0
        )
        self.source_files_df = self.create_rules_df(
            rules_xls, "source_files", index_col=0, concat_axis=0
        )
        self.preprocessed_files_df = self.create_rules_df(
            rules_xls, "preprocessed_files", index_col=0, concat_axis=0
        )
        self.files_links_df = self.create_rules_df(
            rules_xls, "files_links", index_col=0, concat_axis=1
        )
        self.files_ids = set(self.source_files_df.index).union(
            self.preprocessed_files_df.index
        )
        LOGGER.info("rules_xls file correctly opened", func="__init__")

        self.data_config = {
            "global_config_version": self.hashcode,
            "sources": dict(),
            "preprocessed": dict(),
        }
        self.prod_config = dict()
        self.mask_config = dict()

        self.working_dir = working_dir
        self.vapp = vapp
        self.vconf = vconf
        self.experiment = experiment

    @property
    def hashcode(self):
        """hashcode (Property)

        Returns:
            str : hexadecimal hash key of the global_config's MD5 checksum
        """
        return MD5(self.global_config).hash

    @property
    def reference_datetime(self):
        """reference_datetime : Property returning the reference datetime
            (i.e. the rounded drafting datetime to the previous hour)

        Returns:
            datetime.datetime : reference datetime
        """
        return datetime(
            year=self.drafting_datetime.year,
            month=self.drafting_datetime.month,
            day=self.drafting_datetime.day,
            hour=self.drafting_datetime.hour,
        )

    @property
    def production_datetime(self):
        """production_datetime : Property returning the production datetime
            (i.e. the rounded drafting datetime to the next hour)

        Returns:
            datetime.datetime : production datetime
        """
        return self.reference_datetime + self.production_timedelta

    def midnight(self, days_offset=0):
        """midnight : method returning the midnight datetime of the production
        day more or less the days_offset.

        Args:
            days_offset (int, optional): Numer of days between the considered
                day and the production day. Defaults to 0.

        Returns:
            datetime.datetime: Midnight datetime of the production day more or
                less the days_offset.
        """
        return datetime(
            year=self.production_datetime.year,
            month=self.production_datetime.month,
            day=self.production_datetime.day,
        ) + timedelta(days=days_offset)

    @staticmethod
    def int_timedelta(d0, d1):
        """int_timedelta : Static method returning the difference between 2 dates
        in hours (as an integer)

        Args:
            d0 (datetime.datetime): First date
            d1 (datetime.datetime): Second date

        Returns:
            int : Diffence in hours between d0 and d1
        """
        return int((d1 - d0) // timedelta(hours=1))

    @staticmethod
    def str_datetime(str_dt, int_td=0):
        """str_datetime : Static method converting a formated datetime (following the
        DATETIME_FMT) to a proper datetime.datetime format. It is possible to
        add a int_td hours to the given datetime.

        Args:
            str_dt (str): Formated datetime following the DATETIME_FMT standard.
            int_td (int, optional): Optionnal hours to add. Defaults to 0.

        Returns:
            datetime.datetime: Corresponding datetime.datetime
        """
        return datetime.strptime(str_dt, DATETIME_FMT) + timedelta(hours=int_td)

    @staticmethod
    def apply_date(dataframe, current_datetime):
        """apply_date : Change all the dates in a dataframe (extracted from the
        rules_xls file) according to a given current_datetime.

        Args:
            dataframe (pandas.DataFrame): Dataframe to transform
            current_datetime (datatime.datetime): Datetime to apply
        """
        global LOGGER
        LOGGER = LOGGER.bind(current_datetime=current_datetime.strftime(DATETIME_FMT))

        def replace_datestr(datestr):
            """replace_datestr : Local function to apply the format_datestr
            function with the current_datetime and the DATETIME_FMT.

            Args:
                datestr (str): String to format

            Returns:
                str : Formatted string
            """
            return format_datestr(datestr, current_datetime, fmt=DATETIME_FMT)

        # copying the original dataframe
        dataframe_copy = pd.DataFrame()

        # columns processing
        for col in dataframe.columns:
            new_col = replace_datestr(col)
            dataframe_copy[new_col] = dataframe[col].apply(replace_datestr)
            LOGGER.debug(f"Dates changed in column {col}", func="apply_date")

        # index processing
        dataframe_copy.index = pd.Index(map(replace_datestr, dataframe.index))
        LOGGER.debug("Index dates changed", func="apply_date")

        if "date" not in dataframe:
            return dataframe_copy

        def datetimize(column, offset_datetime):
            """datetimize : Local function for changing integer timedeltas
            into datetimes in the dataframe_copy.

            Args:
                column (str): Column name
                offset_datetime (datetime.datetime): Base datetime from which all
                timedeltas are applied
            """
            if column not in dataframe_copy:
                return None
            dataframe_copy[column] = dataframe_copy[column].apply(
                lambda n: timedelta(hours=int(n)) if np.isfinite(n) else np.nan
            )
            dataframe_copy[column] = dataframe_copy[column] + offset_datetime
            return None

        for column in ["date", "dispo_time"]:
            datetimize(column, current_datetime)
            LOGGER.debug(f"Dates changed in column {column}", func="apply_date")

        for column in ["start", "stop"]:
            datetimize(column, dataframe_copy["date"])
            LOGGER.debug(f"Dates changed in column {column}", func="apply_date")

        LOGGER = LOGGER.try_unbind("current_datetime")
        return dataframe_copy

    def create_rules_df(self, rules_xls, sheet_name, index_col=None, concat_axis=0):
        """create_rules_df

        Args:
            rules_xls (str): path to the excel conf file containing all the
                rules to follow in order to choose the correct files for the
                configured situations.
            sheet_name (str) : sheet name to open and process
            index_col (int or list, optional) : Index column index in the dataframe
                to open. Default as None.
            concat_axis (int, optional) : Axis along which will be concatenated
                the various dates dataframes.

        Returns:
            pandas.DataFrame : DataFrame containing all the rules to apply in
                order to select the correct files according to the configured
                situations.
        """
        global LOGGER
        LOGGER = LOGGER.bind(
            rules_xls=rules_xls,
            sheet_name=sheet_name,
            index_col=index_col,
            concat_axis=concat_axis,
        )
        # rules xls file opening
        rules_df = pd.read_excel(
            rules_xls, sheet_name=sheet_name, index_col=index_col
        ).dropna(axis=1, how="all")

        # Applying for production day ...
        today_df = self.apply_date(rules_df, self.midnight())
        # ... and for the day before
        yesterday_df = self.apply_date(rules_df, self.midnight(days_offset=-1))
        # And finally concatenating the two dataframes
        rules_df = pd.concat([yesterday_df, today_df], axis=concat_axis)

        # Finally only keeping the lines of available files
        if "dispo_time" in rules_df and "geometry" in rules_df:
            rules_df = rules_df[rules_df["dispo_time"] <= self.reference_datetime]
            rules_df["mesh_size"] = rules_df["geometry"].apply(
                lambda s: self.geometries_df["mesh_size"].loc[s]
            )
            rules_df = rules_df.sort_values(
                by=["mesh_size", "dispo_time"], ascending=[True, False]
            )

        LOGGER.info(
            f"Sheet '{sheet_name}' from file {rules_xls} processed",
            func="create_rules_df",
        )
        LOGGER = LOGGER.try_unbind(
            "rules_xls", "sheet_name", "index_col", "concat_axis"
        )
        return rules_df

    def compute_period(self, period_config):
        """compute_period

        Args:
            period_config (dict): dictionary containing the description of :
                - a simple period, i.e. a beginning, an end, a name and
                a validity time;
                Or :
                - a period with period elements, each element being described
                like a simple period.
                The beginning and the end are materialized by a integer value
                meaning either the delta in hours between the beginnig/end
                and the reference datetime (key with delta), or the delta
                in hours between the beginning/end and the midnight datetime
                of the production day.

        Returns:
            dict : Dictionnary containing the name of the period, the start
                datetime and the end datetime.
        """
        global LOGGER
        LOGGER = LOGGER.bind(period_id=period_config.get("id"))
        # Finding the corresponding period element
        production_timedelta = int(
            (self.production_datetime - self.midnight()).total_seconds() // 3600
        )
        base_datetime = self.midnight()
        if "periodElements" in period_config:
            period_elements = period_config["periodElements"]
            for element in period_elements:
                if "productionTime_until" not in element:
                    LOGGER.warning(
                        "Missing 'productionTime_until' key in period config.",
                        period_element=element,
                        func="compute_period",
                    )
                    element["productionTime_until"] = 24

            elements_sorted = sorted(
                period_elements, key=(lambda d: d["productionTime_until"])
            )

            # Getting the correct period_element in the list
            base_element = None
            for element in elements_sorted:
                if production_timedelta < element["productionTime_until"]:
                    base_element = deepcopy(element)
                    break

            if base_element is None:
                # If there is no correct period element for the production day
                # We get as base_element the first element of the next day
                base_datetime += timedelta(days=1)
                base_element = deepcopy(elements_sorted[0])

        else:
            base_element = period_config
            if production_timedelta >= period_config["productionTime_until"]:
                base_datetime += timedelta(days=1)

        # Processing of the start datetime
        if "start" in base_element:
            delta = timedelta(hours=base_element["start"])
            start_date = base_datetime + delta
        elif "deltaStart" in base_element:
            delta = timedelta(hours=base_element["deltaStart"])
            start_date = self.production_datetime + delta
        elif "absoluStart" in base_element:
            try:
                start_date = datetime.strptime(
                    base_element["absoluStart"], DATETIME_FMT
                )
            except ValueError as absoluStart_error:
                raise ConfigurationError(
                    f"Wrong absoluStart format : got {base_element['absoluStart']} "
                    f"while expected {DATETIME_FMT}"
                ) from absoluStart_error
        else:
            raise ConfigurationError(
                "Period configuration error. Missing a beginning key.",
                period_id=period_config.get("id"),
            )

        if start_date < self.production_datetime:
            LOGGER.warning(
                f"Configured start_date ({start_date}) is before "
                f"the production_datetime ({self.production_datetime}).",
                func="compute_period",
            )

        # Processing the end datetime
        if "stop" in base_element:
            delta = timedelta(hours=base_element["stop"])
            stop_date = base_datetime + delta
        elif "deltaStop" in base_element:
            delta = timedelta(hours=base_element["deltaStop"])
            stop_date = self.production_datetime + delta
        elif "absoluStop" in base_element:
            try:
                stop_date = datetime.strptime(base_element["absoluStop"], DATETIME_FMT)
            except ValueError as absoluStop_error:
                raise ConfigurationError(
                    f"Wrong absoluStop format : got {base_element['absoluStop']} "
                    f"while expected {DATETIME_FMT}"
                ) from absoluStop_error
        else:
            raise ConfigurationError(
                "Period configuration error. Missing an ending key."
            )

        if stop_date < start_date:
            LOGGER.warning(
                f"Configured stop_date ({stop_date}) is before "
                f"the start_date ({start_date}).",
                func="compute_period",
            )
            # stop_date = start_date

        if stop_date < self.production_datetime:
            LOGGER.warning(
                f"Configured stop_date ({stop_date}) is before "
                f"the production_datetime ({self.production_datetime}).",
                func="compute_period",
            )

        # Post processing the period name
        name_to_format = base_element.get("name", period_config.get("name", ""))
        name = Formatter(text=name_to_format).format_dates(self.production_datetime)
        LOGGER.info(
            f"Computed period : ({name}, {start_date}, {stop_date})",
            func="compute_period",
        )
        LOGGER = LOGGER.try_unbind("period_id")
        return (name, start_date, stop_date)

    def param_to_description(self, param):
        """param_to_description : Get a complete parameter name and returns
        a tuple containing the root parameter name, the parameter level, and
        the accumulation period in hours

        Args:
            param (str): Parameter complete name

        Returns:
            tuple:
                str: Root parameter name
                str: Parameter level
                int: Accumulation period in hours
        """
        root_param, param_level = param.split("__")
        if root_param in self.agg_param_df.index:
            agg_description = self.agg_param_df.loc[root_param]
            return (
                agg_description["param"] + "__" + param_level,
                int(agg_description["accum"]),
            )

        return param, None

    def description_to_param(self, root_param, param_level, accum):
        """description_to_param: Returns the complete parameter name given
        the root parameter name, the parameter level and the accumulation period.

        Args:
            root_param (str): Root parameter name
            param_level (str): Parameter level
            accum (int): Accumulation period in hours

        Returns:
            str: Complete parameter name
        """
        agg_index = self.agg_param_df[self.agg_param_df["param"] == root_param]
        agg_index = agg_index[agg_index["accum"] == accum].index
        if len(agg_index) == 0:
            return root_param + "__" + param_level

        param_name = agg_index[0]
        return param_name + "__" + param_level

    def get_components_params(self, compo_config):
        """get_components_params : Method which returns all the parameters
        used in the component or linked to the component.

        Args:
            compo_config (dict): Component config

        Returns:
            set : Set of parameter used in the component or linked.
        """
        params = []
        if compo_config["type"] == "risk":
            for level in compo_config["levels"]:
                for event in level["elementsEvent"]:
                    full_root_param, accum = self.param_to_description(event["field"])
                    if full_root_param not in self.param_links:
                        params += [event["field"]]
                        continue
                    linked_root_params = [
                        p.split("__")
                        for p in self.param_links[full_root_param].dropna().index
                    ]
                    params += [
                        self.description_to_param(p, l, accum)
                        for p, l in linked_root_params
                    ]
        elif compo_config["type"] == "text":
            for weather in compo_config["weather"]:
                params += set(
                    TEXT_ALGO[weather["id"]][weather.get("algo", "generic")][
                        "fields"
                    ].values()
                )
                if "field" in weather["condition"]:
                    params += [weather["condition"]["field"]]
        else:
            raise ConfigurationError(
                f"Unexpected component type : {compo_config['type']}."
            )
        return set(params)

    def useable_geometries(self, geo_config):
        """useable_geometries : Returns the useable geometries with
        a geographical zone according to its configuration.

        Args:
            geo_config (dict): Geograpghical zone's configuration
        """
        global LOGGER
        LOGGER = LOGGER.bind(geo_id=geo_config["id"])

        def is_feature_in(geometry):
            """is_feature_in : Check if a geographical feature (according to the
            GeoJSON standard) is in the given geometry.

            Args:
                geometry (str): Geometry name

            Returns:
                bool : If the feature is in the geometry
            """

            sh_feature = sh.asShape(geo_config["geometry"])
            if not sh_feature.is_valid:
                LOGGER.error(
                    "Invalid geometry configuration", func="useable_geometries"
                )
                return False
            sh_bounds = sh_feature.bounds
            if not ((-180 <= sh_bounds[0] <= 180) and (-180 <= sh_bounds[2] <= 180)):
                LOGGER.error(
                    f"Longitude value error : given (lon, lat, lon, lat) = {sh_bounds}"
                    ", while longitude values must be in [-180,180].",
                    func="useable_geometries",
                )
                return False
            if not ((-90 <= sh_bounds[1] <= 90) and (-90 <= sh_bounds[3] <= 90)):
                LOGGER.error(
                    f"Latitude value error : given (lon, lat, lon, lat) = {sh_bounds}"
                    ", while latitude values must be in [-90,90].",
                    func="useable_geometries",
                )
                return False

            lat_min, lon_min, lat_max, lon_max = (
                self.geometries_df[["lat_min", "lon_min", "lat_max", "lon_max"]]
                .loc[geometry]
                .values
            )
            grid = sh.Polygon(
                (
                    (lon_min, lat_min),
                    (lon_min, lat_max),
                    (lon_max, lat_max),
                    (lon_max, lat_min),
                    (lon_min, lat_min),
                )
            )
            return grid.contains(sh_feature)

        useable = []
        for geometry in self.geometries_df.index:
            if is_feature_in(geometry):
                useable += [geometry]
        LOGGER = LOGGER.try_unbind("geo_id")
        return useable

    def get_file_info(self, file_id: str) -> dict:
        """get_file_info : returns informations from a given file_id
        (source or preprocessed) extracted from either self.source_files_df
        or self.preprocessed_files_df and adds to the file_info the list of terms.

        Args:
            file_id (str): File ID in the self.source_files_df or in
                self.preprocessed_files_df indexes.

        Returns:
            dict : File info
        """
        if file_id in self.source_files_df.index:
            file_info = self.source_files_df.loc[file_id]

        elif file_id in self.preprocessed_files_df.index:
            file_info = self.preprocessed_files_df.loc[file_id]

        else:
            LOGGER.error(
                f"file_id {file_id} is neither in self.source_files_df nor in "
                "self.preprocessed_files_df",
                file_id=file_id,
                func="get_file_info",
            )
            return dict()

        step = timedelta(hours=int(file_info["step"]))
        term_range = range(int((file_info["stop"] - file_info["start"]) / step) + 1)
        file_info.loc["terms"] = [file_info["start"] + step * i for i in term_range]
        return file_info

    def best_preprocessed_file_term(self, term, geometries, params):
        """best_preprocessed_file_term: Returns the best preprocessed file id
        according to the given term, useable geometries and parameters

        Args:
            term (datetime.datetime): Term
            geometries (iterable of str) : List, tuple or set of useable geometries
            params (iterable of str) : List, tuple or set of parameters

        Returns:
            str : Most conveniant preprocessed file id
        """
        for file_id in self.preprocessed_files_df.index:
            file_info = self.get_file_info(file_id)
            if term not in file_info.get("terms", []):
                continue
            if file_info.get("geometry", "") not in geometries:
                continue
            params_ok = True
            for param in params:
                full_root_param, _ = self.param_to_description(param)
                if full_root_param not in self.files_links_df[file_id].dropna().index:
                    params_ok = False
                    break
            if params_ok:
                return [file_id] + file_info[["date", "stop"]].to_list()
        return None

    def best_preprocessed_files(self, start, stop, geometries, params):
        """best_preprocessed_files

        Args:
            start (datetime.datetime): Start datetime of the period
            stop (datetime.datetime): Stop datetime of the period
            geometries (iterable of str) : List, tuple or set of useable geometries
            params (iterable of str) : List, tuple or set of parameters

        Returns:
            list : list of all the files descriptions needed to cover the period,
                the zone and the parameters.
                A file description (an element of the returned list) is a tuple
                containing the filename, the start term of the file and the stop
                term.
        """
        global LOGGER
        LOGGER = LOGGER.bind(
            start=start.strftime(DATETIME_FMT),
            stop=stop.strftime(DATETIME_FMT),
            geometries=list(geometries),
            params=list(params),
        )
        if start > stop or self.production_datetime > stop:
            LOGGER.error(
                f"No preprocessed file found for start={start}; "
                f"stop={stop}; production={self.production_datetime}",
                func="best_preprocessed_files",
            )
            return []
        correct_start = max(self.production_datetime, start)
        current_best_file = self.best_preprocessed_file_term(
            correct_start, geometries, params
        )
        if current_best_file is None:
            return self.best_preprocessed_files(
                correct_start + timedelta(hours=1), stop, geometries, params,
            )

        filename, date, file_stop = current_best_file
        if file_stop >= stop:
            file_stop = stop
            other_files = []

        else:
            if file_stop < self.midnight(days_offset=2):
                new_start = file_stop + timedelta(hours=1)
            else:
                new_start = file_stop + timedelta(hours=3)

            other_files = self.best_preprocessed_files(
                new_start, stop, geometries, params
            )

        start_time = correct_start.strftime(DATETIME_FMT)
        stop_time = file_stop.strftime(DATETIME_FMT)

        LOGGER = LOGGER.try_unbind("start", "stop", "geometries", "params")
        return [(filename, start_time, stop_time)] + other_files

    def get_alternate(self, file_id: str) -> str:
        try:
            return self.get_file_info(file_id)["alternate"]
        except Exception:
            LOGGER.warning(
                f"Failed to retrieve alternate info from file : {file_id}",
                file_id=file_id,
                exc_info=True,
            )
            return None

    def create_rh(
        self,
        file_id: str,
        term: datetime = None,
        param: str = None,
        alternate: tuple = None,
    ) -> dict:
        # basic infos from dataframe
        file_info = self.get_file_info(file_id)

        # resource_handler
        rh_dico = file_info[RESOURCE_COLUMNS].to_dict()
        rh_dico["date"] = file_info["date"].strftime(DATETIME_FMT)
        rh_dico["vapp"] = file_info.get("vapp", self.vapp)
        rh_dico["vconf"] = file_info.get("vconf", self.vconf)
        rh_dico["experiment"] = file_info.get("experiment", self.experiment)
        rh_dico["block"] = file_info["block"]
        rh_dico["namespace"] = file_info["namespace"]
        rh_dico["format"] = file_info["nativefmt"]

        # case if we create a source rh
        if file_id in self.source_files_df.index and term is not None:
            rh_dico["term"] = self.int_timedelta(file_info["date"], term)
            role_name = f"{file_id} {term.strftime(DATETIME_FMT)}"

        # case if we create a preprocessed rh
        elif file_id in self.preprocessed_files_df.index and param is not None:
            rh_dico["param"] = param
            rh_dico["begintime"] = self.int_timedelta(
                file_info["date"], file_info["start"]
            )
            rh_dico["endtime"] = self.int_timedelta(
                file_info["date"], file_info["stop"]
            )
            rh_dico["step"] = int(file_info["step"])
            role_name = f"{file_id} {param}"

        # role or alternate
        if alternate is None:
            rh_dico["local"] = os.path.join(
                self.working_dir,
                WORKING_DIR["data"],
                TagFormatter().format_tags(LOCAL[file_info["kind"]], rh_dico),
            )
            rh_dico["role"] = role_name

        else:
            rh_dico["alternate"], rh_dico["local"] = alternate

        rh_dico["fatal"] = False
        rh_dico["now"] = True
        return rh_dico

    def create_full_file_config(
        self, file_id: str, term: datetime = None, param: str = None
    ) -> list:
        # creating main rh
        role_rh = self.create_rh(file_id, term=term, param=param)
        rhs = [role_rh]

        # adding alternates
        current_file_id = file_id
        for _ in range(ALTERNATE_MAX):
            current_file_id = self.get_alternate(current_file_id)
            if current_file_id not in self.files_ids:
                break
            rhs += [
                self.create_rh(
                    current_file_id,
                    term=term,
                    param=param,
                    alternate=(role_rh["role"], role_rh["local"]),
                )
            ]

        return rhs

    def source_files_terms(self, file_id: str, param: str, accum: int) -> dict:
        """source_files_terms : Returns source files id and term needed for a given
        preprocessed file id, a complete parameter name, an accumulation period
        and start/stop datetimes. It also computes the source files configurations
        and stores them into the self.data_config['sources'].

        Args:
            file_id (str): Preprocessed file id
            param (str): Complete Parameter name
            accum (int): Accumulation period in hours

        Returns:
            dict : Dictionnary with the following structure :
                {
                    <source_file_id> : [terms]
                }
        """
        preprocessed_file_info = self.get_file_info(file_id)
        source_files_list = [
            source_file_id.strip()
            for source_file_id in self.files_links_df.loc[param][file_id].split(",")
        ]
        source_files_dico = dict()
        current_start = preprocessed_file_info["start"]
        # Adding virtual terms due to accumulation (for preproc files)
        preproc_stop = preprocessed_file_info["stop"]
        preproc_step = timedelta(hours=int(preprocessed_file_info["step"]))
        accum_td = timedelta(hours=int(accum)) if accum is not None else timedelta(0)
        virtual_stop = preproc_stop + accum_td
        virtual_range = range(1, int((virtual_stop - preproc_stop) / preproc_step) + 1)
        virtual_terms = [preproc_stop + preproc_step * i for i in virtual_range]
        for source_file_id in source_files_list:
            # We check until virtual stop in case of accumulation
            if virtual_stop <= current_start:
                break
            if source_file_id not in self.source_files_df.index:
                continue

            source_file_info = self.get_file_info(source_file_id)
            source_files_dico[source_file_id] = {
                "terms": [
                    self.int_timedelta(source_file_info["date"], term)
                    for term in source_file_info["terms"]
                    if (
                        term >= current_start
                        and term in preprocessed_file_info["terms"] + virtual_terms
                    )
                ],
                "step": int(source_file_info["step"]),
            }

            current_start = source_file_info["terms"][-1] + preproc_step

            if source_file_id in self.data_config["sources"]:
                continue

            self.data_config["sources"][source_file_id] = dict()
            for term in source_file_info["terms"]:
                term_int = self.int_timedelta(source_file_info["date"], term)
                self.data_config["sources"][source_file_id][
                    term_int
                ] = self.create_full_file_config(file_id=source_file_id, term=term)

        return source_files_dico

    def change_slice_to_islice(self, slice_time, file_id):
        """change_slice_to_islice: Function which enable to go from slice (in hours)
        to islice (which select elements)

        Args:
            slice_time (tuple): A tuple describing the first and last hour to take
            file_id (str): Preprocessed file id

        Returns:
            [tuple]: New tuple
        """
        file_info = self.get_file_info(file_id)
        date = file_info["date"]
        start = file_info["start"]
        td_start = self.int_timedelta(date, start)
        step = int(file_info["step"])
        n_start = int((slice_time[0] - td_start) / step)
        n_stop = int((slice_time[1] - td_start) / step)
        return (n_start, n_stop + 1)

    def preprocessed_rh(self, file_id, param):
        key = " ".join([file_id, param])
        if key in self.data_config["preprocessed"]:
            return self.data_config["preprocessed"][key]["resource_handler"]

        return self.create_full_file_config(file_id, param=param)

    def list_components_configs(self, prod_idx):
        """list_components_config : list all the components configurations
        contained in a prod_idx configuration.

        Args:
            prod_idx (int): Production index in the self.global_config
        """
        return deepcopy(self.global_config[prod_idx]["components"])

    def get_geo(self, config):
        """get_geo
        Patch to transform config geos config such as it is compatible
        with GeoJSON format
        """
        return config

    def get_configuration_date(self):
        """Retourne la date de configuration d'un bulletin
        Args:
            prod_idx : le bulletin en quesiton
        Returns:
            La date au format promethee.
            Par défaut on retourne le 1er Janvier 1901.
        """
        return self.global_config[0].get(
            "date_config", datetime(1901, 1, 1).strftime(DATETIME_FMT)
        )

    @staticmethod
    def process_agg_kwargs(dico, mask_file):
        if "aggregation" in dico:
            agg_kwargs = dico["aggregation"].get("kwargs")
            if agg_kwargs is not None:
                for central_zone in (
                    "central_mask_id",
                    "centralZone",
                    "centralZoneConditional",
                ):
                    central_mask_id = agg_kwargs.get(central_zone)
                    if central_mask_id is not None:
                        agg_kwargs["central_mask_id"] = {
                            "file": mask_file,
                            "mask_id": central_mask_id,
                        }
                        agg_kwargs.pop(central_zone)
                # On change la condition de dr.
                if "drConditional" in agg_kwargs:
                    agg_kwargs["dr"] = agg_kwargs.pop("drConditional")
                if "drCentralZone" in agg_kwargs:
                    agg_kwargs["dr"] = agg_kwargs.pop("drCentralZone")
                agg_kwargs["dr"] = float(agg_kwargs["dr"])
                if agg_kwargs["dr"] > 1 and agg_kwargs["dr"] <= 100:
                    agg_kwargs["dr"] = agg_kwargs["dr"] / 100.0
        return dico

    def process_single_component(
        self,
        compo_config: dict,
        single_data_config: dict,
        single_mask_config: dict,
        processed_periods: dict,
        processed_hazards: dict,
    ) -> dict:
        """process_single_component :

        Args:
            compo_config (dict): [description]
            single_data_config (dict): [description]
        """
        global LOGGER
        LOGGER = LOGGER.bind(
            compo_id=compo_config.get("id"), period_id=compo_config.get("period"),
        )
        processed_compo_config = deepcopy(compo_config)

        # Component's params
        params = self.get_components_params(compo_config)

        # Component's period
        period_name, start, stop = processed_periods[compo_config["period"]]
        processed_compo_config["period"] = {
            "id": compo_config["period"],
            "name": period_name,
            "start": start.strftime(DATETIME_FMT),
            "stop": stop.strftime(DATETIME_FMT),
        }

        if compo_config["type"] == "risk":
            if "name" not in processed_compo_config:
                LOGGER.info("Filling with hazard name", func="process_single_component")
                processed_compo_config["name"] = processed_hazards[
                    compo_config["hazard"]
                ].get("name")

            if "otherNames" not in processed_compo_config:
                processed_compo_config["otherNames"] = processed_hazards[
                    compo_config["hazard"]
                ].get("otherNames")

        # Component's geos
        files_groups = dict()
        for mask_id in compo_config["geos"]:
            LOGGER = LOGGER.bind(mask_id=mask_id)
            geometries = self.useable_geometries(
                next(
                    dico
                    for dico in single_mask_config["geos"]["features"]
                    if dico["id"] == mask_id
                )
            )
            best_files = tuple(
                self.best_preprocessed_files(
                    start=start, stop=stop, geometries=geometries, params=params,
                )
            )

            if best_files in files_groups:
                files_groups[best_files] += [mask_id]
            else:
                files_groups[best_files] = [mask_id]
        LOGGER = LOGGER.try_unbind("mask_id")

        LOGGER.info(
            "Files group created",
            func="process_single_config",
            params=list(params),
            files_groups=[
                [list(best_files), list(geos)]
                for best_files, geos in files_groups.items()
            ],
        )

        new_components_list = []
        for best_files, geos in files_groups.items():
            for file_id, _, _ in best_files:
                for param in params:
                    if (file_id, param) not in single_data_config:
                        single_data_config[(file_id, param)] = self.preprocessed_rh(
                            file_id, param
                        )

            new_compo = deepcopy(processed_compo_config)
            new_compo["geos"] = {"file": single_mask_config["file"], "mask_id": geos}
            new_compo["time_dimension"] = "valid_time"
            new_compo["production_datetime"] = self.production_datetime.strftime(
                DATETIME_FMT
            )
            if new_compo["type"] == "text":
                for weather in new_compo["weather"]:
                    params = TEXT_ALGO[weather["id"]][weather.get("algo", "generic")][
                        "fields"
                    ].items()
                    weather["fields"] = {
                        key: [
                            {
                                "file": single_data_config[(file_id, param)][0][
                                    "local"
                                ],
                                "td_slice": {"valid_time": (start_time, stop_time)},
                            }
                            for file_id, start_time, stop_time in best_files
                        ]
                        for key, param in params
                    }
                    if "field" in weather["condition"]:
                        param = weather["condition"]["field"]
                        weather["condition"]["field"] = [
                            {
                                "file": single_data_config[(file_id, param)][0][
                                    "local"
                                ],
                                "td_slice": {"valid_time": (start_time, stop_time)},
                            }
                            for file_id, start_time, stop_time in best_files
                        ]
                    self.change_threshold_operator(weather["condition"])
            else:
                new_compo["levels"] = []
                for file_id, start_time, stop_time in best_files:
                    LOGGER = LOGGER.bind(file_id=file_id)
                    for level in processed_compo_config["levels"]:
                        LOGGER = LOGGER.bind(level=level.get("level"))
                        level_copy = deepcopy(level)
                        level_copy = self.process_agg_kwargs(
                            level_copy, single_mask_config["file"]
                        )
                        for event in level_copy["elementsEvent"]:
                            param = event["field"]
                            LOGGER = LOGGER.bind(param=param)
                            root_param, _ = param.split("__")
                            if root_param in self.agg_param_df.index:
                                # Cas special des arguments pour riskBertrand
                                tmp_param, accum = self.param_to_description(param)
                                base_param, tmp_l = tmp_param.split("__")
                                model_step = int(self.get_file_info(file_id)["step"])
                                LOGGER.info(
                                    f"Parameter {param} is inside agg_param_list. "
                                    "Checking for Bertrand risk.",
                                    func="process_single_component",
                                )
                                if accum > model_step:
                                    param_base = self.description_to_param(
                                        base_param, tmp_l, model_step
                                    )
                                    LOGGER.info(
                                        "Adding information for Bertrand risk. "
                                        f"Param de base a cette echeance {param_base}",
                                        func="process_single_component",
                                    )
                                    event["process"] = "Bertrand"
                                    if (file_id, param_base,) not in single_data_config:
                                        single_data_config[
                                            (file_id, param_base)
                                        ] = self.preprocessed_rh(file_id, param_base)
                                    event["field_1"] = {
                                        "file": single_data_config[
                                            (file_id, param_base)
                                        ][0]["local"],
                                        "td_slice": {
                                            "valid_time": (start_time, stop_time)
                                        },
                                    }
                                    event["cum_period"] = accum
                                    # On n'utilise pas dask pour le field
                                    # (permettant ainsi d'économiser du temps)
                                    # event["chunks_mask"] = {"id":10}  On utilise pas
                                    # dask pour les masques. Permet de ne pas faire
                                    # bugguer le code.
                            event["field"] = {
                                "file": single_data_config[(file_id, param)][0][
                                    "local"
                                ],
                                "td_slice": {"valid_time": (start_time, stop_time)},
                            }
                            event = self.process_agg_kwargs(
                                event, single_mask_config["file"],
                            )
                            self.change_threshold_operator(event)
                        new_compo["levels"] += [level_copy]
                        LOGGER = LOGGER.try_unbind("param")
                    LOGGER = LOGGER.try_unbind("level")
                LOGGER = LOGGER.try_unbind("file_id")
            new_components_list += [new_compo]
        return new_components_list

    def process_single_config(self, prod_idx):
        """process_single_config

        Args:
            config (dict): configuration of a single bulletin

        Returns:
            (dict, set, dict): Tuple containing :
                * the extracted mask configuration dictionnary (geoJSON
                standard)
                * the set of all the data_files (filename and params) necessary
                for production
                * the extracted production configuration dictionnary
        """
        config = self.global_config[prod_idx]
        prod_hashcode = MD5(config).hash
        prod_id = config.get("production_id", prod_idx)
        prod_name = config.get("production_name", f"production_{prod_id}")
        global LOGGER
        LOGGER = LOGGER.bind(prod_id=prod_id)
        LOGGER.info("Starting process_single_config", func="process_single_config")
        # Mask configs
        mask_hashcode = MD5(config["geos"]).hash
        mask_file = os.path.join(self.working_dir, WORKING_DIR["mask"], f"{prod_id}.nc")
        single_mask_config = {
            "file": mask_file,
            "id": prod_id,
            "name": prod_name,
            "config_hash": self.hashcode,
            "prod_hash": prod_hashcode,
            "mask_hash": mask_hashcode,
            "geos": self.get_geo(config["geos"]),
            "resource_handler": {
                "role": f"mask_{prod_id}",
                "fatal": False,
                "kind": "promethee_mask",
                "promid": prod_id,
                "version": mask_hashcode,
                "namespace": "vortex.cache.fr",
                "experiment": self.experiment,
                "vapp": self.vapp,
                "vconf": self.vconf,
                "block": "masks",
                "format": "netcdf",
                "local": mask_file,
            },
        }
        LOGGER.info("Mask config done", func="process_single_config")

        # Processing periods
        processed_periods = dict()
        for period_config in config["periods"]:
            processed_periods[period_config["id"]] = self.compute_period(period_config)
        LOGGER.info("Periods created", func="process_single_config")

        # Processing Hazards
        processed_hazards = dict()
        for hazard in config["hazards"]:
            LOGGER = LOGGER.bind(hazard_id=hazard["id"])
            if isinstance(hazard["id"], (list, tuple)):
                LOGGER.warning(
                    "Given hazard['id'] as list (or tuple)",
                    func="process_single_config",
                )
                if len(hazard["id"]) > 1:
                    raise ConfigurationError(
                        "Given hazard['id'] as list or tuple,"
                        f"of length {len(hazard['id'])} > 1",
                        func="process_single_config",
                    )
                hazard["id"] = hazard["id"][0]
            processed_hazards[hazard["id"]] = deepcopy(hazard)
        LOGGER = LOGGER.try_unbind("hazard_id")
        LOGGER.info("Hazards created", func="process_single_config")

        # Data and prod configs
        single_prod_config = {
            "id": prod_id,
            "name": prod_name,
            "config_hash": self.hashcode,
            "prod_hash": prod_hashcode,
            "mask_hash": mask_hashcode,
            "components": [],
        }

        single_data_config = dict()
        components_configs = self.list_components_configs(prod_idx)
        LOGGER.info(
            f"{len(components_configs)} components to process",
            func="process_single_config",
        )
        for compo_config in components_configs:
            try:
                single_prod_config["components"] += self.process_single_component(
                    compo_config,
                    single_data_config,
                    single_mask_config,
                    processed_periods,
                    processed_hazards,
                )
            except BaseException:
                LOGGER.error(
                    f"Failed to process component {compo_config.get('id')}",
                    compo_id=compo_config.get("id"),
                    exc_info=True,
                )
        LOGGER = LOGGER.try_unbind(
            "prod_id", "compo_id", "period_id", "mask_id", "param", "level", "file_id"
        )
        return single_mask_config, single_data_config, single_prod_config

    def append_config(self, single_processed_configs):
        """append_config : callback method for parallel processing
        of individual processed configs

        Args:
            single_processed_configs ([type]): [description]
        """
        prod_id = single_processed_configs[0]["id"]
        # Mask
        self.mask_config[prod_id] = single_processed_configs[0]

        # Prod
        self.prod_config[prod_id] = single_processed_configs[2]

        # Data
        for (file_id, param), rh_dico in single_processed_configs[1].items():
            key = " ".join([file_id, param])
            if key in self.data_config["preprocessed"]:
                continue

            # sources
            full_root_param, accum = self.param_to_description(param)
            sources_dico = self.source_files_terms(
                file_id=file_id, param=full_root_param, accum=accum,
            )

            self.data_config["preprocessed"][key] = {
                "resource_handler": rh_dico,
                "sources": sources_dico,
                "agg": {"param": full_root_param, "accum": accum},
            }

    @staticmethod
    def change_threshold_operator(event: dict):
        """change_threshold_operator : trying to change threshold if not a correct
        threshold format, and change operator the way the event is described, to be
        compliant with the risk module.

        Args:
            event (dict): event dictionnary
        """
        global LOGGER
        LOGGER = LOGGER.bind(event_id=event.get("id"))

        def convert_single_threshold(threshold):
            if isinstance(threshold, (bool, int, float, complex)):
                return threshold

            log_text = f"Given threshold '{threshold}' of type {type(threshold)}, "
            log_text += "while (bool, int, float, complex) expected. "

            converted = None
            if isinstance(threshold, str):
                if threshold.lower() == "true":
                    converted = True
                elif threshold.lower() == "false":
                    converted = False

            if converted is None:
                try:
                    converted = int(threshold)
                except ValueError:
                    try:
                        converted = float(threshold)
                    except ValueError:
                        try:
                            converted = complex(threshold)
                        except ValueError:
                            pass

            if converted is None:
                log_text += "Failed to convert."
                converted = threshold
            else:
                log_text += f"Converted to {type(converted)}."

            LOGGER.warning(log_text, func="change_threshold_operator")
            return converted

        if "plain" in event:
            if isinstance(event["plain"]["threshold"], (list, tuple)):
                # Changement des threshold
                event["plain"]["threshold"] = [
                    convert_single_threshold(t) for t in event["plain"]["threshold"]
                ]
                # Adaptation des opérateurs de comparaison '==' qui devient isin
                if event["plain"]["comparisonOp"] in ("egal", "==", "="):
                    event["plain"]["comparisonOp"] = "isin"
            else:
                event["plain"]["threshold"] = convert_single_threshold(
                    event["plain"]["threshold"]
                )

        if "mountain" in event:
            if isinstance(event["mountain"]["threshold"], (list, tuple)):
                # Changement des threshold
                event["mountain"]["threshold"] = [
                    convert_single_threshold(t) for t in event["mountain"]["threshold"]
                ]
                # Adaptation des opérateurs de comparaison '==' qui devient isin
                if event["mountain"]["comparisonOp"] in ("egal", "==", "="):
                    event["mountain"]["comparisonOp"] = "isin"
            else:
                event["mountain"]["threshold"] = convert_single_threshold(
                    event["mountain"]["threshold"]
                )

        # On sort mountainThreshold du dico altitude.
        if "altitude" in event:
            if isinstance(event["altitude"], list):
                event["mountainThreshold"] = event["altitude"][0]["mountainThreshold"]
            elif isinstance(event["altitude"], dict):
                event["mountainThreshold"] = event["altitude"]["mountainThreshold"]
            else:
                raise ConfigurationError(
                    "altitude in configuration file should be either dict or list. "
                    f"Got {type(event['atltitude'])}",
                    func="change_threshold_operator",
                )
        if "alt_min" in event or "alt_max" in event and "mountain" not in event:
            event.pop("mountainThreshold")
            event.pop("altitude")
        LOGGER = LOGGER.try_unbind("event_id")

    def process_all(self, nproc=16):
        """process_all

        Returns:
            (dict, dict, dict): Tuple of all three expected configuration
                dictionnaries (mask, data and production)
        """
        len_config = len(self.global_config)
        my_error_handler = partial(handle_error, log=LOGGER)
        pool = Pool(nproc)
        for i in range(len_config):
            LOGGER.info(
                f"Processing production configuration n°{i+1}/{len_config}",
                func="process_all",
            )
            pool.apply_async(
                self.process_single_config,
                args=(i,),
                callback=self.append_config,
                error_callback=my_error_handler,
            )
        pool.close()
        pool.join()
        return self.mask_config, self.data_config, self.prod_config


class ConfigMetronomeProcessor(ConfigProcessor):
    """ConfigMetronomeProcessor : Class which parses a global configuration
    issued from Metronome, reshape it like an original Promethee config and
    produces three configurations out of it (like the ConfigProcessor) :
        * a mask configuration (self.mask_config)
        * a data configuration (self.data_config)
        * a production configuration (self.prod_config)
    """

    def __init__(
        self,
        global_config_filename,
        rules,
        drafting_datetime,
        production_delta=1,
        working_dir="./",
        vapp="promethee",
        vconf="msb",
        experiment="TEST",
    ):
        """__init__

        Args:
            global_config_filename (str): path to the global configuration file
            rules_xls (str): path to the excel conf file containing all the
                rules to follow in order to choose the correct files for the
                situations.
            drafting_datetime (datetime.datetime): Promethee's drafting datetime
            production_delta (int, optionnal) : Delta in hour between the reference
                and the production datetimes. Defaults to 1.
            working_dir (str, optionnal) : Path to the working directory.
                Defaults to "./".
        """
        super().__init__(
            global_config_filename,
            rules,
            drafting_datetime,
            production_delta,
            working_dir,
            vapp,
            vconf,
            experiment,
        )

    def get_geo(self, config):
        """get_geo
        Enable to transform METRONOME area config dictionnary such that
        it is compatible with GeoJSON format.

        Args:
            config ([dict): The input config
        Returns:
            [dict]: [The config in GeoJSON format]
        """
        res = {"type": "FeatureCollection", "features": []}
        for elt in config:
            res["features"].append(elt)
        return res

    @staticmethod
    def base_component_config(compo_config):
        """base_component_config: Returns the base configuration of a component
        given its original configuration

        Args:
            compo_config (dict): Component's original configuration

        Returns:
            dict: Component's base configuration
        """
        output = {}
        if "alt_min" in compo_config["data"]:
            output["alt_min"] = compo_config["data"].get("alt_min")
        if "alt_max" in compo_config["data"]:
            output["alt_max"] = compo_config["data"].get("alt_max")
        output.update(
            {
                "id": compo_config["id"],
                "type": compo_config["data"]["type"],
                "name": compo_config["name"],
                "customer": compo_config["customer"],
                "customer_name": compo_config.get("customer_name", "unknown"),
                "production_id": compo_config.get("production_id", "UnknownProdId"),
                "production_name": compo_config.get(
                    "production_name", "UnknownProdName"
                ),
                "geos_descriptive": compo_config.get(
                    "geos_descriptive", compo_config["data"].get("geos_descriptive", [])
                ),
                "compass_split": compo_config["data"].get("compass_split", True),
                "altitude_split": compo_config["data"].get("altitude_split", True),
                "product_comment": compo_config["data"].get("product_comment", True),
            }
        )

        return output

    def reshape_hazard(
        self, hazard_idx: int, hazard: dict, columns: list
    ) -> pd.DataFrame:
        global LOGGER
        hazard_df = pd.DataFrame(columns=columns)
        LOGGER = LOGGER.bind(hazard_id=hazard["id"])
        if isinstance(hazard["id"], (list, tuple)):
            LOGGER.warning("Given hazard_id as list (or tuple)", func="reshape_hazard")
            if len(hazard["id"]) > 1:
                raise ConfigurationError(
                    "Given hazard_id as list or tuple, "
                    f"of length {len(hazard['id'])} > 1",
                    func="reshape_hazard",
                )
            hazard["id"] = hazard["id"][0]

        hazard_name = hazard.get("label", hazard.get("technical_name", "unknown"))
        if hazard_name == "unknwon":
            LOGGER.warning("Given hazard as no known name")
        for level_idx, level in enumerate(hazard["levels"]):
            for config_idx, config in enumerate(level["configs"]):
                for geo in config["geos"]:
                    for period in config["periods"]:
                        hazard_df = hazard_df.append(
                            other={
                                "hazard": hazard["id"],
                                "period": period,
                                "hazardName": hazard_name,
                                "geo": geo,
                                "hazard_idx": hazard_idx,
                                "level_idx": level_idx,
                                "config_idx": config_idx,
                            },
                            ignore_index=True,
                        )
        LOGGER = LOGGER.try_unbind("hazard_id")
        return hazard_df

    def reshape_risk_component(self, compo_config):
        """reshape_risk_component: Transform a risk component configuration
        into a Promethee's structure given a component configuration with
        a Metronome structure.
        TO DO : Explain changes between component's configurations structures

        Args:
            compo_config (dict): Component's configuration
                (with a Metronome's structure)

        Returns:
            list of dict : List of component's configuration following
                Promethee's structure
        """
        global LOGGER
        LOGGER = LOGGER.bind(compo_id=compo_config.get("id"))
        component_df = pd.DataFrame(
            columns=[
                "hazard",
                "period",
                "hazardName",
                "geo",
                "hazard_idx",
                "level_idx",
                "config_idx",
            ]
        )
        for hazard_idx, hazard in enumerate(compo_config["data"]["hazards"]):
            try:
                component_df = component_df.append(
                    self.reshape_hazard(hazard_idx, hazard, component_df.columns),
                    ignore_index=True,
                )
            except BaseException:
                LOGGER.error(
                    "Failed to reshape hazard.",
                    hazard_id=hazard.get("id"),
                    exc_info=True,
                )
        component_df = component_df.set_index(
            ["hazard", "period", "hazardName", "geo"]
        ).sort_index()

        grouped_components_dict = dict()
        for idx in set(component_df.index):
            key = idx[:3] + tuple(component_df.loc[idx].values.reshape(-1))
            if key in grouped_components_dict:
                grouped_components_dict[key] += [idx[3]]
            else:
                grouped_components_dict[key] = [idx[3]]

        reshaped_components = []
        base_component_config = self.base_component_config(compo_config)
        for key, value in grouped_components_dict.items():
            new_compo = {
                "hazard": key[0],
                "period": key[1],
                "geos": value,
                "hazardName": key[2],
                "levels": [],
            }
            new_compo.update(base_component_config)
            levels_indices = component_df.loc[
                (key[0], key[1], key[2], value[0])
            ].values.reshape((-1, 3))
            for hidx, lidx, cidx in levels_indices:
                current_level = compo_config["data"]["hazards"][hidx]["levels"][lidx]
                current_config = current_level["configs"][cidx]
                current_level_config = {"level": current_level["level"]}
                current_level_config.update(current_config["dataModel"])
                new_compo["levels"] += [current_level_config]

            reshaped_components += [new_compo]
        LOGGER = LOGGER.try_unbind("compo_id", "hazard_id")
        return reshaped_components

    def reshape_text_component(self, compo_config):
        """reshape_text_component: Transform a text component configuration
        into a Promethee's structure given a component configuration with
        a Metronome structure.
        TO DO : Explain changes between component's configurations structures

        Args:
            compo_config (dict): Component's configuration
                (with a Metronome's structure)

        Returns:
            list of dict : List of component's configuration following
                Promethee's structure
        """
        component_df = pd.DataFrame(
            columns=["period", "geo", "weather_idx", "config_idx"]
        )
        for weather_idx, weather in enumerate(compo_config["data"]["weather"]):
            for config_idx, config in enumerate(weather["configs"]):
                for period in config["periods"]:
                    for geo in config["geos"]:
                        component_df = component_df.append(
                            other={
                                "period": period,
                                "geo": geo,
                                "weather_idx": weather_idx,
                                "config_idx": config_idx,
                            },
                            ignore_index=True,
                        )
        component_df = component_df.set_index(["period", "geo"]).sort_index()
        reshaped_components = []
        base_component_config = self.base_component_config(compo_config)
        for idx in set(component_df.index):
            new_compo = {"period": idx[0], "geos": [idx[1]], "weather": []}
            new_compo.update(base_component_config)
            component_df_idx = component_df.loc[idx]
            if isinstance(component_df_idx, pd.Series):
                current_weather = compo_config["data"]["weather"][
                    component_df_idx["weather_idx"]
                ]
                current_config = current_weather["configs"][
                    component_df_idx["config_idx"]
                ]
                new_compo["weather"] += [
                    {
                        "id": current_weather["id"],
                        "condition": current_config["dataModel"]["text"],
                    }
                ]
            else:
                for widx, cidx in component_df.loc[idx].values:
                    current_weather = compo_config["data"]["weather"][widx]
                    current_config = current_weather["configs"][cidx]
                    new_compo["weather"] += [
                        {
                            "id": current_weather["id"],
                            "condition": current_config["dataModel"]["text"],
                        }
                    ]
            reshaped_components += [new_compo]

        return reshaped_components

    def list_components_configs(self, prod_idx):
        """list_components_config : list all the components configurations
        contained in a prod_idx configuration.

        Args:
            prod_idx (int): Production index in the self.global_config
        """
        global LOGGER
        LOGGER = LOGGER.bind(prod_id=self.global_config[prod_idx].get("id", prod_idx))
        components_list = []
        for compo_config in self.global_config[prod_idx]["components"]:
            # On va rajouter les clés de production si elles sont présentes
            # Pour l'instant on ne sait pas si elles sont au niveau du composant ou au
            # dessus.
            compo_config.setdefault(
                "production_id",
                self.global_config[prod_idx].get(
                    "production_id", "UnknownProductionID"
                ),
            )
            compo_config.setdefault(
                "production_name",
                self.global_config[prod_idx].get(
                    "production_name", "UnknownProductionName"
                ),
            )
            try:
                LOGGER = LOGGER.bind(compo_id=compo_config.get("id"))
                compo_type = compo_config["data"]["type"]
                if compo_type == "risk":
                    components_list += self.reshape_risk_component(compo_config)
                elif compo_type == "text":
                    components_list += self.reshape_text_component(compo_config)
                else:
                    raise ConfigurationError(
                        f"Unexpected component type : {compo_type}.",
                        func="list_components_configs",
                    )
            except ConfigurationError:
                LOGGER.error(
                    "Configuration Error caught.",
                    func="list_components_configs",
                    exc_info=True,
                )
            except BaseException:
                LOGGER.error(
                    "Exception caught.", func="list_components_configs", exc_info=True
                )

        LOGGER = LOGGER.try_unbind("prod_id", "compo_id")
        return components_list


if __name__ == "__main__":
    cfp = ConfigMetronomeProcessor(
        "../examples/remotes/BDPE/new_14797.tgz", "default", datetime.today()
    )
    cfp.process_all()

    with open("mask.json", "w") as f:
        json.dump(cfp.mask_config, f, **JSON_KWARGS)

    with open("data.json", "w") as f:
        json.dump(cfp.data_config, f, **JSON_KWARGS)

    with open("prod.json", "w") as f:
        json.dump(cfp.prod_config, f, **JSON_KWARGS)
