"""
@package
Interpretor module

History :
   V. Chabot : Ajout d'un "cover_period" sur les risques et les évènements.
        Permet d'avoir la période couverte par un evt
        (via les metadonnées des fichiers concernés)
        Suppose que la dimension temporelle est "valid_time".
   V. Chabot : Ajout d'un changement de selector pour les fields présents.
        Si les champs existent déjà dans la selection, les remplace.
        Sinon, les rajoute.
"""
from typing import Union
import copy
import operator
import xarray as xr
from ftap_autom.promethee.src.conf.app_configuration import MFLOG
from ftap_autom.promethee.src.utils.hash import MD5
from ftap_autom.promethee.src.utils.xr_utils import ARRAY_LOADER
import mflog

LOGGER = mflog.getLogger("interpret.mod")
mflog.set_config(**MFLOG)
LOGGER = LOGGER.bind(module="interpret")


COMPUTE_LIST = ["density", "extrema", "representative", "summary"]

COMPARISON_OP = [">", "sup", ">=", "supegal", "<", "inf", "<=", "infegal"]
comp_op = {
    ">": operator.gt,
    "sup": operator.gt,
    ">=": operator.ge,
    "supegal": operator.ge,
    "<": operator.lt,
    "inf": operator.lt,
    "<=": operator.le,
    "infegal": operator.le,
    "==": operator.eq,
    "egal": operator.eq,
    "!=": operator.ne,
    "dif": operator.ne,
    "isin": lambda da, l: da.isin(l),
}


def or_operator(first_field, second_field):
    """
       Prend le maximum des champs présents (en chaque point)
    """
    if first_field.count() == second_field.count():
        res = operator.or_(first_field, second_field)
    else:
        LOGGER.warning(" On passe par le max pour simuler le or")
        dfirst = first_field.expand_dims("place").assign_coords(place=[1])
        dsecond = second_field.expand_dims("place").assign_coords(place=[2])
        concat = xr.concat([dfirst, dsecond])
        res = concat.max("place")
    return res


def and_operator(first_field, second_field):
    """
    On ne fait rien pour l'instant.
    A voir comment on doit traiter ce cas
    """
    return operator.and_(first_field, second_field)


def update_next_critical(dict_comparison, d2_comparison, variable, kind="plain"):
    """
    Cette fonction update le dictionnaire qu'on envoit ensuite vers le module de texte.
    Il permet de rajouter une info (next_critical) à savoir l'information
    sur le seuil du niveau supérieur.
    Si une valeur next_critical existe, on regarde si la valeur de ce niveau là est
    moins critique.

    Args:
        dict_comparison (dict): Le dictionnaire du niveau de risque à mettre à jour
        d2_comparison (dict): Le dictionnaire d'un niveau de risque supérieur
        variable (str): La variable qui nous intéresse
        kind (str, optional): Le type ("plain"/"mountain"). Defaults to "plain".
    """
    var1 = dict_comparison[variable]
    varOther = d2_comparison[variable]
    operator = var1[kind]["comparisonOp"]
    if (
        kind in varOther
        and operator in COMPARISON_OP
        and operator == varOther[kind]["comparisonOp"]
    ):
        # Dans ce cas, on va vérifier que la valeur est plus
        # critique que celle de base et moins critique que la suivante
        this_critical = float(var1[kind].get("threshold"))
        next_critical = float(varOther[kind].get("threshold"))
        # On verifie que la valeur est plus critique que celle actuelle
        if comp_op[operator](next_critical, this_critical):
            # LOGGER.warning("On est plus critique que la valeur du niveau ")
            if "next_critical" in var1[kind]:
                # LOGGER.info("Il exsite deja une valeur")
                current_critical = float(var1[kind].get("next_critical"))
                # S'il existe deja une valeur next_critical.
                # Si la nouvelle valeur critique est moins critique.
                if comp_op[operator](current_critical, next_critical):
                    dict_comparison[variable][kind]["next_critical"] = next_critical
            else:
                dict_comparison[variable][kind]["next_critical"] = next_critical


# logical_op = {"and": operator.and_, "or": operator.or_}
logical_op = {"and": and_operator, "or": or_operator}

# Identifiants pertinents
ID_KEY = [
    "mask_hash",
    "config_hash",
    "prod_hash",
    "hazard",
    "prod_id",
    "id",
    "bulletin",
    "customer",
    "production_datetime",
]


def get_bi_dico(son, father, key, default=None):
    """get_bi_dico
    Fonction qui permet de chercher l'information dans un premier dictionnaire, puis,
    si cette info n'est pas présente dans un second dictionnaire.

    Permet d'implémenter un "get" à bidictionnaire

    Args:
        son (dict): Dictionnaire le plus imbriqué
        father (dict): Information venant du niveau du dessus
        key (str): Clé recherchée
        default ([type], optional): Retour par défaut si rien n'est trouvé.
            Defaults to None.

    Returns:
        [type]: [description]
    """
    res = default
    if key in son:
        res = son.get(key)
    elif key in father:
        res = father.get(key)
    return res


class EventInterp:
    """
    L'interpreteur d'evenement.
    """

    son_key = ["aggregation"]
    default = {"compute_list": COMPUTE_LIST}

    def __init__(self, event, father={}):
        self.event = event
        self.father = father
        self.level = get_bi_dico(self.event, self.father, "level")

    def __eq__(self, other):
        """__eq__ Surcharge du egal pour les évènements

        On va virer les "fields" de la comparaison
        (car les fichiers dovient être différents).

        Returns:
            [bool]:
        """
        LOGGER.debug("Entering event comparison. Beta function.", func="__eq__")
        if not isinstance(other, type(self)):
            return NotImplemented

        evt1 = copy.deepcopy(self)
        evt2 = copy.deepcopy(other)
        for var in vars(evt1):
            if var == "event":
                v1 = evt1.__getattribute__(var)
                v2 = evt2.__getattribute__(var)
                # On enleve les fields (qu'on ne compare pas)
                pop_k = [key for key in v1 if "field" in key]
                for key in pop_k:
                    v1.pop(key)
                    v2.pop(key, None)
            else:
                v1 = evt1.__getattribute__(var)
                v2 = evt2.__getattribute__(var)
            if v1 != v2:
                return False
        return True

    def __hash__(self):
        """__hash__
        Surcharge du hash.
        Va permettre de faire des sets d'EventInterp

        Returns:
             [int]: Le nombre correspondant à la clé de hashage
        """
        dict_var = {}
        for var in vars(self):
            if var == "event":
                v1 = self.__getattribute__(var)
                pop_k = [key for key in v1 if "field" in key]
                for key in pop_k:
                    v1.pop(key)
                dict_var[var] = v1
            else:
                dict_var[var] = self.__getattribute__(var)
        obj = MD5(dict_var)
        return hash(obj.hash)

    def update(self, new_dict):
        """
        Permet d'updater l'event avec un nouveau dictionnaire
        Args:
            new_dict (dict) : Le dictionnaire pour updater
        """
        self.event.update(new_dict)

    def setdefault(self, new_dict):
        for key in new_dict:
            self.event.setdefault(key, new_dict[key])

    @staticmethod
    def list_optional_key():
        dict_opt = {
            "process": "The kind of event. Default : None",
            "compute_list": (
                "The list of element we need to compute on top of risk occurrence. "
                "Default : everything is computed."
            ),
            "category": (
                "Variable category. Could be qualtitative/quantitative/... "
                "Default None."
            ),
        }
        return dict_opt

    def generic_get(self, key, default=None):
        """
        Permet d'avoir un getter generic.
        On va chercher dans les clés de l'event et les clés du pere.

        Args:
            key (str): La clé qu'on recherche
            default (type, optional): La valeur par défaut de la clé.

        Returns:
            La valeur correspondant à la clé
        """
        if default is None:
            default = self.default.get(key)
        if hasattr(self, "get_" + key):
            res = self.__getattribute__("get_" + key)()
            if res is None:
                res = default
        elif key in self.son_key:
            res = self.event.get(key, default)
        else:
            res = get_bi_dico(self.event, self.father, key, default)
        return res

    def doing_aggregation(self):
        res = False
        if "aggregation" in self.event.keys() and self.event["aggregation"] is not None:
            res = True
        return res

    def altitude_condition(self):
        res = False
        if self.get_mountain() is not None:
            res = True
        return res

    def get_mountainThreshold(self):
        """get_mountainThreshold Get the moutainThreshold

        Enable to get it event if he is inside altitude block.

        Returns:
            [int]: The altitude threshold (in m) above which we are in mountain.
        """
        res = None
        if "mountainThreshold" in self.event:
            res = self.event.get("mountainThreshold")
        elif "altitude" in self.event:
            res = self.event["altitude"][0].get("mountainThreshold")
        else:
            LOGGER.debug("No mountain threshold set", func="get_mountainThreshold")
        return res

    def get_compute_list(self):
        """get_compute_list

        Return the list of element to compute.
        This function check that risk module is able to compute things and raise
        an error otherwise.

        Raises:
            AttributeError: Raise an error if an element of compute_liste is not part
            of the the posibilities

        Returns:
            [list] : return things to compute. By default everything is computed.
        """
        compute_list = get_bi_dico(
            self.event, self.father, "compute_list", COMPUTE_LIST
        )
        for x in compute_list:
            if x not in COMPUTE_LIST:
                raise AttributeError(
                    f"The risk module is not able to compute this {x}. "
                    f"Possibile options are {COMPUTE_LIST}."
                )
        return compute_list

    def get_mountain(self):
        """get_mountain
        Ici on va en plus checker que l'evenement est bien une condition
        sur l'altitude.
        Sinon ça n'a pas de sens d'avoir une définion de seuil en montagne.

        Returns:
            [dict]: Le dictionnaire décrivant l'evenement en montagne
        """
        res = None
        if self.generic_get("mountainThreshold") is not None:
            res = self.event.get("mountain")
        return res

    def get_aggregation_aval(self):
        """get_aggregation_aval
        Retourne l'aggrégation au niveau du risque (si elle existe).
        Permet quand l'aggrégation est définie au niveau du risque de calculer
        des valeurs équivalentes.

        Returns:
            [dict]: Retourne l'aggrégation si celle-ci est au niveau du risque.
        """
        return self.father.get("aggregation", None)

    def get_cover_period(self):
        """Return the period cover by the event.
        To do so, we will need to open the DataArray
        """
        ds = ARRAY_LOADER(self.event["field"])
        times = copy.deepcopy(ds[self.generic_get("time_dimension", "valid_time")])
        ds.close()
        return times

    def get_grid_name(self):
        # On va implémenter différentes manière d'avoir le nom de la grille
        # 1. A partir d'une clé  dans la configuration du field
        if "grid" in self.event["field"]:
            grid_name = self.event["field"].get("grid")
        else:
            ds = ARRAY_LOADER(self.event["field"])
            if hasattr(ds, "PROMETHEE_z_ref"):
                grid_name = ds.attrs["PROMETHEE_z_ref"]
            else:
                grid_name = "Unknown"
                LOGGER.warning(
                    "Unknown grid_name : grid name is neither in the conf"
                    " nor in the key (PROMETHEE_Z_REF)",
                    func="get_grid_name",
                )
            ds.close()
        return grid_name

    def get_field_name(self) -> Union[str, None]:
        """
        Try to get the field name.

        Returns:
            Union[str, None]: The name of the field
        """

        if isinstance(self.event["field"], dict):
            full_path = self.event["field"].get("file", "")
            l_name = full_path.split("/")
            variable = l_name[-1].split(".")[0]
        elif isinstance(self.event["field"], xr.DataArray):
            variable = self.event["field"].name
        else:
            LOGGER.error(
                "field is neither a dictionnary nor a dataArray instance "
                f"{type(self.event['field'])}"
            )
            variable = None
        return variable

    def get_comparison(self) -> dict:
        """
        Enable to get comparison operator for an event

        Returns:
            dict: Dictionnary of comparison operator (on plain or mountain)
        """
        dict_out = {}
        if "plain" in self.event:
            dict_out["plain"] = self.event["plain"]
            dict_out["plain"]["category"] = self.event.get("category")
        if "mountain" in self.event:
            dict_out["mountain"] = self.event["mountain"]
            dict_out["mountain"]["category"] = self.event.get("category")
        dict_out["category"] = self.event.get("category")
        dict_out["mountain_threshold"] = self.get_mountainThreshold()
        # On recupere la fonction d'aggregation. Permettra les controles plus tard.
        if "aggregation" in self.event:
            agg = self.event.get("aggregation")
        else:
            agg = self.get_aggregation_aval()
        dict_out["aggregation"] = agg
        return dict_out

    def update_field_sel(self, new_dict):
        """
        Cette fonction permet d'updater la seleciton lors de l'ouverture du fichier
        (si on ne travail que sur une partie de l'evenement).

        Args:
            new_dict ([dict]): Dictionnaire décrivant la nouvelle sélection

        """
        key_sel = set(["sel", "isel", "islice", "slice", "td_slice"])
        # On va recupere la clé des champs (Field, Field_1) en théorie
        field_key = [key for key in self.event.keys() if key.startswith("field")]
        # On va aussi recuperer toutes les dimensions qu'on doit changer
        l_change = []
        [l_change.extend(new_dict[key].keys()) for key in new_dict.keys()]
        new_selector = new_dict.keys()
        # On vérifie que chacune des clés est présent dans key_sel (qui sont les
        # seuls valides)
        if set(new_selector).intersection(key_sel) != set(new_selector):
            difference = set(new_selector).difference(key_sel)
            raise ValueError(
                f"New dictionnary contains key undesirable for selection {difference}."
            )
        # Pour chacun des champs, on va maintenant :
        # 1. Enlever les clés de l_change déjà présente
        #       dans les differents elts de key_sel
        # 2. Rajouter les clés de selection appropriés
        for field in field_key:
            selector_present = set(self.event[field].keys()).intersection(key_sel)
            for selector in selector_present:
                # On regarde s'il y a une clé à enlever et on enleve
                current_dim = set(self.event[field][selector].keys())
                dim_to_delete = current_dim.intersection(set(l_change))
                if dim_to_delete == current_dim:
                    # Si un niveau n'a plus de clé
                    del self.event[field][selector]
                else:
                    # Si on enleve qu'une partie des clés
                    for dim in current_dim:
                        del self.event[field][selector][dim]
            for selector in new_selector:
                # On regade s'il y a une clé à ajouter
                if selector in self.event[field]:
                    self.event[field][selector].update(new_dict[selector])
                else:
                    self.event[field][selector] = new_dict[selector]


class RiskInterp:
    """
    Classe permettant l'interprétation d'un risk
    We will use all the information from the configuration
    """

    son_key = []
    default = {"compute_list": COMPUTE_LIST}

    def __init__(self, risk, father={}):
        """__init__

        Args:
            risk (dict): The risk configuration
            father (dict): The alea configuration (default = {})
        """
        self.risk = risk
        self.nb_event = self.check_nbElementaryEvent()
        # On va maintenant définir des events pour chacun des evenements dans la liste
        self.father = father

        self.list_event = self._define_event()

    def _define_event(self):
        risk_copy = copy.deepcopy(self.risk)
        risk_copy.pop("elementsEvent", [])
        for key in self.father:
            risk_copy.setdefault(key, self.father[key])
        return [
            EventInterp(event, risk_copy)
            for event in self.risk.get("elementsEvent", [])
        ]

    def get_singleEvt_comparison(self) -> Union[dict, None]:
        """
        Enable to get, for a single event, the comparison operator

        Returns:
            Union[dict, None]: A list of comparison operator.
                None if there is several event.
        """
        res = None
        if len(self.list_event) == 1:
            res = self.list_event[0].get_comparison()
        else:
            #  On verifie que les evenement dans la liste ne sont pas identique,
            # i-e ils différent seulement par les chemins de fichiers.
            evt0 = self.list_event[0]
            all_the_same = True
            for evt in self.list_event:
                if evt != evt0:
                    all_the_same = False
            if all_the_same:
                res = self.list_event[0].get_comparison()
        return res

    def get_comparison(self) -> Union[dict, None]:
        dout = dict()
        for event in self.list_event:
            field = event.get_field_name()
            comparison = event.get_comparison()
            if field is not None and field not in dout:
                dout[field] = comparison
            elif field in dout and dout[field] != comparison:
                LOGGER.error(
                    f" Current  {dout[field]} is different of new one {comparison}. "
                    "Don't know what to do in this case. "
                )
        return dout

    def get_ids(self):
        id_dict = {}
        for x in ID_KEY:
            if x in self.risk:
                id_dict[x] = self.risk.get(x)
            elif x in self.father:
                id_dict[x] = self.father.get(x)
        level = get_bi_dico(self.risk, self.father, "level")
        if level is not None:
            id_dict["riskLevel"] = level
        return id_dict

    def get_cover_period(self):
        return self.list_event[0].get_cover_period()

    def get_grid_name(self):
        """We here make the hypothesis that all file are based on the same grid"""
        return self.list_event[0].get_grid_name()

    def update(self, new_dict):
        self.risk.update(new_dict)
        # On met à jour aussi les events.
        for event in self.list_event:
            event.update(new_dict)

    def update_field_sel(self, new_dict):
        """Permet d'updater la periode de tous les évènements.

        Args:
            new_dict (dict): Dictionnaire de selection
        """
        for event in self.list_event:
            event.update_field_sel(new_dict)

    def setdefault(self, new_dict):
        for key in new_dict:
            self.risk.setdefault(key, new_dict[key])

        for event in self.list_event:
            event.setdefault(new_dict)

    def __eq__(self, other):
        """__eq__ Surcharge de l'egalité entre deux risks.
        On vire la liste "elementsEvent" de la comparaison.
        Par contre on va aussi checker que la liste d'EventInterp soit concordante.

        Beta function ....
        Does it really works when we have a list of different ElementaryEvent ?
        Does it required that the list is "sorted' ?

        Args:
            other ([type]): [description]

        Returns:
            [type]: [description]
        """
        LOGGER.debug("Entering risk comparison. Beta function", func="__eq__")
        res = True
        elt1 = copy.deepcopy(self)
        elt2 = copy.deepcopy(other)
        elt1.risk.pop("elementsEvent")
        elt2.risk.pop("elementsEvent")
        for var in vars(elt1):
            v1 = elt1.__getattribute__(var)
            v2 = elt2.__getattribute__(var)
            if var == "list_event":
                # On va faire quelque chose de special sur la liste.
                # On va comparer deux listes d'évènements uniques triés par niveaux
                # Ne fonctionne pas s'il y a des doublons de niveau ...
                # Dans ce cas pour un même niveau ca arrive dans l'ordre d'apparition
                v1 = sorted(list(set(v1)), key=operator.attrgetter("level"))
                v2 = sorted(list(set(v2)), key=operator.attrgetter("level"))

            if v1 != v2:
                res = False
        return res

    def generic_get(self, key, default=None):
        if default is None:
            default = self.default.get(key)
        if hasattr(self, "get_" + key):
            return self.__getattribute__("get_" + key)()
        elif key in self.son_key:
            return self.risk.get(key, default)
        else:
            return get_bi_dico(self.risk, self.father, key, default)

    def get_compute_list(self):
        """get_compute_list

        Return the list of element to compute.
        This function check that risk module is able to compute things and raise
        an error otherwise.

        Raises:
            AttributeError: Raise an error if an element of compute_liste
                (in configuration) is not part of the the posibility

        Returns:
            [list] : return things to compute. By default everything is computed.
        """
        compute_list = get_bi_dico(self.risk, self.father, "compute_list", COMPUTE_LIST)
        for x in compute_list:
            if x not in COMPUTE_LIST:
                raise AttributeError(
                    f"The risk module is not able to compute this {x}. "
                    f"Possibile options are {COMPUTE_LIST}."
                )
        return compute_list

    @staticmethod
    def list_optional_key():
        dict_opt = {
            "compute_list": (
                "The list of element we need to compute on top of risk occurrence. "
                "Default : everything is computed."
            ),
            "level": (
                "Information about the risk level. Used only for log message. "
                "Default : None."
            ),
        }
        return dict_opt

    def check_nbElementaryEvent(self):
        nb_t = len(self.risk.get("elementsEvent", []))
        if "nbElementaryEvent" in self.risk:
            nb = self.risk.get("nbElementaryEvent")
            if nb != nb_t:
                raise AttributeError(
                    f"The number of elements ({nb}) is not consistent "
                    f"with the len of the list ({nb_t})."
                )
        return nb_t

    def get_nbElementaryEvent(self):
        return self.nb_event

    def get_logicalOpList(self):
        """get_logicalOpList
        Return the list of logical operator (and/or)
        Raises:
            AttributeError: Raise an error if the lenght is not consistant
                with elementEvents len.

        Returns:
            [list]: list of logical operator
        """
        logical_op = self.risk.get("logicalOpList", [])
        nb_elt = self.get_nbElementaryEvent()
        if len(logical_op) != nb_elt - 1:
            raise AttributeError(
                f"The number of logical operator ({len(logical_op)}) is not consistent"
                f" with the len of element list (n={nb_elt}. Should be {nb_elt-1}."
            )
        return logical_op

    def get_event(self, position=0):
        """
        Permet de retourer l'evenement en position fournie
        Returns:
            EventInterp
        """
        return self.list_event[position]

    def is_bertrand(self):
        """Check if an event of the list is Bertrand Kind."""
        for event in self.list_event:
            if event.generic_get("process") == "Bertrand":
                return True
        return False


class AleaInterp:
    """Classe permettant de suivre la config d'un alea."""

    son_key = []
    default = {}

    def __init__(self, component, father={}):
        self.component = component
        self.father = father
        self.risk_level = self._define_risk_level()

    def _define_risk_level(self):
        component_copy = copy.deepcopy(self.component)
        component_copy.pop("levels")
        for key in self.father:
            component_copy.setdefault(key, self.father[key])
        risk_level = [
            RiskInterp(level, component_copy) for level in self.component.get("levels")
        ]
        return risk_level

    def generic_get(self, key, default=None):
        if default is None:
            default = self.default.get(key)
        if key in self.son_key:
            res = self.component.get(key, default)
        else:
            res = get_bi_dico(self.component, self.father, key, default)
        return res

    def get_ids(self):
        id_dict = {}
        for x in ID_KEY:
            if x in self.component:
                id_dict[x] = self.component.get(x)
            elif x in self.father:
                id_dict[x] = self.component.get(x)
        return id_dict

    def get_risk_level(self, position=0):
        """Retourn RiskInterp pour la position fournie

        Args:
            position (int, optional): La position dans la liste des risques.
                Defaults to 0.

        Returns:
            RiskInterp: Le risque
        """
        return self.risk_level[position]

    def update(self, new_dict):
        self.component.update(new_dict)
        # On met a jours les risques
        for risk_level in self.risk_level:
            risk_level.update(new_dict)

    def setdefault(self, new_dict):
        for key in new_dict:
            self.component.setdefault(key, new_dict[key])
        for risk_level in self.risk_level:
            risk_level.setdefault(new_dict)

    def get_levels(self):
        """ Cette fonction retourne quels sont les niveaux présents pour le risque.
        """
        level_list = []
        for risk_level in self.risk_level:
            level_list.append(risk_level.generic_get("level"))
        levels = list(set(level_list))
        levels.sort()
        return levels

    def get_comparison(self, level=1):
        """
        On va chercher le dictionnaire de comparaison.

        Args:
            level (int, optional): [description]. Defaults to 1.
        """
        # LOGGER.info(f"Level {level.values}")
        this_level = self.select_risk_level(level=level)
        # LOGGER.warning(f"Niveau accessibles {self.get_levels()}")
        levels = [i for i in self.get_levels() if i > level.values]

        # On recupere le dictionnaire de comparaison du niveau qui nous intéresse
        dict_comparison = this_level[0].get_comparison()

        # Ensuite on va boucler sur chacune des variables et
        # voir si on trouve des variables identiques à la notre
        for variable in dict_comparison.keys():
            for elt in levels:
                other_level = self.select_risk_level(level=elt)
                d2_comparison = other_level[0].get_comparison()
                # LOGGER.warning(f"Dictionnaire 1 {dict_comparison}")
                # LOGGER.warning(f"Dictionnaire 2 {d2_comparison}")
                if variable in dict_comparison and variable in d2_comparison:
                    if "plain" in dict_comparison[variable]:
                        update_next_critical(
                            dict_comparison,
                            d2_comparison,
                            variable=variable,
                            kind="plain",
                        )
                    if "mountain" in dict_comparison[variable]:
                        update_next_critical(
                            dict_comparison,
                            d2_comparison,
                            variable=variable,
                            kind="mountain",
                        )
        return dict_comparison

    def select_risk_level(self, level=1):
        """select_risk_level

        Retourne une liste de riskInterp.
        Les éléments de cette liste correspondent tous au niveau demandé
        (plusieurs définition sont possible quand on utilise différents fichiers).

        Args:
            level (int, optional): Le niveau de risque requis. Defaults to 1.

        Returns:
            [list]: Liste de riskInterp.
        """
        res_list = []
        level_list = []
        for risk_level in self.risk_level:
            if risk_level.generic_get("level") == level:
                res_list.append(risk_level)
            level_list.append(risk_level.generic_get("level"))
        if res_list == []:
            LOGGER.warning(
                f"There is no level corresponding to {level}. "
                f"Possible levels are {set(level_list)}",
                level=level,
                func="select_risk_level",
            )
            raise ValueError("No level {level} found")
        return copy.deepcopy(res_list)


class AleaFromList(AleaInterp):
    """
    Permet la definition d'un alea a partir d'une liste de risk
    """

    def __init__(self, risk_list, father={}):
        self.component = {}
        self.father = father
        self.risk_level = risk_list
