import os
from datetime import datetime
from io import BytesIO
import tarfile
import json
import imageio
import mflog

from ftap_autom.promethee.src.lib.localisation.visu_localisation import (
    visualize_as_table,
    plot_localisation,
)
from ftap_autom.promethee.src.conf.app_configuration import (
    MFLOG,
    WORKING_DIR,
    DATETIME_FMT_CDP,
    DATETIME_FMT,
)

# Logging
log = mflog.getLogger("outputLocalisation.mod")
mflog.set_config(**MFLOG)
log = log.bind(module="outputLocalisation")


class OutputLocalisationFormatter:
    def __init__(self, component_config, locHandler, risk_level, max_risk):
        # On suppose que la localisation a tourner sur un unique "id"
        self.component_config = component_config
        self.locHandler = locHandler
        self.final_areas = self.locHandler.get_localised_area()
        self.risk_level = risk_level
        self.max_risk = max_risk

    @property
    def formatted_component(self):
        component_base = {
            "prod_id": self.component_config.generic_get("prod_id"),
            # A modifier pour utiliser "PeriodId" quand ce dernier sera present
            # dans le fichier de config.
            "Period": {
                "PeriodId": self.component_config.generic_get("period").get("id"),
                "DateDebutPeriode": datetime.strptime(
                    self.component_config.generic_get("period").get("start"),
                    DATETIME_FMT,
                ).strftime(DATETIME_FMT_CDP),
                "DateFinPeriode": datetime.strptime(
                    self.component_config.generic_get("period").get("stop"),
                    DATETIME_FMT,
                ).strftime(DATETIME_FMT_CDP),
            },
            "ComponentName": self.component_config.generic_get("name"),
            "CustomerId": self.component_config.generic_get("customer"),
            "CustomerName": self.component_config.generic_get(
                "customerName", "GenericCustomer"
            ),
            "HazardName": self.component_config.generic_get("hazardName"),
        }
        component_base["Period"]["PeriodName"] = (
            "Du "
            + component_base["Period"]["DateDebutPeriode"]
            + " au "
            + component_base["Period"]["DateFinPeriode"]
        )

        dout = self.locHandler.get_spatial_table()
        # On selectionne le champ correspondant au premier evenement du risque.
        # On recupere le fichier gif.
        # Ce fichier devra être sauvé.
        log.warning("Before generating gif file. Could take time")
        component_base["image"] = plot_localisation(
            self.locHandler.full_risk, self.final_areas, dout, save=False
        )
        log.warning("Gif file generated (in cache)")
        dr = (
            self.component_config.risk_level[0]
            .generic_get("aggregation")
            .get("kwargs")
            .get("dr")
        )
        xl_file = BytesIO()
        self.max_risk.name = "Risque"
        visualize_as_table(
            self.max_risk,
            dout,
            unique_table=self.locHandler.get_unique_table(),
            unique_table_name=self.locHandler.get_unique_name(),
            fname=xl_file,
            thresh_dens=dr,
        )
        component_base["excel"] = xl_file
        return component_base


class OutputLocalisationExporter:
    def __init__(self, formatted_components, working_dir="./"):
        conf_file = os.path.join(working_dir, WORKING_DIR["configs"]["version"])
        if os.path.exists(conf_file):
            with open(conf_file) as fp:
                version = json.load(fp)

        else:
            log.error("No version file found. Filling to be robust")
            version = {
                "reference_datetime": datetime(1901, 1, 1).strftime(DATETIME_FMT),
                "configuration_datetime": datetime(1901, 1, 1).strftime(DATETIME_FMT),
            }
        self.date_bulletin = datetime.strptime(
            version["reference_datetime"], DATETIME_FMT
        ).strftime("%Y%m%d")
        # Fixer encore de facon brutale. Peut être la supprimer ?
        self.date_config = datetime.strptime(
            version.get("configuration_datetime", "19710110T000000"), DATETIME_FMT
        ).strftime(DATETIME_FMT_CDP)
        self.date_production = datetime.today().strftime(DATETIME_FMT_CDP)
        self.output_dir = os.path.join(working_dir, WORKING_DIR["output"])
        os.makedirs(self.output_dir, exist_ok=True)
        self.formatted_components = formatted_components

    def export_all(self):
        list_dir = []
        for component in self.formatted_components:
            output_dir = os.path.join(
                self.output_dir, component["CustomerName"], self.date_bulletin
            )
            list_dir.append(output_dir)
            os.makedirs(output_dir, exist_ok=True)
            gif_name = os.path.join(
                output_dir,
                "{}_{}_{}_{}.gif".format(
                    component["ComponentName"],
                    component["HazardName"],
                    component["CustomerName"],
                    self.date_bulletin,
                ),
            )
            excel_name = os.path.join(
                output_dir,
                "{}_{}_{}_{}.xls".format(
                    component["ComponentName"],
                    component["HazardName"],
                    component["CustomerName"],
                    self.date_bulletin,
                ),
            )
            # Sauvegarde du fichier gif
            imageio.mimsave(gif_name, component["image"], fps=1)
            # On relit le excel et on le sauve
            xl_file = component["excel"]
            xl_file.seek(0)
            xl_info = xl_file.read()
            # Sauvegarde du fichier excel
            with open(excel_name, "wb") as f:
                f.write(xl_info)
        # On archive les fichiers creer
        archive_name = os.path.join(self.output_dir, "promethee_msb_poc.tgz")
        with tarfile.open(archive_name, "w:gz") as tar:
            for dirname in list_dir:
                full_path = os.path.abspath(dirname)
                base_name = os.path.basename(dirname)
                tar.add(full_path, arcname=base_name)
