import copy
import json
import os

import mflog
import xarray as xr
from ftap_autom.promethee.src.lib.localisation.core_spatial import (
    get_n_area,
    LocalisationError,
)
from ftap_autom.promethee.src.utils.xr_utils import MASK_LOADER

from ftap_autom.promethee.src.conf.app_configuration import MFLOG
from ftap_autom.promethee.src.lib.interpretor import AleaFromList, RiskInterp
import ftap_autom.promethee.src.lib.parameter_generator as pg
import ftap_autom.promethee.src.lib.risk_generator as rg

LOGGER = mflog.getLogger("localisation")
mflog.set_config(**MFLOG)
LOGGER = LOGGER.bind(module="localisation")


class SpatialIngredient:
    """
    Object to store several information on spatial localisation.
    Can be saved and loaded.
    """

    def __init__(
        self,
        domain: xr.DataArray,
        full_area_list: xr.DataArray,
        localised_area: xr.DataArray,
        grid_name: str,
    ):
        self.domain = domain
        self.full_area_list = full_area_list
        self.localised_area = localised_area
        self.grid_name = grid_name

    @classmethod
    def load(cls, basename: str):
        """
        Enable to load "SpatialIngredient" from an autosaved.

        Args:
            basename (str): Base du nom de fichier de sauvegarde
                (incluant le path) sans extension

        Returns:
            [type]: [description]
        """
        print("In spatial ingredient, basename is ", basename)
        with open(basename + ".json", "r") as fp:
            geodict = json.load(fp)
        grid_name = geodict["geos"]["grid_name"]
        mask_a = MASK_LOADER(geodict["geos"])
        domain = mask_a.sel(id=geodict["geos"]["domain_id"])
        full_area_list = mask_a.sel(id=geodict["geos"]["full_list_id"])
        localised_area = MASK_LOADER(geodict["localized"])
        return cls(domain, full_area_list, localised_area, grid_name)

    def auto_save(self, basename: str):
        """

        Args:
            basename (str): Base du nom de fichier de sauvegarde
                (incluant le path) sans extension
        """
        LOGGER.info(f"Basename is {basename}")
        fname = basename + ".nc"
        loca_fname = basename + "localized.nc"
        geodict = {}
        geodict["geos"] = {
            "file": fname,
            "grid_name": self.grid_name,
            "domain_id": self.domain.id.values.tolist(),
            "full_list_id": self.full_area_list.id.values.tolist(),
        }
        geodict["localized"] = {
            "file": loca_fname,
            "grid_name": self.grid_name,
        }
        dmerged = xr.concat([self.domain, self.full_area_list], dim="id")
        if os.path.exists(fname):
            os.remove(fname)
        dmerged.to_netcdf(fname)
        if os.path.exists(loca_fname):
            os.remove(loca_fname)
        self.localised_area.to_netcdf(loca_fname)
        with open(basename + ".json", "w") as fp:
            json.dump(geodict, fp)


class SpatialLocalisation:
    """
    This module is responsible of the spatial localisation.

    Usage :
    1. Init
    2. Localise :
        After initialisation, you can use the localise() function.
        This require to know which risk_level is of interest and which
        is the period of the risk.
    3. compute_information_on_area
    """

    def __init__(self, alea_conf, geo_id, between_authorized=False):
        """
        Initialisation

        Args:
            alea_conf (AleaInterp): An alea interpretor.
            geo_id (str): The id of the area to consider.
               This "id" should be present in the mask configuration.
            between_authorized (bool, optional):
                This say if we let the spatialLoclaisation alogrithm choose the.
                    Defaults to False.

        Others variables initialised :
           - full_risk (DataArray): A 3D (lat/lon/time) DataArray of the current
                level of risk
           - domain (DataArray) : A 2D (lat/lon) DataArray telling us what is
                the current domain
           - best_element (RiskInterp) : The interpretor for the risk used for
                localisation purpose.
                Indeed, as there is several sources of risk,
                several definition can be found for the same level
                (covering different period).
                This element indicates which one is chosen.
        """
        self.alea_conf = alea_conf
        self.time_dimension = alea_conf.generic_get("time_dimension")
        self.geo_id = geo_id
        self.between_authorized = between_authorized

        if self.time_dimension is None:
            mflog.info(
                "Time dimension has been set to default to valid_time."
                "Please check if this is correct."
            )
            self.time_dimension = "valid_time"
            # On rajoute une valeur par défault
            alea_conf.setdefault({"time_dimension": self.time_dimension})

        # On initialise avec une coquille vide
        self.full_risk = xr.DataArray()
        self.best_elt = RiskInterp({}, {})

    def localise(self, risk_level, risk_period: set):
        """
        This functionnality enable to localise a risk.

        Args:
            risk_level (int): The risk_level wanted for localisation.
                This level should exist in the configuration.
            risk_period (set): This set indicate which timestep should be used
                for the localisation.

        Raises:
            LocalisationError: If the geo_id is not present in RiskInterp.
            LocalisationError: If the risk is upstream

        Returns:
            [xr.DataArray]: The localisedArea.
        """

        # On recupere une liste de risque pour un niveau donné
        risk_list = self.alea_conf.select_risk_level(level=risk_level)
        # On recupere le periodes couverte par chaque niveau
        best_elt, common_period = self._find_best_conf_match(risk_list, risk_period)
        agg = best_elt.generic_get("aggregationType")
        if agg != "downStream":
            raise LocalisationError(
                f"We are able to localise only downStream risk. Get {agg}"
            )
        # ==========================
        # On va filtrer pour mettre seulement ceux de la zone qui nous intéresse
        geos_save = best_elt.generic_get("geos")
        geo_new = copy.deepcopy(geos_save)
        if self.geo_id in geos_save.get("mask_id"):
            geo_new["mask_id"] = [self.geo_id]
        else:
            raise LocalisationError("mask_id is not present")
        best_elt.update({"geos": geo_new, "compute_list": []})
        # ==========================
        # On va réadapter la periode.
        # Cela permettra de calculer le risque que sur la période précisée.
        # On va faire cela sauf pour les risques de type Bertrand.
        # Ces derniers requièrent d'avoir l'ensemble du dataset pour être calculés
        # par contre ensuite on va revenir sur la période courte.
        if best_elt.is_bertrand():
            LOGGER.info("Risk is bertrand kind. So we compute all")
        else:
            best_elt.update_field_sel({"sel": {self.time_dimension: common_period}})
        self.best_elt = best_elt
        # ==========================
        # On va cacluler le risk
        risk_handler = rg.RiskGenericEvent(
            risk_conf=best_elt, aggregate_dim=["latitude", "longitude"]
        )
        self.full_risk = risk_handler.return_spatial_information()
        if best_elt.is_bertrand():
            self.full_risk = self.full_risk.sel({self.time_dimension: common_period})
        # log.error(f"Full risk values {self.full_risk.max().values}")
        # ==========================
        # On va maintenant "localiser ce risque", c'est à dire recupérer les
        # "bests zones"
        domain, localisation_area = self._get_poss_localisation_area(best_elt)
        domain_no_id = domain.reset_coords("id", drop=True)
        self.full_risk = self.full_risk.squeeze("id").reset_coords("id", drop=True)
        localised_area = get_n_area(
            self.full_risk,
            domain_no_id,
            localisation_area,
            n=2,
            cdt=0.001,
            between_authorized=self.between_authorized,
        )

        ingredient = SpatialIngredient(
            domain, localisation_area, localised_area, best_elt.get_grid_name()
        )
        return ingredient

    def compute_information_on_area(self, risk_level, da_area):
        """
        This enable the computation of the risk (for a given level)
        over the "localised area".
        This computation is performed for the full period


        Args:
            risk_level (int): The wanted level
            da_area (xr.DataArray): The localised_area.
                Use self.localised_area instead ?

        ToDo :
          Ajouter un moyen d'interpoler entre la grille d'entrée
          et de sortie pour différent niveaux.
          Comment passer les infos ?
             => risk.get_grid_name()
             => self.best_elt.get_grid_name()

        Returns:
            ParameterGeneratorTransit: The parameter compute on the localised_area
        """
        risk_list = copy.deepcopy(self.alea_conf.select_risk_level(level=risk_level))
        for risk in risk_list:
            # On force les zones géographique et les éléments à calculer
            risk.update(
                {"geos": da_area, "compute_list": ["density", "representative"]}
            )
        localiseconf = AleaFromList(
            risk_list, father={"time_dimension": self.time_dimension}
        )
        param_handler = pg.ParameterGeneratorTransit(localiseconf)
        param_handler.compute_risks()
        return param_handler.get_hazard()

    def _get_poss_localisation_area(self, best_elt):
        """Retourne les localisations possibles.


        Args:
            best_elt (RiskInterp): Un element de configuration
                (un peu updater dans localise)

        Raises:
            ValueError: [description]

        Returns:
            [dataArray, DataArray]: Le domain, Les zones de localisation
        """

        geo = best_elt.generic_get("geos")
        grid_name = best_elt.get_grid_name()
        # Chargement des masques en précisant la grille
        geo.update({"grid_name": grid_name})
        full_list = MASK_LOADER(geo)  # , grid_name = grid_name)
        domain = full_list.sel(id=self.geo_id)
        # On va ensuite selectionner les id commençant par geo_id
        id_list = [
            idi
            for idi in full_list.id.values
            if idi.startswith(self.geo_id) and idi != self.geo_id
        ]
        selected_area = full_list.sel(id=id_list)
        drop_ids = []
        # On va rajouter compass_split et altitude_split
        if not best_elt.generic_get("compass_split", True):
            compass_idx = selected_area["areaType"] == "compass"
            drop_ids.extend(selected_area.sel(id=compass_idx).id.values)

        if not best_elt.generic_get("altitude_split", True):
            alt_idx = selected_area["areaType"] == "Altitude"
            drop_ids.extend(selected_area.sel(id=alt_idx).id.values)

        id_list.extend(geo.get("descriptive_id", []))
        id_list.extend(best_elt.generic_get("geos_descriptive", []))
        id_list = list(set(id_list).difference(set(drop_ids)))

        if len(id_list) > 0:
            localisation_area = full_list.sel(id=id_list).dropna("id", how="all")
        else:
            raise LocalisationError(
                "There is no area for localisation process. "
                "So no localisation is performed."
            )
        return domain, localisation_area

    @staticmethod
    def _find_best_conf_match(risk_list, risk_period: set):
        """Enable to find the best "risk" in the list.
            This is based on common period
        Args:
            risk_list ([type]): List of riskInterp
            risk_period ([type]): Period we are interested in
        Returns:
            [riskInterp]: The best element of the list for localisation.
            [list] : common period : The list of period to localise for this risk
        """
        input_period = risk_period
        first_time = None
        best_element = None

        for elt in risk_list:
            elt_period = set(elt.get_cover_period().values)
            intersect = input_period.intersection(elt_period)
            # log.error("Period intersection %s",intersect)
            # Nombre d'élement commun
            nb_common = len(intersect)
            if nb_common > 0 and first_time is None:
                best_element = copy.deepcopy(elt)
                best_common = nb_common
                l_inter = sorted(intersect)
                first_time = l_inter[0]
                common_period = l_inter
            elif nb_common > 0:
                l_inter = sorted(intersect)
                current_first = l_inter[0]
                # Conditions pour detronner le meilleur actuel
                if current_first < first_time and nb_common >= best_common / 4:
                    # On souhaite l'evt le plus jeune (s'il a un nombre de
                    # correspondance importante)
                    best_common = nb_common
                    best_element = copy.deepcopy(elt)
                    common_period = l_inter
                elif current_first > first_time and nb_common >= 4 * best_common:
                    # On est aussi d'accord pour prendre un evt
                    # avec une grille moins résolue.
                    # si le nombre d'échéance dépasse un certain seuil.
                    best_common = nb_common
                    best_element = copy.deepcopy(elt)
                    common_period = l_inter

        if best_element is None:
            raise ValueError("Best conf not found")
        return best_element, common_period
