import six
from datetime import datetime
from io import BytesIO
import matplotlib as mpl
import matplotlib.cm as cm
from matplotlib.colors import rgb2hex, ListedColormap, BoundaryNorm
import matplotlib.pyplot as plt
import imageio
import numpy as np
import xarray as xr
import pandas as pd


def visualize_as_table(
    da_risk,
    ds_subdivision,
    unique_table: xr.DataArray,
    unique_table_name="unique",
    fname="test.xls",
    thresh_dens=0.2,
):
    """[summary]

    Args:
        da_risk ([type]): [description]
        ds_subdivision ([type]): [description]
        unique_table (xr.DataArray): [description]
        unique_table_name (str, optional): [description]. Defaults to "unique".
        fname (str, optional): [description]. Defaults to "test.xls".
        thresh_dens (float, optional): [description]. Defaults to 0.2.
    """
    df = da_risk.to_dataframe().reset_index()
    df = df.drop(["id", "areaType"], axis=1)
    df["date"] = df["valid_time"].dt.strftime("%Y%m%d %H")
    dfp = df.pivot(index="date", columns="areaName", values="Risque")  # .transpose()
    dfp.reset_index(inplace=True)
    dfp = dfp.set_index("date")

    excel_file = fname
    writer = pd.ExcelWriter(excel_file, engine="xlsxwriter")
    export_df_as_excel(
        dfp, writer=writer, sheet_name="Domaine", GlobalText="Risk", risk=3, index=True
    )

    # On traite maintenant le fichier sur les zones
    ds_area = ds_subdivision.swap_dims(
        {"id": "areaName"}
    ).squeeze()  # .isel(valid_time=[1,2,10,21])
    ds_area = ds_area.reset_coords(["id", "areaType"], drop=True)
    # On va le brodcaster vers les dimensions du fichier original
    ds_broadcasted = ds_area.broadcast_like(da_risk)
    # On va maintenant ecrire des choses dans le fichier excel
    df_area = ds_broadcasted.to_dataframe().reset_index()
    df_area["date"] = df_area["valid_time"].dt.strftime("%Y%m%d %H")
    df_area["occurrence"] = df_area["occurrence"].astype(int)
    dfp_occurrence = df_area.pivot(
        index="date", columns="areaName", values="occurrence"
    )
    dfp_density = df_area.pivot(index="date", columns="areaName", values="risk_density")
    export_df_as_excel(
        dfp_occurrence,
        writer=writer,
        sheet_name="Subdivision",
        GlobalText="Occurrence",
        index=True,
        risk=True,
    )
    export_df_as_excel(
        dfp_density,
        add_col=6,
        writer=writer,
        sheet_name="Subdivision",
        GlobalText="Densité (seuil :%s)" % thresh_dens,
        index=True,
        risk=thresh_dens,
    )
    nb_elt = 2
    if "rep_value_plain" in df_area:
        nb_elt = nb_elt + 1
        dfp_plain = df_area.pivot(
            index="date", columns="areaName", values="rep_value_plain"
        )
        export_df_as_excel(
            dfp_plain,
            add_col=3 * nb_elt,
            writer=writer,
            sheet_name="Subdivision",
            GlobalText="Rep value plaine",
            index=True,
        )
    if "rep_value_mountain" in df_area:
        nb_elt = nb_elt + 1
        dfp_mountain = df_area.pivot(
            index="date", columns="areaName", values="rep_value_mountain"
        )
        export_df_as_excel(
            dfp_mountain,
            add_col=3 * nb_elt,
            writer=writer,
            sheet_name="Subdivision",
            GlobalText="Rep value Montagne",
            index=True,
        )

    # On va aussi ecrire des choses pour les tables
    df_unique = unique_table.to_dataframe().reset_index()
    dfp_unique = df_unique.pivot(index="areaName", columns="period", values="elt")
    dfp_unique = dfp_unique.reindex(unique_table.areaName.values)

    export_df_as_excel(
        dfp_unique,
        writer=writer,
        sheet_name="Simple Table",
        GlobalText=unique_table_name,
        index=True,
        risk=True,
    )
    writer.save()


def export_df_as_excel(
    df_new, writer, sheet_name="Sheet1", GlobalText="Trimester value", **kwargs
):
    """
    """
    workbook = writer.book

    sh = df_new.shape
    if GlobalText == "":
        srow = 1
    else:
        srow = 2
    if kwargs.get("index", True):
        scol = 1
    else:
        scol = 0
    if "add_col" in kwargs:
        # Permet de décaler l'écriture
        scol = scol + kwargs["add_col"]
    length = len(df_new.columns)
    if length <= 5:
        largeur = 25
    else:
        largeur = 15
    largeur = kwargs.get("largeur", largeur)

    df_new.to_excel(
        writer,
        sheet_name=sheet_name,
        startrow=srow,
        startcol=scol,
        header=False,
        index=False,
    )
    worksheet = writer.sheets[sheet_name]

    merge_format = workbook.add_format(
        {
            "text_wrap": True,
            "bold": 1,
            "border": 1,
            "align": "center",
            "valign": "vcenter",
            "fg_color": "#668cff",
        }
    )

    header_format = workbook.add_format(
        {
            "text_wrap": True,
            "bold": True,
            "border": 1,
            "valign": "vcenter",
            "fg_color": "#b3c6ff",
        }
    )
    abc = workbook.add_format({"bg_color": "#b3c6ff"})
    red_format = workbook.add_format(
        {"bg_color": "#FFC7CE", "font_color": "#9C0006", "fg_color": "#b3c6ff"}
    )
    # Add a format. Green fill with dark green text.
    green_format = workbook.add_format({"bg_color": "#C6EFCE", "font_color": "#006100"})
    if "risk" in kwargs:
        worksheet.conditional_format(
            0,
            scol,
            sh[0] + srow - 1,
            scol + sh[1] - 1,
            {"type": "blanks", "format": abc},
        )
        worksheet.conditional_format(
            0,
            scol,
            sh[0] + srow - 1,
            scol + sh[1] - 1,
            {
                "type": "cell",
                "criteria": ">=",
                "value": kwargs["risk"],
                "format": red_format,
            },
        )
        worksheet.conditional_format(
            0,
            scol,
            sh[0] + srow - 1,
            scol + sh[1] - 1,
            {
                "type": "cell",
                "criteria": "<",
                "value": kwargs["risk"],
                "format": green_format,
            },
        )
    for col_num, value in enumerate(df_new.columns.values):
        worksheet.write(srow - 1, col_num + scol, value)

    worksheet.write(srow - 1, 0, df_new.index.names[0])

    if kwargs.get("index", True):
        for row_num, value in enumerate(df_new.index.values):
            worksheet.write(row_num + srow, 0, value)
    if GlobalText != "":
        if length > 1:
            worksheet.merge_range(
                0, scol, 0, length - 1 + scol, GlobalText, merge_format
            )
        else:
            worksheet.write(0, scol, GlobalText, merge_format)
    #        worksheet.set_row(0, 55, merge_format)
    worksheet.set_row(srow - 1, 50, header_format)
    worksheet.set_column(0, 0, 20)
    worksheet.set_column(scol, col_num + scol, largeur)
    return writer


def render_mpl_table(
    data_in,
    col_width=5.0,
    row_height=0.625,
    font_size=12,
    header_color="#40466e",
    row_colors=["#f1f1f2", "w"],
    edge_color="w",
    bbox=[0, 0, 1, 1],
    header_columns=0,
    ax=None,
    fname="Test.png",
    hname=["Month"],
    **kwargs
):
    """
    kwargs:
      plt_online
      tot_row
    """
    data = data_in.copy(deep=True)

    plt_online = kwargs.pop("plt_online", False)
    Total_row = kwargs.pop("tot_row", False)
    nrows, _ = data.values.shape

    for col in data.keys():
        if all(isinstance(elem, (np.floating, float)) for elem in data[col].values):
            data[col] = data[col].map("{:,.2f}".format)

    if ax is None:
        size = (np.array(data.shape[::-1]) + np.array([0, 1])) * np.array(
            [col_width, row_height]
        )
        fig, ax = plt.subplots(figsize=size)
        ax.axis("off")

    plt.ioff()
    mpl_table1 = plt.table(cellText=[[""]], colLabels=hname, bbox=[0, 0.9, 1, 0.2])
    mpl_table1.auto_set_font_size(False)
    mpl_table1.set_fontsize(font_size + 2)

    mpl_table = plt.table(
        cellText=data.values, bbox=bbox, colLabels=data.columns, **kwargs
    )
    mpl_table.auto_set_font_size(False)
    mpl_table.set_fontsize(font_size)

    for k, cell in six.iteritems(mpl_table._cells):
        cell.set_edgecolor(edge_color)
        if k[0] == 0 or k[1] < header_columns:
            cell.set_text_props(weight="bold", color="w")
            cell.set_facecolor(header_color)
        elif Total_row and (k == nrows - 1):
            cell.set_facecolor(["orange"])
        else:
            cell.set_facecolor(row_colors[k[0] % len(row_colors)])

    for k, cell in six.iteritems(mpl_table1._cells):
        cell.set_edgecolor(edge_color)
        if k[0] == 0 or k[1] < header_columns:
            cell.set_text_props(weight="bold", color="w")
            cell.set_facecolor(header_color)
        else:
            cell.set_facecolor(row_colors[k[0] % len(row_colors)])
    plt.ioff()

    if plt_online:
        plt.show()
    else:
        fig.savefig(fname)
    plt.close()


def colorbar_definition(N):
    """
    Defini une colorbar specifique
    N : nombre d'element de la colorbar
    """
    name = "Pastel1"
    cmap = cm.get_cmap(name, N)
    newcolors_l = [rgb2hex(x) for x in cmap(range(N))]
    newcolors = xr.DataArray(newcolors_l)
    newcolors_rgb = [cmap(i) for i in range(N)]
    return newcolors, newcolors_rgb


def plot_decompo(areas, fig, ax, legend="id"):
    N = areas.id.size
    ncolors, rgb_colors = colorbar_definition(N)
    newcmp = ListedColormap(ncolors.values)
    l_bounds = sorted(np.arange(N) - 0.5)
    l_bounds.append(N)
    norm = BoundaryNorm(l_bounds, newcmp.N)
    cax = fig.add_axes(
        [
            ax.get_position().x1 * 1.01,
            ax.get_position().y0 + 0.1 * (ax.get_position().y1 - ax.get_position().y0),
            0.05 * (ax.get_position().x1 - ax.get_position().x0),
            0.7 * (ax.get_position().y1 - ax.get_position().y0),
        ]
    )
    # On fait la palette de couleur
    color_id = xr.Dataset()
    color_id["color"] = (("id"), range(N))
    # On effectue le plot
    ((areas * color_id["color"]).sum("id", min_count=1)).plot(
        ax=ax, cmap=newcmp, norm=norm, add_colorbar=False
    )
    bounds = np.linspace(0, N, N + 1)
    ticks_new = bounds[:-1] + 0.5
    cb = mpl.colorbar.ColorbarBase(cax, cmap=newcmp, boundaries=bounds, ticks=ticks_new)
    cb.ax.set_yticklabels(areas[legend].values)
    cb.ax.tick_params(labelsize=16)
    handles = []
    leg = ax.legend(
        handles=handles,
        bbox_to_anchor=(0.5, 1.02),
        loc="lower center",
        labelspacing=2,
        fontsize=12,
        handlelength=0,
        handletextpad=0,
        fancybox=True,
    )
    for line, text in zip(leg.get_lines(), leg.get_texts()):
        text.set_color(line.get_color())


def plot_localisation(
    spatial_risk, areas, d1, field=None, save=True, fgif="my_movie.gif"
):
    """
    Permet de plotter la localisation dans un fig

    Args:
        spatial_risk ([da]): The risk
        areas ([da]): The areas chosen to localise the risk [description]
        d1 ([ds]): Le dataset de sortie sur ces zones
        field ([da], optional): Le champ à partir duquel on crée le risque.
            Defaults to None.
        save (bool, optional): Doit on sauver le resultat sous forme de fichier fig.
            Defaults to True.
        fgif (str, optional): Nom du fichier de sauvegarde. Defaults to "my_movie.gif".

    Returns:
        [type]: [description]
    """
    # On verifie que tout est sur la même grille
    time_sel = set(spatial_risk.valid_time.values).intersection(
        set(d1.valid_time.values)
    )
    l_sel = sorted(time_sel)
    # Determination de ce qu'on va plotter.

    images = []
    for time_s in l_sel:
        image_file = BytesIO()
        fig, axs = plt.subplots(nrows=2, ncols=3, figsize=(20, 15))
        axes = [a for a in axs.flat]

        # spatial_risk
        dtemp = spatial_risk.sel(valid_time=time_s)
        dtemp.plot(ax=axes[0], levels=[0, 0.99, 1.01], add_colorbar=True)
        axes[0].set_title("Risk brute")

        # Densité
        to_plot = areas * d1["risk_density"].sel(valid_time=time_s)

        to_plot = to_plot.sum("id", min_count=1)
        to_plot.plot(
            ax=axes[1], levels=[0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.01]
        )
        axes[1].set_title("Densité")

        # Decompo
        plot_decompo(areas, fig, axes[2], legend="areaName")
        axes[2].set_title("Découpage")
        # Occurrence
        to_plot = areas * d1["occurrence"].sel(valid_time=time_s)
        to_plot.sum("id", min_count=1).plot(
            ax=axes[3], levels=[0, 0.99, 1.01], add_colorbar=False
        )
        axes[3].set_title("Occurrence")
        # Rep1
        if "rep_value_plain" in d1:
            min_v = d1["rep_value_plain"].isel(evt=0).min().values
            max_v = d1["rep_value_plain"].isel(evt=0).max().values
            levels_v = [
                np.round(min_v, 1) + i * np.round(max_v - min_v, 1) / 10
                for i in range(12)
            ]
            to_plot = areas * d1["rep_value_plain"].sel(valid_time=time_s).isel(evt=0)
            to_plot.sum("id", min_count=1).plot(
                ax=axes[4], levels=levels_v, cmap="Reds"
            )
            axes[4].set_title("Valeur representative")
        # Field
        if field is not None:
            max_f = field.min().values
            min_f = field.max().values
            levels_f = [
                np.round(min_f, 1) + i * np.round(max_f - min_f, 1) / 10
                for i in range(12)
            ]
            field.sel(valid_time=time_s).plot(ax=axes[5], levels=levels_f, cmap="Reds")
            axes[5].set_title("Valeurs brutes")
        fig.suptitle(time_s.astype("M8[ms]").astype(datetime).strftime("%Y%m%d - %HH"))
        plt.savefig(image_file)
        plt.close()
        image_file.seek(0)
        images.append(imageio.imread(image_file))
    if save:
        imageio.mimsave(fgif, images, fps=1)
        return None
    return images
