import json
import numpy as np
import xarray as xr


def concat_masks(masks_ncpaths, output_ncpath=None):
    """
    Concatenates a list of netcdf masks (masks_ncpaths) into a single dataarray
    (output_ncpath) and a corresponding dictionnary containing the attributes.
    Parameters :
        masks_ncpaths : list of str
            list of all the netcdf masks paths that are to be concatenated
        output_ncpath : str (optional)
            path to save the output concatenated mask
    Returns :
        output_da : xarray.DataArray
            3D-DataArray (latitude, longitude, id) containing all the masks
        output_dict : dict
            Corresponding dictionnary containing all the attributes
            (keys of this dictionnary are the ids of the output_da)
    """
    data_arrays = [xr.open_dataarray(m).load() for m in masks_ncpaths]
    output_dict = dict()
    for i, da in enumerate(data_arrays):
        key_id = "{:03d}".format(i)
        da["id"] = np.array([key_id])
        output_dict[key_id] = da.attrs
        output_dict[key_id]["origin_filename"] = masks_ncpaths[i].split("/")[-1]

        if "nom" in output_dict[key_id]:
            output_dict[key_id]["label"] = output_dict[key_id]["nom"]
        elif "name" in output_dict[key_id]:
            output_dict[key_id]["label"] = output_dict[key_id]["name"]
        elif "zone" in output_dict[key_id]:
            output_dict[key_id]["label"] = output_dict[key_id]["zone"]
        elif "NUTS_NAME" in output_dict[key_id]:
            output_dict[key_id]["label"] = output_dict[key_id]["NUTS_NAME"]
        elif "LbSousSect" in output_dict[key_id]:
            output_dict[key_id]["label"] = output_dict[key_id]["LbSousSect"]
        elif "LbSecteurH" in output_dict[key_id]:
            output_dict[key_id]["label"] = output_dict[key_id]["LbSecteurH"]
        elif "LbRegionHy" in output_dict[key_id]:
            output_dict[key_id]["label"] = output_dict[key_id]["LbRegionHy"]
        else:
            output_dict[key_id]["label"] = output_dict[key_id][
                "origin_filename"
            ].rstrip(".nc")

    output_da = xr.concat(data_arrays, dim="id").drop(
        ["valid_time", "heightAboveGround", "time", "step", "depthBelowGround"],
        errors="ignore",
    )

    if output_ncpath is not None:
        try:
            output_da.to_netcdf(output_ncpath)
        except Exception as e:
            print("ERROR : Failed to export as netcdf : {}".format(e))

        try:
            with open(
                output_ncpath.replace(".nc", ".json"), "w", encoding="utf-8"
            ) as f:
                json.dump(output_dict, f, default=str, indent=4, ensure_ascii=False)
        except Exception as e:
            print("ERROR : Failed to export as json : {}".format(e))

    return output_da, output_dict
