import shapely.geometry as geom
import shapefile
import xarray as xr
import json
import numpy as np
import os
from osgeo import ogr, osr
import logging
import datetime as dt
from PIL import Image, ImageDraw

# Logging
log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
TS = dt.datetime.utcnow().strftime("%Y%m%d_%H%M")
file_handler = logging.FileHandler("log_" + TS)
file_handler.setLevel(logging.WARNING)
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
ch.setFormatter(formatter)
file_handler.setFormatter(formatter)
log.addHandler(ch)
log.addHandler(file_handler)


def create_mask(poly1, ds):
    """
        Return an DataArray from a shape and a Dataset/Dataarray
        Add return_2d_ds to be sure that the dataset contains only lat/lon
        coordinates.
        This help getting results quickly.

        :param ploy1: The polygon to treat
        :param ds: The corresponding dataset

        Warning : Lat/Lon coordinates should be regular.
                  The function needs to be changed for irregular grid
    """
    # 1. Get the bounds of the polygone
    min_lon, min_lat, max_lon, max_lat = poly1.bounds
    dwork = ds.copy(deep=True)
    if "lat" in dwork.coords:
        dwork = dwork.rename({"lat": "latitude"})
    way_lat = int(np.sign(dwork.latitude[1].values - dwork.latitude[0].values))
    if "lon" in dwork.coords:
        dwork = dwork.rename({"lon": "longitude"})
    way_lon = int(np.sign(dwork.longitude[1].values - dwork.longitude[0].values))
    stepX = abs(dwork.latitude.values[0] - dwork.latitude.values[1]) / 2.0
    stepY = abs(dwork.longitude.values[0] - dwork.longitude.values[1]) / 2.0
    dtemp = (
        dwork.sel(latitude=slice(min_lat - 2 * stepX, max_lat + 2 * stepX, way_lat))
        .sel(longitude=slice(min_lon - 2 * stepY, max_lon + 2 * stepY, way_lon))
        .squeeze()
    )
    df = dtemp.to_dataframe().reset_index()
    values = []
    for coord in df.values:
        # A rereflechir pour transformer en mode multi scale.
        # Peut etre utiliser interp entre plusieurs grilles et tester ensuite
        # uniquement les points critiques non binaires ?
        # Integrer de meme une facon d'apprehender les grilles non regulieres
        # (se debarasser de stepX et stepY)
        # Enfin, penser a la facon d'agglomerer (par ex. pour les donnees de type
        # population)
        point = geom.MultiPolygon(
            [
                (
                    (
                        (coord[1] - stepY, coord[0] - stepX),
                        (coord[1] - stepY, coord[0] + stepX),
                        (coord[1] + stepY, coord[0] + stepX),
                        (coord[1] + stepY, coord[0] - stepX),
                        (coord[1] - stepY, coord[0] - stepX),
                    ),
                    (),
                )
            ]
        )
        values.append(point.intersection(poly1).area / point.area)
    df["mask"] = values
    return df.set_index(["latitude", "longitude"]).to_xarray()["mask"]


def lon_lat_to_img(point, g):
    """
    Transform to apply to an img to get coordinates
    :param point: lon/lat of point
    :param g: geometry grid as dict
    """

    lon, lat = point[0], point[1]
    # x / lon transform
    a_x = (g["nb_c"] - 1) / (g["last_lon"] - g["first_lon"])
    b_x = -g["first_lon"] * a_x
    x = a_x * lon + b_x
    # y  / Lat transform
    a_y = (g["nb_l"] - 1) / (g["last_lat"] - g["first_lat"])
    b_y = -g["first_lat"] * a_y
    y = a_y * lat + b_y
    return (x, y)


def lon_lat_shape_to_img(poly, g):
    return [
        lon_lat_to_img(point, g)
        for point in poly.exterior.__geo_interface__["coordinates"]
    ]


def create_mask_PIL(poly, g, ds_grid):
    """
    Create mask using PIL library
    """
    img = Image.new("1", (int(g["nb_c"]), int(g["nb_l"])))
    if poly.geometryType() == "Polygon":
        img_shape = lon_lat_shape_to_img(poly, g=g)
        ImageDraw.Draw(img).polygon(img_shape, fill=1, outline=1)
    elif poly.geometryType() == "MultiPolygon":
        for small in poly:
            img_shape = lon_lat_shape_to_img(small, g=g)
            ImageDraw.Draw(img).polygon(img_shape, fill=1, outline=1)
    ds = xr.Dataset()
    ds["latitude"] = ds_grid.latitude
    ds["longitude"] = ds_grid.longitude
    ds["mask"] = (("latitude", "longitude"), np.array(img))
    return (
        ds["mask"]
        .where(ds["mask"] > 0)
        .dropna(dim="latitude", how="all")
        .dropna(dim="longitude", how="all")
    )


class generate_grid_mask:
    """
    This class enable to generate a grid mask (saved as netcdf file)
    from a polygon mask.
    """

    def __init__(self, input_file, input_grid_file, output_directory, **kwargs):
        """
        Mandatory argument :
            input_file : File containing the masks
            input_grid_file : File containing the grid on which we want to
                project the mask. It can be "grib" or "netcdf" format.
            output_directory : Directory in which mask are saved
        Optional (specfiy through kwargs):
            - input_grid_format : Format of the grid file.
                Could either be "grib" (default) or "netcdf".

        """
        log.debug("Init grid mask ")
        self.input_file = input_file
        self.input_grid_file = input_grid_file
        self.output_directory = output_directory
        self.input_grid_format = kwargs.get("input_grid_format", "grib")
        self.saving_strategy = kwargs.get("saving_strategy", "one_per_file")
        self.test = kwargs.get("test_mode", True)  # Are we in test/dev mode or not
        self.force_creation = kwargs.get("force_creation", False)
        self.use_PIL = kwargs.get("use_PIL", False)
        # self.output_format = "nc" #Not yet used
        # self.input_format = "geojson" # Not yet used
        # self.id_getter = None # Not yet used

    def load_geojson(self):
        """
        On load l'ensemble du fichier en memoire
        Pas top mais pour l'instant ca ira.
        """
        with open(self.input_file) as geojson1:
            self.poly_geojson = json.load(geojson1)

    def load_shapefile(self):
        """
        Open the shapefile
        Save the list of available field (on top of polygon)
        """
        self.shp_file = shapefile.Reader(self.input_file)
        self.available_fields = list(np.asarray(self.shp_file.fields[1:])[:, 0])
        # We need to check that crs is regular lat/lon grid
        driver = ogr.GetDriverByName("ESRI Shapefile")
        data_source = driver.Open(self.input_file)
        layer = data_source.GetLayer()
        sourceprj = layer.GetSpatialRef()

        targetprj = osr.SpatialReference()
        targetprj.ImportFromEPSG(4326)
        if targetprj.IsSame(sourceprj) == 1:
            print("We are working with the expected CRS")
        else:
            raise ValueError(
                "Shapefile doesn't seem to be projected with the good CRS. "
                "Transformation between CRS is not yet included."
            )

    def read_grid(self):
        """
        Read the grib file.
        Choose the method according to the input type.
        Ensure that the grid is bidimensionnal (by squeezing other variables)
        """
        log.info(f"Reading grid from file  {self.input_grid_file}.")
        if self.input_grid_format == "grib":
            self.ds_grid = xr.open_dataset(self.input_grid_file, engine="cfgrib")
        elif self.input_grid_format == "netcdf":
            self.ds_grid = xr.open_dataset(self.input_grid_file)
        # On change les coordonnées si besoin.
        if "lat" in self.ds_grid.coords:
            self.ds_grid = self.ds_grid.rename({"lat": "latitude"})
        if "lon" in self.ds_grid.coords:
            self.ds_grid = self.ds_grid.rename({"lon": "longitude"})
        # Delate other dimension than latitude and longitude
        for x in self.ds_grid.dims:
            if x not in ["latitude", "longitude"]:
                self.ds_grid = self.ds_grid.isel({x: 0}).squeeze()

        self.g = {}
        self.g["first_lat"] = self.ds_grid.latitude[0].values.round(4)
        self.g["last_lat"] = self.ds_grid.latitude[-1].values.round(4)
        self.g["step_lat"] = (self.g["last_lat"] - self.g["first_lat"]) / (
            self.ds_grid.latitude.size - 1
        )
        self.g["nb_l"] = self.ds_grid.latitude.size
        self.g["first_lon"] = self.ds_grid.longitude[0].values.round(4)
        self.g["last_lon"] = self.ds_grid.longitude[-1].values.round(4)
        self.g["step_lon"] = (self.g["last_lon"] - self.g["first_lon"]) / (
            self.ds_grid.longitude.size - 1
        )
        self.g["nb_c"] = self.ds_grid.longitude.size
        print(self.g)

    def get_output_name(self, name):
        return self.output_directory + name + ".nc"

    def save_mask(self, mask, name):
        """
        Save the mask in the specify directory

        Input :
           - mask : a dataset or dataarray with a "mask" variable
           - name : output filename for the mask

        Information :
            Use self.saving_strategy to know how to save the masks.
        """
        if self.saving_strategy == "one_per_file":
            fname = self.get_output_name(name)
            log.debug(f"Saving mask at {fname}")
            maxi = mask.max().values
            mini = mask.min().values
            if maxi > 1:
                raise ValueError(f"Problem found in mask : maximum is {maxi} >1")
            if mini < 0:
                raise ValueError(f"Problem found in mask : maximum is {mini} <0")

            mask.to_netcdf(fname, mode="w", encoding={"mask": {"dtype": "float32"}})
        else:
            raise ValueError(
                f"Saving strategy not yet implemented : {self.saving_strategy}"
            )

    def loop_for_shapefile(self, id_key, **kwargs):
        """
            Loop over all the polygon include in the shapefile to create a mask
        """
        self.load_shapefile()
        if id_key not in self.available_fields:
            raise ValueError(
                f"The ID key {id_key} is not in the shapefile available keys.\n"
                f"Please choose one among avaialbe keys {self.available_fields}."
            )
        name_index = self.available_fields.index(id_key)
        self.read_grid()
        log.info("Grid has been read")
        # On recupere le nombre d'enregistrement
        num_records = self.shp_file.numRecords
        if self.test:
            num_records = 3
        # On recupere les indices des metadonnees a conserver
        if "prop_keys" in kwargs:
            other_index = []
            prop_keys = kwargs.get("prop_keys")
            if not isinstance(prop_keys, list):
                prop_keys = [prop_keys]
            for key in prop_keys:
                other_index.append(self.available_fields.index(key))
        else:
            log.warning(
                "Getting only geometry information from the shapefile. "
                "You need to specify 'prop_keys' list in the configuration "
                f"if you don't want that. Available fields are {self.available_fields}"
            )

        # On boucle pour creer les masques
        for id_area in np.arange(num_records):
            other_data = self.shp_file.record(id_area)
            name = other_data[name_index]
            fname = self.get_output_name(name)
            log.debug(f"Doing {name}")
            if not os.path.exists(fname) or self.force_creation:
                poly = geom.asShape(self.shp_file.shapeRecord(id_area).shape)
                if self.use_PIL:
                    mask_region = create_mask_PIL(poly, g=self.g, ds_grid=self.ds_grid)
                else:
                    mask_region = create_mask(poly, self.ds_grid)
                if "prop_keys" in kwargs:
                    dict_prop = {}
                    for i, key in enumerate(prop_keys):
                        dict_prop[key] = other_data[other_index[i]]
                    self.add_information_to_mask(mask_region, dict_prop)
                mask_region = (
                    mask_region.expand_dims("id")
                    .assign_coords(id=[name])
                    .where(mask_region > 0)
                )
                if "glob_attr" in kwargs:
                    mask_region = self.add_glob_attr(
                        mask_region, kwargs.get("glob_attr")
                    )

                if len(mask_region.latitude) == 0 or len(mask_region.longitude) == 0:
                    log.warning(f"Mask is not present in the original grid {name}")
                else:
                    self.save_mask(mask_region, name)
            else:
                log.info("File already present : nothing has been done.")

    @staticmethod
    def add_glob_attr(mask, dict_attr):
        for key in dict_attr:
            mask.attrs[key] = dict_attr[key]
        return mask

    @staticmethod
    def add_information_to_mask(mask, dict_prop):
        """
        Enable to add information present in the SIG file into the netcdf file.
        """
        if int(xr.__version__.split(".")[1]) < 13:
            log.info("Properties are stored as attribute and not as coordinates")
        for key in dict_prop:
            if int(xr.__version__.split(".")[1]) >= 13:
                dict_temp = {}
                dict_temp[key] = [dict_prop.pop(key)]
                mask = mask.expand_dims(key).assign_coords(coords=dict_temp)
            else:
                mask.attrs[key] = dict_prop[key]

    def loop_for_geojson(self, id_key, **kwargs):
        """Fait une boucle pour créer un masque pour chaque fichier geosjson
        qui existe.

        :param id_key: Key to look as ID in geojson file.
            Parameters in kwargs enable to say at which level it is.
        kwargs :
            enable to handle the fact that Id keys is not known
            and can be at different level
        """
        self.load_geojson()
        self.read_grid()
        lev_id = kwargs.get("lev_id", 1)

        num_records = len(self.poly_geojson["features"])
        if self.test:
            num_records = 1
        for id_area in np.arange(num_records):
            # Getting the polygon definining the mask
            poly = geom.asShape(self.poly_geojson["features"][id_area]["geometry"])
            if lev_id == 1:
                name = str(self.poly_geojson["features"][id_area][id_key])
                if "prop_keys" in kwargs:
                    prop = self.poly_geojson["features"][id_area]
            elif lev_id == 2:
                if "lev1_id" not in kwargs:
                    raise ValueError(
                        "You specify that the ID should be two level "
                        "from features but do not specify lev1_id."
                    )
                name = str(
                    self.poly_geojson["features"][id_area][kwargs.get("lev1_id")][
                        id_key
                    ]
                )
                if "prop_keys" in kwargs:
                    prop = self.poly_geojson["features"][id_area][kwargs.get("lev1_id")]
            elif lev_id == 3:
                if "lev1_id" not in kwargs:
                    raise ValueError(
                        "You specify that the ID should be 3 level "
                        "from features but do not specify lev1_id."
                    )
                if "lev2_id" not in kwargs:
                    raise ValueError(
                        "You specify that the ID should be 3 level "
                        "from features but do not specify lev2_id."
                    )
                name = str(
                    self.poly_geojson["features"][id_area][kwargs.get("lev1_id")][
                        kwargs.get("lev2_id")
                    ][id_key]
                )
                if "prop_keys" in kwargs:
                    prop = self.poly_geojson["features"][id_area][
                        kwargs.get("lev1_id")
                    ][kwargs.get("lev2_id")]
            else:
                raise ValueError(
                    "lev_id > 3 has not been developped. Please develop it.\n"
                    f"lev_id is {lev_id}"
                )
            # To be move elsewhere ?
            log.debug(f"Starting mask creation for {name} ")
            fname = self.get_output_name(name)
            if not os.path.exists(fname) or self.force_creation:
                if self.use_PIL:
                    mask_region = create_mask_PIL(poly, g=self.g, ds_grid=self.ds_grid)
                else:
                    mask_region = create_mask(poly, self.ds_grid)
                if "prop_keys" in kwargs:
                    # removing useless key
                    unwant = set(prop) - set(kwargs.get("prop_keys"))
                    for unwanted_key in unwant:
                        del prop[unwanted_key]
                    # Adding informaiton to the mask
                    self.add_information_to_mask(mask_region, prop)
                if len(mask_region.latitude) == 0 or len(mask_region.longitude) == 0:
                    log.warning(f"Mask is not present in the original grid {name}")
                else:
                    mask_region = (
                        mask_region.expand_dims("id")
                        .assign_coords(id=[name])
                        .where(mask_region > 0)
                    )
                    self.save_mask(mask_region, name)
            else:
                log.debug(
                    f"Masked already exist {fname} and will not be recreated."
                    "Set force_creation to true in namelist to force the mask creation"
                )
