"""
@package lib.output

Output formatting module
Aims at formatting results from the parameter generator
and the comment_generator (xarray.Dataset) into a JSON file
"""

# Standard packages
import os
import json
from datetime import datetime
import tarfile
from functools import partial
from multiprocessing import Pool

import mflog


# Third parties packages


# Own package
from ftap_autom.promethee.src.utils.hash import MD5
from ftap_autom.promethee.src.utils.exception import handle_error
from ftap_autom.promethee.src.conf.app_configuration import (
    MFLOG,
    WORKING_DIR,
    DATETIME_FMT_CDP,
    DATETIME_FMT,
    JSON_KWARGS,
)
from ftap_autom.promethee.src.lib.comment_localisation.comment_manager import (
    CommentManager,
)


# Logging
LOGGER = mflog.getLogger("output.mod")
mflog.set_config(**MFLOG)
LOGGER = LOGGER.bind(module="output")


class OutputFormatter:
    """ OutputFormatter
    This class aims at formatting a "compacted" component into a dictionnary
    structured as required by the "CDP Bulletin".
    By "compacted" we mean a component description that is applied to several
    masks (this kind of component only exists for Risk components). It means that
    the component_config has several 'uid' and several corresponding 'mask_id'.
    """

    def __init__(self, component_config, component):
        """
        __init__ : OutputFormatter initilization

        Args:
            component_config (dict): Component configuration as it could be
            found in a "production" configuration after the original config
            processing.
            component (dict): Dictionnary with the following structure
            depending on the component type :
                * Text : {
                    'comments' : {component_id : text}
                }
                * Risk : {
                    'risks' : risks_dataset,
                    'comments' : {component_id : text}
                }
        """
        self.component_config = component_config
        self.component = component

    @property
    def component_type(self):
        """component_type : Component type property

        Returns:
            str : Type of the component
        """
        return self.component_config["type"]

    @property
    def formatted_component(self):
        """formatted_component : Formatted component property

        Returns:
            str : List of single (i.e. not compacted components) formatted
            in dictionnaries as required for the "CDP Bulletin".
        """
        global LOGGER
        LOGGER = LOGGER.bind(
            prod_id=self.component_config.get("prod_id"),
            compo_id=self.component_config.get("id"),
            period_id=self.component_config.get("period", {"id": None}).get("id"),
        )
        results = []
        component_base = {
            "prod_id": self.component_config["prod_id"],
            # A modifier pour utiliser "PeriodId" quand ce dernier sera present
            # dans le fichier de config.
            "Period": {
                "PeriodId": self.component_config["period"]["id"],
                "DateDebutPeriode": datetime.strptime(
                    self.component_config["period"]["start"], DATETIME_FMT
                ).strftime(DATETIME_FMT_CDP),
                "DateFinPeriode": datetime.strptime(
                    self.component_config["period"]["stop"], DATETIME_FMT
                ).strftime(DATETIME_FMT_CDP),
            },
            "ComponentName": self.component_config["name"],
            "CustomerId": self.component_config["customer"],
            "CustomerName": self.component_config.get(  # to Snake Case
                "customer_name", "AutomaticFilling"
            ),
            "ProductionName": self.component_config.get(  # to Snake Case
                "production_name", "AutomaticFilling"
            ),
            "ProductionId": self.component_config.get(  # to Snake Case
                "production_id", "AutomaticFilling"
            ),
        }
        component_base["Period"]["PeriodName"] = (
            "Du "
            + component_base["Period"]["DateDebutPeriode"]
            + " au "
            + component_base["Period"]["DateFinPeriode"]
        )
        if self.component_type in ["Text", "text"]:
            # Case if component_type is Text
            # This part is just a first draft
            component_id = self.component_config["id"]

            component_dico = {
                **component_base,
                "ComponentType": "Text",
                "ComponentId": component_id,
                "SyntText": self.component["comments"][component_id],
            }
            component_dico["GeoId"] = self.component_config["geos"]["mask_id"]
            results += [component_dico]

        elif self.component_type in ["Risk", "risk"]:
            # Case if component_type is Risk
            # Again, this part is a first draft, a lot of features are missing
            # or mis-implemented
            comment_handler = CommentManager(self.component, self.component_config)

            for geo_id in self.component_config["geos"]["mask_id"]:
                # Binding geo_id to the LOGGER
                LOGGER = LOGGER.bind(geo_id=geo_id)
                LOGGER.debug(
                    "Processing risk",
                    type_id=repr(type(self.component_config.get("id"))),
                    func="formatted_component",
                )
                component_id = self.component_config["id"]
                if not isinstance(self.component_config["id"], str):
                    LOGGER.error(
                        "Component_id should be a string",
                        type_id=repr(type(self.component_config.get("id"))),
                        func="formatted_component",
                    )
                # On prevoit le cas où la génération du commentaire échoue
                # Dans ce cas, on log l'erreur et on le spécifie dans le commentaire.
                # Peut être renvoyer un commentaire vide "" ?
                try:
                    LOGGER.debug(
                        "On lance le commentaire", func="formatted_component",
                    )
                    detail_comment = comment_handler.get_comment(geo_id)
                    if detail_comment != "R.A.S.":
                        LOGGER.info(
                            "Commentaire calculé",
                            comment=detail_comment,
                            func="formatted_component",
                        )
                except Exception as excpt:
                    handle_error(excpt, log=LOGGER)
                    detail_comment = (
                        "Erreur dans output lors de la generation de ce commentaire. "
                        "Voir les logs pour plus de details."
                    )
                    LOGGER.debug(
                        "On va continuer la creation du fichier de sortie malgré "
                        "l'erreur dans la génération du commentaire détaillé.",
                        func="formatted_component",
                    )

                period_update = self.component.get_risk_period()
                component_base["Period"].update(period_update)
                geoName = self.component.get_geo_name(geo_id)
                component_dico = {
                    **component_base,
                    "ComponentType": "Aleas",
                    "ComponentId": component_id,
                    "HazardId": self.component_config["hazard"],
                    "HazardName": self.component.get_hazard_name(),
                    "GeoId": geo_id,
                    "GeoName": geoName,
                    "Dataset": {
                        "ShortSummary": self.component.get_short_summary(geo_id),
                        "Summary": self.component.get_summary(geo_id),
                    },
                    "DetailComment": detail_comment,
                }
                LOGGER.debug("Dico généré.", func="formatted_component")
                results += [component_dico]
        else:
            LOGGER.error(
                f"Unknown component type : {self.component_type}",
                func="formatted_component",
            )
        LOGGER.debug("Going out of OutputFormatter", func="formatted_component")
        LOGGER = LOGGER.try_unbind("prod_id", "compo_id", "period_id", "geo_id")
        return results


class OutputExporter:
    """ OutputExporter
    This class aims at exporting to JSON files productions which components
    have previously been formatted to a dictionnary structured as required by
    the "CDP Bulletin".
    """

    def __init__(self, formatted_components, working_dir="./"):
        """__init__ : OutputExporter initialization

        Args:
            formatted_components (list): list of all dictionnaries. Every
            dictionnary being a single formatted component.
            working_dir (str, optional): Path to the working directory.
            Defaults to './'.
        """
        # On va recuperer la date de reference :
        conf_file = os.path.join(working_dir, WORKING_DIR["configs"]["version"])
        with open(conf_file) as fp:
            version = json.load(fp)
        self.date_bulletin = datetime.strptime(
            version["reference_datetime"], DATETIME_FMT
        ).strftime(DATETIME_FMT_CDP)
        # Fixer encore de facon brutale. Peut être la supprimer ?
        self.date_config = datetime.strptime(
            version.get("configuration_datetime", "19710110T000000"), DATETIME_FMT
        ).strftime(DATETIME_FMT_CDP)
        self.date_production = datetime.today().strftime(DATETIME_FMT_CDP)
        self.grouped_components = formatted_components
        self.output_dir = os.path.join(working_dir, WORKING_DIR["output"])
        os.makedirs(self.output_dir, exist_ok=True)

    @property
    def grouped_components(self):
        """grouped_components : In the list of all the formatted components,
        it finds and groups all the components that belongs to the same
        production into the same dictionnary

        Returns:
            dict : Dictionnary of all the productions identified by their
            production id {uid : production}
        """
        return self._grouped_components

    @grouped_components.setter
    def grouped_components(self, formatted_components):
        """grouped_components : In the list of all the formatted components,
        it finds and groups all the components that belongs to the same
        production into the same dictionnary

        Returns:
            dict : Dictionnary of all the productions identified by their
            production id {uid : production}

        History :  Some change are made.
           A customer may have different production. Therefore we introduce "prod_id".

        """
        grouped = dict()
        for compo in formatted_components:
            component = compo.copy()
            ProdId = component.pop("prod_id")  # to Snake Case
            CustomerId = component.pop("CustomerId")  # to Snake Case
            CustomerName = component.pop("CustomerName")  # to Snake Case
            ComponentType = component.pop("ComponentType")  # to Snake Case
            ProductionId = component.pop("ProductionId")  # to Snake Case
            ProductionName = component.pop("ProductionName")  # to Snake Case
            if ProdId in grouped:
                if ComponentType in grouped[ProdId]["Components"]:
                    grouped[ProdId]["Components"][ComponentType] += [component]
                else:
                    grouped[ProdId]["Components"].update({ComponentType: [component]})
            else:
                grouped[ProdId] = {
                    "CustomerId": CustomerId,
                    "CustomerName": CustomerName,
                    "ProductionId": ProductionId,
                    "ProductionName": ProductionName,
                    "DateBulletin": self.date_bulletin,
                    "DateProduction": self.date_production,
                    "DateConfiguration": self.date_config,
                    "Components": {ComponentType: [component]},
                }
        self._grouped_components = grouped

    #        return grouped

    def export_production(self, prod_id):
        """export_production : Export a single production into a JSON file

        Args:
            prod_id (dict): identifier of a single production
        Returns :
            str : The filename where the results are stored.
        """
        if prod_id not in self.grouped_components:
            LOGGER.error(
                f"prod_id '{prod_id}' not in self.grouped_components",
                prod_id=prod_id,
                func="export_production",
            )
            return False

        prod_hash = MD5(self.grouped_components[prod_id]).hash
        filename = os.path.join(self.output_dir, f"prom_{prod_id}_{prod_hash}.json")
        with open(filename, "w") as out_f:
            json.dump(self.grouped_components[prod_id], out_f, **JSON_KWARGS)
        return filename

    def export_all(self, nproc):
        """export_all : Exports all production into JSON files (a single file
        per production). Export done using parallel processing.
        In the end of the process, create an tar_gz archive.

        Args:
            nproc (int): number max of processor to use for parallel processing
        """
        list_files = []
        if nproc > 1:
            my_error_handler = partial(handle_error, log=LOGGER)
            pool = Pool(nproc)
            for elt in self.grouped_components:
                pool.apply_async(
                    self.export_production,
                    args=(elt,),
                    callback=list_files.append,
                    error_callback=my_error_handler,
                )
            pool.close()
            pool.join()
        else:
            for elt in self.grouped_components:
                fname = self.export_production(elt)
                list_files.append(fname)

        # On archive les fichiers creer
        archive_name = os.path.join(self.output_dir, "promethee_msb_poc.tgz")
        with tarfile.open(archive_name, "w:gz") as tar:
            for filename in list_files:
                full_path = os.path.abspath(filename)
                base_name = os.path.basename(filename)
                # On ajoute le fichier en utilisant son path global
                # Par contre on lui donne le nom local
                tar.add(full_path, arcname=base_name)
                # Doit-on supprimer le fichier après l'avoir ajouter à l'archive ???
