"""parameter_generator.py.

Parameter Generator module
"""
import pandas as pd
import xarray as xr
import numpy as np

import mflog
from ftap_autom.promethee.src.lib.risk_generator import RiskGenericEvent
from ftap_autom.promethee.src.conf.app_configuration import DATETIME_FMT_CDP, MFLOG

# Logging
LOGGER = mflog.getLogger("alea")
mflog.set_config(**MFLOG)
LOGGER = LOGGER.bind(module="Alea")


# Patch pour convertir les données de lamyaa
def convert_risk_level_to_potential_occurrence(da):
    """
    patch utilisé normalement que dans les tests (à voir si toujours nécessaire)
    """
    levels = set(np.unique(da.values)).difference({0})
    ds = xr.merge(
        [
            (da == level).expand_dims("risk_level").assign_coords(risk_level=[level])
            for level in levels
        ]
    )
    return ds.rename({da.name: "occurrence"})["occurrence"]


# Fin du patch. A supprimer une fois que les donnees utilisées par lamyaa
# auront été converties...

# Variable qui ne peuvent pas se merger facilement (car le temps n'est
# plus une dimension)
SPECIAL_VAR = {"summarized_density": "max", "risk_summarized_density": "max"}


def special_merge(dfirst, dsecond):
    """
    Va permettre de merger des variables "non mergeables".
    Les variables non mergeables doivent être déclarées dans SPECIAL_VAR.
    Pour ces variables, on prendra soit le max, soit le min, soit la moyenne
    (selon le cas).
    Pour l'instant seul le max est implémenté. Implémentation des autres cas à prévoir.
    """

    dvars = set(dsecond.data_vars).intersection(set(SPECIAL_VAR.keys()))
    dvars1 = set(dfirst.data_vars).intersection(set(SPECIAL_VAR.keys()))
    inter = dvars.intersection(dvars1)
    dout = xr.Dataset()

    for var in inter:
        var1 = dfirst[var]
        lev1 = set(var1.risk_level.values)
        var2 = dsecond[var]
        lev2 = set(var2.risk_level.values)
        lev_inter = lev2.intersection(lev1)
        if lev_inter != set():
            if SPECIAL_VAR[var] == "max":
                dout[var] = xr.ufuncs.fmax(var1, var2.broadcast_like(var1))
                dfirst = dfirst.drop_vars(var)
                dsecond = dsecond.drop_vars(var)
            else:
                raise ValueError("Method not implemented")
    return xr.merge([dfirst, dsecond, dout])


class ParameterGeneratorTransit:
    """Classe permettant de gérer un paramètre. Calcul les différents de risque
    associé. Permet de retourner les informations soit :

      - sous forme de dictionnaire (pour les sorties vers le CDP,
      - sous forme d'array (pour permettre la liaison avec
             le module de génération de texte détaillé).

    Modif :
       02/09/2020 (V. Chabot) :
           Ajout des fonctions permettant d'exporter des dictionnaires pour les sorties.
    """

    _unique_warning = True

    def __init__(self, alea_conf):
        """__init__

        Arguments:

            alea_conf {object} --
            L'objet decrivant le paramètre à calculer
            (créer via interpretor.ParamInterp)
        """
        self.alea_conf = alea_conf
        self.time_dimension = alea_conf.generic_get("time_dimension")
        if self.time_dimension is None:
            LOGGER.info(
                "Time dimension has been set to default. "
                "Please check if this is correct.",
                func="__init__",
            )
            self.time_dimension = "step"
            # On rajoute une valeur par défault
            alea_conf.setdefault({"time_dimension": self.time_dimension})
        # On definit le dataset de sortie.
        self.risks = xr.Dataset()

    @property
    def unique_warning(self):
        return type(self)._unique_warning

    @unique_warning.setter
    def unique_warning(self, val):
        type(self)._unique_warning = val

    def is_risks_empty(self):
        return not bool(self.risks)

    def get_risk(self, level):
        """get_risk : Create a risk handler and computes the corresponding
        risk DataArray

        Arguments:
            level (RiskInterp)  -- An configuration interpretor for the given level

        Returns:
            xarray.DataSet (time_dimension, id, evt) -- Get the risk dataset.

        """
        aggregate_dim = ["latitude", "longitude"]
        risk_handler = RiskGenericEvent(risk_conf=level, aggregate_dim=aggregate_dim)
        output_risk = risk_handler.make_output_dataset()
        output_risk.attrs["level"] = int(level.generic_get("level"))
        for dim in aggregate_dim:
            if hasattr(output_risk, dim):
                output_risk = output_risk.reset_coords(dim, drop=True)
        return output_risk

    def load_risk(self, ds):
        """load_risk
        This function enable to initialise an Hazard with already computed hazard

        Args:
            ds (xr.Dataset): The hazard
        """
        self.risks = ds
        self.compute_final_risk_level()

    def get_hazard(self):
        """
        This function enable the acess to the risks (once it is loaded or calculated)
        """
        return self.risks

    def compute_risks(self):
        """compute_risks : Computes all the risks configured in the
        self.dict_risk and create the self.risks attribute

        Chaque niveau de risque est calculé en faisant appel à get_risk.
        Il est possible de découper un même niveau en plusieurs blocs.
        Par exemple car on a des données sur des grilles différentes à différentes
        échéances.
        L'array du risque est étendue (pour ajouter une coordonnée risk_level).

        De plus, des informations sur les zones sont rajoutées

        Evolution :
          - avec le nouveau MaskLoader, AreaName et AreaType sont des coordonnées.
        """
        hazard = self.alea_conf.generic_get("hazard")
        prod_id = self.alea_conf.generic_get("prod_id")
        if hazard is not None or prod_id is not None:
            LOGGER.debug(
                "Computing hazard",
                hazard=hazard,
                bulletin=prod_id,
                func="compute_risks",
            )

        risks_ds = xr.Dataset()
        for level in self.alea_conf.risk_level:
            risk = self.get_risk(level)
            risk = risk.expand_dims(dim="risk_level").assign_coords(
                risk_level=[int(level.generic_get("level"))]
            )
            try:
                # Faire quelque chose de specifique pour la densite resumee...
                risks_ds = special_merge(risks_ds, risk)
            except Exception as excpt:
                LOGGER.exception(
                    "Error in merging dataset",
                    hazard=hazard,
                    bulletin=prod_id,
                    func="compute_risks",
                    exc_info=True,
                )
                raise ValueError from excpt

        if "areaType" in risks_ds.data_vars:
            risks_ds["areaType"] = risks_ds["areaType"].isel(risk_level=0)

        if "areaName" in risks_ds.data_vars:
            risks_ds["areaName"] = risks_ds["areaName"].isel(risk_level=0)
        self.risks = risks_ds
        self.compute_final_risk_level()

    @staticmethod
    def replace_middle(x, axis, **kwargs):
        """Cette fonction remplace la valeur du milieu d'un risque si elle est
        inférieur à ses voisins. On "scan" et on remplace. Ainsi [2,1,2] =>

        [2,2,2]

        [5,1,4] => [5,4,4]
        [5,4,1] = [5,4,1]

        C'est cette fonction qui bouche les trous.
        Par contre, comme vu avec les specificateurs,
        c'est pas grave que les autres valeurs ne soient pas cohérentes ...

        Args:
            x ([ndarray]):
                L'array comprenant les risques à boucher.
                Cette array doit être passée par un rolling (sur 3 time_dimension).
                Ainsi on a une array qui possède une dimension de plus
                (par rapport à l'originale)
            axis ([axis]):
                       Passé directement par reduce (on dirait).
                        Concerne l'axe sur lequel on a "roller".
            kwargs : Obligatoire pour être compatible avec reduce de xarray
        Returns:
            [ndarray]: Une array de la dimension originale (avant rolling).
        """
        if isinstance(axis, tuple) and len(axis) == 1:
            axis = axis[0]
        x_borders = np.min(x.take([0, 2], axis=axis), axis=axis)
        x_middle = x.take(1, axis=axis)
        x_out = np.nanmax([x_borders, x_middle], axis=0)
        return x_out

    # Accesseurs pour le risque final.
    def get_period_for_level(self, my_id, level, coord=None):
        """get_period_for_level Return the period for which a given level is activated

        The period is either the time_dimension (in this module)
        or any other coordinate (indexed with the time_dimension)

        Args:
            my_id ([str]): An information for the area selection
            level ([int]): An integer telling which is the level of interest
            coord ([type], optional): The coordinate of interest.
                Defaults to None (so self.time_dimension)

        Returns:
            [xr.DataArray]: The coordinate selected for the given period
        """

        if coord is None:
            coord = self.time_dimension
        if self.is_risks_empty():
            return xr.DataArray({coord: [], "id": None})
        risk_area = self.final_risk.sel(id=my_id)
        if coord not in self.final_risk.coords:
            raise AttributeError(
                f"Dimension {coord} is not part of coordinates. "
                f"Coordinates list is {self.final_risk.coords}."
            )
        period = risk_area[risk_area == level][coord]
        return period

    def get_final_max_risk_period(self, my_id, coord=None):
        """
        Return, for a given id, the maximum of the risk as well as the step where
        it occured
        """
        if coord is None:
            coord = self.time_dimension
        risk_max = self.get_final_risk_max_level(my_id)
        period = self.get_period_for_level(my_id, level=risk_max, coord=coord)
        return period

    def get_period_for_risk(self, my_id, coord=None):
        """get_period_with_risk

        Return the period when a risk is raised (i-e risk is not 0)

        Args:
            my_id (str): Id of the area
            coord (str) : [Optional]. Can be any coordinates.
                If None, set to self.time_dimension
        Return:
            [np.ndarray] : The period of risk. The type is the one of
                self.time_dimension
        """
        if coord is None:
            coord = self.time_dimension
        risk_area = self.final_risk.sel(id=my_id)
        return risk_area[risk_area != 0][coord]

    def get_final_risk_max_level(self, my_id):
        """get_final_risk_max_level

        Return maximum level for a given area

        Args:
            my_id ([str]): The area id

        Returns:
            [int]: The maximum
        """
        if self.is_risks_empty():
            return 0
        return max(self.final_risk.sel(id=my_id).values)

    def get_final_risk_min_level(self, my_id):
        """get_final_risk_max_level

        Return minimum level for a given area
        Args:
            my_id ([str]): The area id
        Returns:
            [int]: The minimum
        """
        if self.is_risks_empty():
            return 0
        return min(self.final_risk.sel(id=my_id).values)

    def get_final_risk_level(self):
        """get_final_risk_level
        Return a dataArray describing the final risk level.

        Returns:
            [xr.DataArray]: [The final risk level]
        """
        return self.final_risk

    # Fin des accesseurs pour le risque final.

    def compute_final_risk_level(self):
        """compute_risk_level : Computes the maximum risk levels out of the
        previously computed self.risks Fait en sorte de lisser un peu le niveau
        de risque (en remplaçant les descente remonté trop brutales). Utilise
        la fonction replace_middle.

        Returns:
            xarray.DataArray (time_dimension, id) -- Maximum risk level DataArray
        """
        self.final_risk = xr.DataArray({self.time_dimension: [], "id": None})
        if self.is_risks_empty():
            LOGGER.warning(
                "Computed risks are empty. "
                "self.final_risk kept as an empty xr.DataArray.",
                func="compute_final_risk_level",
            )
        else:
            risk_only = self.risks["occurrence"] * self.risks.risk_level
            hourly_risk = risk_only.max(dim="risk_level", skipna=True)
            self.final_risk = hourly_risk.rolling(
                {self.time_dimension: 3}, center=True, min_periods=1
            ).reduce(self.replace_middle)

    def get_risk_period(self):
        """get_risk_period : Permet d'acceder à la période couverte par le risque

        Returns:
            [dictionnaire]: Le dictionnaire comprend la date de début
                et de fin de période;
        """
        if self.is_risks_empty():
            period = dict()
        elif self.time_dimension == "step":
            period = {
                "DateDebutPeriode": pd.to_datetime(
                    (self.risks.time + self.risks.step).min().values
                ).strftime(DATETIME_FMT_CDP),
                "DateFinPeriode": pd.to_datetime(
                    (self.risks.time + self.risks.step).max().values
                ).strftime(DATETIME_FMT_CDP),
            }
        else:
            period = {
                "DateDebutPeriode": pd.to_datetime(
                    self.risks[self.time_dimension].min().values
                ).strftime(DATETIME_FMT_CDP),
                "DateFinPeriode": pd.to_datetime(
                    self.risks[self.time_dimension].max().values
                ).strftime(DATETIME_FMT_CDP),
            }
        period["PeriodName"] = (
            "Du " + period["DateDebutPeriode"] + " au " + period["DateFinPeriode"]
        )
        return period

    def get_extended(self, geo_id):
        """get_extended : Permet de retourner le dataset etendu
        (pour le JSON de sortie)

        Ce dataset comprend toutes les informations pour tous les niveaux.
        Il n'est pas forcément cohérent avec le risque maximum en chaque heure
        (du au lissage des descentes/remontées).

        Args:
            geoId (str):  Le geoId dont on veut connaitre les informations
        Returns:
            [dictionnaire]: Dictionnaire contenant le résultat.
        """
        if self.is_risks_empty():
            return []

        dtemp = self.risks.sel(id=geo_id)
        if "summarized_density" in dtemp:
            dtemp = dtemp.drop("summarized_density")
        if "risk_summarized_density" in dtemp:
            dtemp = dtemp.drop("risk_summarized_density")

        data_list = []
        stepsize = self.get_stepsize(dtemp)
        for step in dtemp[self.time_dimension].values:
            step_dict = {}
            # Rajouter des options dans la fonction pour faire cela ?
            if self.time_dimension == "step":
                validity_date = dtemp.time + step
                step_dict["ValidityDate"] = pd.to_datetime(
                    validity_date.values
                ).strftime(DATETIME_FMT_CDP)
            else:
                step_dict["ValidityDate"] = pd.to_datetime(step).strftime(
                    DATETIME_FMT_CDP
                )
            risk_list = []
            for level in dtemp.risk_level.values:
                risk = {}
                risk["Level"] = int(level)
                risk["Values"] = self.get_values(
                    dtemp.sel({self.time_dimension: step, "risk_level": level}),
                    stepsize.stepsize.sel({self.time_dimension: step}),
                )
                risk_list.append(risk)
            step_dict["RiskLevels"] = risk_list
            data_list.append(step_dict)
        return data_list

    def get_geo_name(self, geo_id):
        LOGGER.debug(
            "Giving name of area.", geo_id=geo_id, func="get_geo_name",
        )
        res = "N.A"
        if not self.is_risks_empty():
            res = str(self.risks.sel(id=geo_id)["areaName"].values)
        return res

    def get_hazard_name(self):
        return self.alea_conf.generic_get("hazardName")

    def get_summary(self, geo_id):
        """get_summary Permet d'avoir un premier résumé. Pour chaque step, on
        donne l'information sur le niveua le plus élevé.

        Args:
            geo_id ([type]): Id de la zone d'intérêt.

        Returns:
            [dictionnaire]: Dictionnaire contenant le résultat.

        History :
           23/02/2021 :
                Adding values of first level even if activated level is 0
                (ie no risk is activated).
        """
        if self.is_risks_empty():
            return []

        dtemp = self.risks.sel(id=geo_id)
        if "summarized_density" in dtemp:
            dtemp = dtemp.drop("summarized_density")
        if "risk_summarized_density" in dtemp:
            dtemp = dtemp.drop("risk_summarized_density")

        stepsize = self.get_stepsize(dtemp)
        dfinal = self.final_risk.sel(id=geo_id)
        data_list = []
        for step in dtemp[self.time_dimension].values:
            step_dict = {}

            if self.time_dimension == "step":
                validity_date = dtemp.time + step
                step_dict["ValidityDate"] = pd.to_datetime(
                    validity_date.values
                ).strftime(DATETIME_FMT_CDP)
            else:
                step_dict["ValidityDate"] = pd.to_datetime(step).strftime(
                    DATETIME_FMT_CDP
                )
            level = int(dfinal.sel({self.time_dimension: step}).values)
            step_dict["Level"] = level

            if level > 0:
                summary_level = level
            elif level == 0:
                summary_level = dtemp.risk_level.min().values
            else:
                raise ValueError(f"Level should be >=0. Got {level}.")

            step_dict["Values"] = self.get_values(
                dtemp.sel({self.time_dimension: step, "risk_level": summary_level}),
                stepsize=stepsize.stepsize.sel({self.time_dimension: step}),
            )

            data_list.append(step_dict)
        return data_list

    def get_critical_value(self, da: xr.DataArray, operator: str) -> xr.DataArray:
        """
        Get the most critical values over time

        Args:
            da (xr.DataArray): The dataArray to look at
            operator (str): The comparison operator

        Raises:
            ValueError: If the comparison operator is not among
            [>,<,>=,<=,"inf","sup","supegal","infegal"]

        Returns:
            xr.DataArray: The critical value.
        """
        res = None
        if operator in [">", "sup", ">=", "supegal"]:
            res = da.max(self.time_dimension)
        elif operator in ["<", "inf", "<=", "infegal"]:
            res = da.min(self.time_dimension)
        else:
            raise ValueError(
                "Operator is not understood when trying to find the critical "
                f"representative_values {operator}"
            )
        return res

    def get_short_summary(self, geo_id):
        """get_short_summary
        Donne un résumé minimum (pour la V1) du risque.

        Args:
            geo_id (str): Id de la zone d'intérêt.

        Returns:
            [dictionnaire]: Dictionnaire contenant le résultat.
        History :
           23/02/2021 : Adding values of first level even if activated level is 0.
           June 2021 :  On remplace risk_density par density

        """
        if self.is_risks_empty():
            return dict()

        try:
            dtemp = self.risks.sel(id=geo_id)
        except Exception as excpt:
            LOGGER.exception(
                "Exception in selection",
                present_id=self.risks.id,
                required_id=geo_id,
                func="get_short_summary",
                exc_info=True,
            )
            raise excpt
        dfinal = self.final_risk.sel(id=geo_id)
        dict_out = {}
        level = int(dfinal.max().values)
        dict_out["Level"] = level

        # Permet de selectionner le plus petit niveau et ce même si aucun niveau
        # n'est activé.
        if level > 0:
            short_summary_level = level
            dmax, _ = xr.align(
                dtemp.sel(risk_level=short_summary_level),
                dfinal.where(dfinal == short_summary_level, drop=True),
            )
        elif level == 0:
            short_summary_level = dtemp.risk_level.min().values
            dmax = dtemp.sel(risk_level=short_summary_level)
        else:
            raise ValueError(
                f"In get_short_summary : level should be >=0. Got {level}."
            )

        # On va differencier les varaible pour savoir quoi retrourner
        # De base on prend la moyenne pour toutes les variables
        level_ds = dmax.mean(self.time_dimension)
        # On va changer pour certaines variables
        data_vars = dmax.data_vars
        max_var = [name for name in data_vars if "max" in name]
        if max_var != []:
            for var in max_var:
                level_ds[var] = dmax[var].max(self.time_dimension)
                LOGGER.debug(
                    "Taking max", func="get_short_summary", geo_id=geo_id, var=var
                )

        min_var = [name for name in data_vars if "min" in name]
        if min_var != []:
            for var in min_var:
                level_ds[var] = dmax[var].min(self.time_dimension)
                LOGGER.debug(
                    "Taking min", func="get_short_summary", geo_id=geo_id, var=var
                )
        # On va faire quelque chose de très spécifique
        # pour les valeurs représentatives :
        # prendre la plus critique
        risk_level = self.alea_conf.select_risk_level(level=short_summary_level)
        dict_comparison = risk_level[0].get_singleEvt_comparison()
        if dict_comparison is not None:
            for key in dict_comparison.keys():
                if "rep_value_" + key in level_ds and dict_comparison.get(
                    "category"
                ) in ["quantitative", "restrictedQuantitative"]:
                    try:
                        level_ds["rep_value_" + key] = self.get_critical_value(
                            dmax["rep_value_" + key],
                            dict_comparison[key]["comparisonOp"],
                        )
                    except Exception as e:
                        LOGGER.error(
                            f"Dictionnaire ayant provoqué l'erreur {dict_comparison}"
                        )
                        raise e

        LOGGER.debug(f"The comparison dictionnary is {dict_comparison}")

        if "summarized_density" in data_vars:
            level_ds["density"] = dmax["summarized_density"]
            level_ds = level_ds.drop("summarized_density")
            LOGGER.debug(
                "Replacing density by summarized density in short summary",
                func="get_short_summary",
                geo_id=geo_id,
            )
        if "risk_summarized_density" in data_vars:

            level_ds["density"] = dmax["risk_summarized_density"]
            level_ds = level_ds.drop("risk_summarized_density")
            if self._unique_warning:
                LOGGER.debug(
                    "Overwriting density by risk_summarized_density in short_summary",
                    func="get_short_summary",
                    geo_id=geo_id,
                    warning=self._unique_warning,
                )
                self.unique_warning = False

        LOGGER.debug(
            f"Variables de sortie {list(level_ds.data_vars)}",
            func="get_short_summary",
            geo_id=geo_id,
        )
        if "units" in dmax:
            level_ds["units"] = dmax["units"]
        level_ds["weatherVarName"] = dmax["weatherVarName"]
        # On s'assure que l'occurrence est bien True
        if level > 0:
            level_ds["occurrence"] = True
        dict_out["Values"] = self.get_values(level_ds)
        return dict_out

    def get_stepsize(self, ds_in):
        """Permet de récupérer le pas de temps du modèle (pour chaque
        échéance). La récupération de ce pas de temps permet ensuite de savoir
        "de combien sont écartés les échéances". On suppose que le dernier pas
        de temps (non renseigné) et le même que l'avant dernier.

        Args:
            ds_in (xr.dataset) : Le dataset pour lequel on doit recuperer la stepsize

        ToDo : Bouger cette fonction ailleurs ?
               On risque d'en avoir besoin dans le module de prétraitement...
        """
        stepsize = (
            ds_in[self.time_dimension]
            .diff(self.time_dimension, label="lower")
            .dt.seconds
            / 3600
        ).astype(np.int)
        stepsize.name = "stepsize"
        stepsize.attrs["units"] = "hours"
        step_comp = (
            stepsize.broadcast_like(ds_in[self.time_dimension])
            .shift({self.time_dimension: 1})
            .isel({self.time_dimension: -1})
            .expand_dims(self.time_dimension)
        )
        stepout = xr.merge([stepsize, step_comp]).astype(np.int)
        return stepout

    def get_values(self, ds_in, stepsize=None):
        """Fonction permettant à partir d'un dataset (pour un niveau et une
        zone donnée) d'extraire les informations extrayable (vis à vis des
        specs) pour le CDP.

        Args:
            ds_in (xr.dataset): Dataset with only 'evt' as dimension.
            stepsize (int) : model stepsize pour ce paramètre à cette échéance.
        """
        compo_id = self.alea_conf.generic_get("id")
        # On va maintenant vérifier que dst n'a que evt
        if set(ds_in.dims) != {"evt"}:
            raise ValueError(
                "In get_values we expect ds to have only 'evt' as dimension. "
                f"However we get {ds_in.dims}."
            )

        # On commence par virer les évènement nul
        # Ces évènements sont possibles car le dataset comprend, pour chaque niveau,
        # autant d'évènement qu'il y a pour le niveau contenant le plus d'évènement.

        dst = ds_in.dropna("evt", how="all")
        # On va regarder les variables vide
        # Il est possible que certaines varaibles soient vide
        # (si par ex. à un niveau on a pas la définition de cette variable)
        dc = dst.count()
        empty = []
        for var in dc.data_vars:
            if dc[var].values == 0:
                empty.append(var)
                LOGGER.debug(
                    f"Detected an empty variable {var}",
                    compo_id=compo_id,
                    var=var,
                    func="get_values",
                )

        # Parmis les variables certaines ne doivent pas apparaitre dans la sortie
        useless = set(
            [
                "areaName",
                "areaType",
                "units",
                "occurrence_event",
                "weatherVarName",
                "risk_density",
                "risk_summarized_density",
            ]
        )
        coords = set(dst.coords)
        # On recupere le set de variable pouvant apparaitre dans la sortie
        variable = sorted(
            set(dst.variables)
            .difference(coords)
            .difference(useless)
            .difference(set(empty))
        )
        value_list = []
        for var in variable:
            dict_value = {}
            if dst[var].size < 2:
                if var in [
                    "occurrence",
                    "density",
                    "summarized_density",
                    "risk_density",
                    "risk_summarized_density",
                ]:
                    dict_value["ValueType"] = var
                    dict_value["Unit"] = "1"
                    if var == "occurrence":
                        dict_value["Value"] = int(dst[var].values)
                    elif "density" in var:
                        dict_value["Value"] = round(float(dst[var].values), 2)
                    else:
                        raise ValueError(
                            f"Cas non traité {var}. "
                            "Revoir le code pour ajouter le cas dans le switch."
                        )
                else:
                    # Cas des valeurs sur les plaines/montagnes etc...
                    # Les retours sont effectués dans les unités qui
                    # nous ont été spécifiées pour le seuil.
                    dict_value["ValueType"] = var
                    if "units" in dst:
                        dict_value["Unit"] = dst["units"].values.tolist()[0]
                    else:
                        LOGGER.error(
                            "Unit not known. This very bad. "
                            f"This is for {dst['weatherVarName'].values}.",
                            compo_id=compo_id,
                            var=var,
                            func="get_values",
                        )
                        dict_value["Unit"] = "Not known ...."
                    dict_value["Value"] = round(float(dst[var].values), 2)
                    # On regarde si on a une info sur la "stepSize".
                    # Si on n'en a pas, on ne la renseigne pas.
                    dict_value["Param"] = {
                        "Name": dst["weatherVarName"].values.tolist()[0]
                    }

                    if stepsize is not None:
                        dict_value["Param"]["Stepsize"] = 0
                        if np.isfinite(stepsize.values) and 0 <= stepsize.values <= 255:
                            dict_value["Param"]["Stepsize"] = int(stepsize.values)
                        else:
                            LOGGER.error(
                                "Unvalid 'Stepsize' in 'Param' dict : "
                                f"{stepsize.values} given. Changed to 0.",
                                compo_id=compo_id,
                                var=var,
                                stepsizes=list(stepsize.values),
                                func="get_values",
                            )
                value_list.append(dict_value)
            else:
                LOGGER.debug(
                    f"Too many information to go the CDP for this variable {var}",
                    compo_id=compo_id,
                    var=var,
                    func="get_values",
                )
        return value_list
