"""
@package core.risk_generator
Risk generator module
"""
import copy
import operator
import numpy as np
import xarray as xr
from ftap_autom.promethee.src.conf.app_configuration import ALTITUDES_NC, MFLOG
from ftap_autom.promethee.src.utils.unit_converter import (
    convert_threshold,
    convert_dataarray,
)
from ftap_autom.promethee.src.utils.xr_utils import ARRAY_LOADER, MASK_LOADER, rounding
from ftap_autom.promethee.src.lib.aggregator import Aggregator, AGGREGATION_AFTER
from ftap_autom.promethee.src.lib.interpretor import logical_op, comp_op
import mflog


LOGGER = mflog.getLogger("risk.mod")
mflog.set_config(**MFLOG)
LOGGER = LOGGER.bind(module="risk")
xr.set_options(keep_attrs=True)
np.warnings.filterwarnings("ignore")


def is_aggregation_after(aggregation):
    res = False
    if aggregation is not None:
        if aggregation["method"] in AGGREGATION_AFTER.keys():
            res = True
    return res


def is_aggregation_before(aggregation):
    res = False
    if aggregation is not None:
        if aggregation["method"] not in AGGREGATION_AFTER.keys():
            res = True
    return res


class UnitaryEvent:
    """
     Classe gérant les évènement unitaires.
     Ces évènements peuvent sont calculés à partir d'un champ spécifique.
     Ils peuvent inclurent une différenciation plaine/montagne.
    """

    def __init__(self, event_conf, aggregate_dim=["latitude", "longitude"]):
        """ Initialisation d'un evenement elementaire simple

        Args:
            event_conf ([EventInterp]): L'interpreteur de l'evenement
            aggregate_dim : Liste des dimensions sur lesquels on aggrége.
              Est utilisé seulement si on a une fonction d'aggrégation.
        """
        self.event_conf = event_conf
        self.aggregate_dim = aggregate_dim
        self.field = self.loading_field("field")
        self.altitude_field = self.loading_altitude(
            self.event_conf.generic_get("altitude_file", ALTITUDES_NC)
        )
        self.loading_mask()

        if self.event_conf.doing_aggregation():
            LOGGER.info("Doing aggregation at event scale", func="__init__")
        self.representative_field = self.field * self.mask
        self.representative_field.name = self.field.name

        # On charge le fichier permettant de décrire l'altitude si besoin
        # if self.event_conf.generic_get("mountain") is not None:
        #    self.altitude_field = self.loading_altitude(
        #        self.event_conf.generic_get("altitude_file", ALTITUDES_NC)
        #    )
        self.event_value = xr.Dataset()

    def changing_threshold(self, dico):
        """This function change the threshold to match units of the field

        Args:
            dico (dict): Dictionnaire contenant le threshold et les unités du threshold

        Returns:
            dict: Le dictionnaire modifié.

        To Do : rajouter une information (sup ou inf) pour les variables quantitatives
        discretes (type Beaufort)
        """
        dico_t = dico.copy()
        if dico_t["units"] is not None:
            if dico_t["units"] != 1:
                dico_t["threshold"] = convert_threshold(
                    self.field, dico_t["threshold"], dico_t["units"]
                )
        return dico_t

    def get_grid_name(self):
        geo = self.event_conf.generic_get("geos")
        # 1. S'il faut ouvrir le masque (cas ou geos est un dictionnaire)
        #   on va compléter les infos et ouvrir le masque
        if isinstance(geo, dict) and "grid_name" in geo:
            grid_name = geo["grid_name"]
        elif hasattr(self.field, "PROMETHEE_z_ref"):
            grid_name = self.field.attrs["PROMETHEE_z_ref"]
        elif isinstance(geo, xr.DataArray):
            grid_name = geo.name
        else:
            raise ValueError(
                "Field should have an attribute named PROMETHEE_z_ref, "
                "or grid_name should be present in the geos dictionnary, "
                "or geos should be a named dataArray"
            )
        return grid_name

    def loading_mask(self):
        """ Enable to load mask.
            Mask can be already loaded in the configuration.
        Raises:
            ValueError: If all dimension for aggregation are not in the mask,
                raise an error.
        """
        geo = copy.deepcopy(self.event_conf.generic_get("geos"))
        # 1. S'il faut ouvrir le masque (cas ou geos est un dictionnaire)
        #   on va compléter les infos et ouvrir le masque
        if isinstance(geo, dict):
            # On s'assure d'avoir un nom de grille
            geo["grid_name"] = self.get_grid_name()
            # On verifie qu'on a bien l'info du chunking
            chunks = geo.get("chunks")
            if chunks is None:
                chunks = self.event_conf.generic_get("chunks_mask")
                geo["chunks"] = chunks
            # On specifie qu'on veut selectionner mask_id
            if geo.get("select_key") is None and "mask_id" in geo:
                geo["select_key"] = "mask_id"

        # On ouvre le mask. Attention geo peut deja etre un masque chargé (dans ce
        # cas on effectue uniquement le travail d'alignement)
        self.mask = MASK_LOADER(geo, self.field)
        for dim in self.aggregate_dim:
            if dim not in self.mask.dims:
                raise ValueError(f"{dim} should be in mask dimension")
        # On va maintenant regarder s'il faut aussi restreindre vis à vis de l'altitude
        alt_min = self.event_conf.generic_get("alt_min")
        alt_max = self.event_conf.generic_get("alt_max")
        if alt_min is not None or alt_max is not None:
            domain_altitude = self.altitude_field * self.mask

            if alt_min is not None:
                domain_altitude = (
                    domain_altitude.where(domain_altitude >= alt_min) * self.mask
                )
            if alt_max is not None:
                domain_altitude = (
                    domain_altitude.where(domain_altitude <= alt_max) * self.mask
                )
            restricted_mask = domain_altitude > -10000
            restricted_mask = restricted_mask.where(restricted_mask > 0)
            LOGGER.debug(
                f"We will restrict the mask to this altitude range {alt_min} {alt_max}",
                func="loading_mask",
            )
            # C'est pas très beau mais c'est moins moche ....
            self.mask = restricted_mask

    def loading_altitude(self, fname):
        """ Charge le champ d'altitude

        Args:
            fname (str): Nom du fichier a ouvrir
        """
        grid = self.get_grid_name()
        altitude_field = xr.open_dataset(fname)[grid]
        # On renomme les dimension pour avoir latitude et longitude (au lieu de
        # latitude_globe05)
        dict_dims = {}
        for x in altitude_field.dims:
            dict_dims[x] = x.split("_")[0]
        altitude_field = altitude_field.rename(dict_dims)
        altitude_field = rounding(altitude_field)
        return altitude_field

    def loading_field(self, field_key):
        """ Charge le champ correspondant et arrondie les latitudes/longitudes

        Args:
            field (DataArray,dict): DataArray déjà ouvert ou dictionnaire d'ouverture
            chunks (dict, optional): Dictionnaire permettant l'utilisation de dask.
                Defaults to None.
        Obsolete. Should be in reader of the dictionnary.
        """
        field = self.event_conf.generic_get(field_key)
        #  On va mettre un patch pour que ça fonctionne encore
        #        LOGGER.warning("Entering loading field", field =field)

        if (
            isinstance(field, dict)
            and "slice" in field
            and isinstance(field["slice"], (list, tuple))
        ):
            period_slice = [
                np.timedelta64(field["slice"][0], "h"),
                np.timedelta64(field["slice"][1], "h"),
            ]
            field["slice"] = {"step": period_slice}

            LOGGER.info(
                "Patch activated",
                patch=1,
                file=field.get("file"),
                func="loading_field",
            )

        if isinstance(field, dict):
            # On modifie les chunks
            if "reader" in field:
                field["reader"].setdefault(
                    "chunks", self.event_conf.generic_get("chunks_field")
                )
            else:
                field["reader"] = {
                    "chunks": self.event_conf.generic_get("chunks_field")
                }
        # Fin du patch.
        # Ce patch a vocation a etre supprimé

        da = ARRAY_LOADER(field)
        for dim in self.aggregate_dim:
            if dim not in da.dims:
                print(da)
                raise ValueError(f"{dim} should be in field dimension")
        # On va virer les dimnesion
        return da

    @staticmethod
    def get_risk(comp_op, field, threshold, **kwargs):
        """" Function created to be able for other children class to implement
        the risk another way

        Args:
            comp_op ([type]): Function which describe the comparison to perform
            field ([data_array]):  The field
            threshold ([float]): The threshold applied

        Returns:
            [data_array]: The risk. It returns a boolean for every pixel.
        Be carreful : there is no nan inside.  nan with a logical produce a False.
        So you may need to restrict this results to area where the mask is valid.

        """
        return comp_op(field, threshold)

    def compute_density(self, risk_field):
        """
        Calcul la densité d'un evenement.
        Ceci dans le but de renvoyer la densité quelque soit le cas
        (qu'on passe par une aggréagation par la moyenne/par la drr ...)

        Args:
            risk_field (data_array): Le champ de risque

        Returns:
            density: La densité de l'évènement.
        """
        agg_handler = Aggregator(risk_field, aggregate_dim=self.aggregate_dim)
        density = agg_handler.compute("density")
        return density

    def compute_summarized_density(self, risk_field):
        """
        Permet de calculer la densité résumée...
        """

        agg_handler_time = Aggregator(
            risk_field, aggregate_dim=self.event_conf.generic_get("time_dimension")
        )
        max_risk = agg_handler_time.compute("max") * self.mask
        agg_handler_space = Aggregator(max_risk, aggregate_dim=self.aggregate_dim)
        density = agg_handler_space.compute("density")
        return density

    def get_extremevalues(self, masked_field, original_unit):
        """
        get_extremevalues : Retourne les valeurs extremes

        Arguments:
            masked_field {data_array} -- Le data_array pour lequel on recherche
                le min et le max

        Returns:
            [(data_array,data_array)] -- Les deux data_array representant les extremes
        """
        category = self.get_variable_type()
        if category in ["quantitative", "restrictedQuantitative"]:
            agg_handler = Aggregator(masked_field, aggregate_dim=self.aggregate_dim)
            maxi = agg_handler.compute("max")
            mini = agg_handler.compute("min")
            mini = convert_dataarray(mini, original_unit)
            maxi = convert_dataarray(maxi, original_unit)
            return (mini, maxi)
        else:
            return (None, None)

    def get_variable_type(self):
        """
        Retourne le type de variable.
        S'il n'y  a pas de type, on retourne quantitative par défaut.
        """
        category = self.event_conf.generic_get("category")
        if category is None:
            category = "quantitative"
            LOGGER.info(
                "No category set. Setting quantitative by default.",
                func="get_variable_type",
            )

        return category

    def get_representative_values(
        self, masked_field, comparisonOp=None, original_unit=None
    ):
        """
        Permet de retourner la valeur representative du champ.
        Gere le cas d'aggrégation par la moyenne et par les drr
        (si l'operateur de comparaison est dans [<,<=,>,>=]).

        Args:
            masked_field ([type]): [description]
            comparisonOp : L'opérateur de comparaison (si besoin dans les drr)
            original_unit : L'unité dans laquelle on doit reconvertir les sorties
        Return:
            data_array : La valeur representative. Return None si la variable est
                qualitative/booleénne ou qu'une erreur s'est produite
        Log :
            erreur : Si l'operateur de comparaison n'est pas dans
                [<,<=,>,>=, inf, infegal, sup ,supegal] et la methode d'aggregation
                est drr.

        """
        category = self.get_variable_type()
        rep_value = None
        if category in ["quantitative", "restrictedQuantitative"]:
            agg_handler = Aggregator(masked_field, aggregate_dim=self.aggregate_dim)
            # On va mainteant recuperer la methode d'aggregation.
            # Soit on la connait deja, soit elle et dans la cle
            # aggregation_aval.
            if not self.event_conf.doing_aggregation():
                aggregation_aval = self.event_conf.generic_get("aggregation_aval")
                if aggregation_aval is not None:
                    LOGGER.debug(
                        "On a une info sur la methode bien qu'on soit "
                        f"en aval {aggregation_aval}",
                        func="get_representative_values",
                    )
                    # Cette methode d'agregation va devoir etre changee si on
                    # est dans le cas d'une methode qui fait intervenir les
                    # zones centrales.
                    aggregation = aggregation_aval
                else:
                    aggregation = {}
                    aggregation["method"] = "mean"
                    LOGGER.warning(
                        "No aggregation method supply for rep value. Taking mean.",
                        func="get_representative_values",
                    )
            else:
                aggregation = self.event_conf.generic_get("aggregation")

            LOGGER.debug(
                "Aggregation method",
                aggregation=aggregation,
                place="event",
                func="get_representative_values",
            )

            if aggregation["method"] not in AGGREGATION_AFTER.keys():
                aggregation_kwargs = aggregation.get("kwargs", {})
                rep_value = agg_handler.compute(
                    aggregation["method"], **aggregation_kwargs
                )
            elif aggregation["method"].startswith("requiredDensity"):
                thresh = aggregation["kwargs"]["dr"]
                my_operator = comparisonOp
                try:
                    if my_operator in [">", ">=", "sup", "supegal"]:
                        quantile = 1 - thresh
                        rep_value = agg_handler.compute(
                            "quantile", q=quantile
                        ).drop_vars("quantile")
                    elif my_operator in ["<", "<=", "inf", "infegal"]:
                        quantile = thresh
                        rep_value = agg_handler.compute(
                            "quantile", q=quantile
                        ).drop_vars("quantile")
                    else:
                        LOGGER.error(
                            f"get_representative_value : cas non connu {my_operator}",
                            func="get_representative_values",
                        )
                        return None
                except Exception as excpt:
                    LOGGER.error(
                        f"Echec de l'aggregation sur le field : {masked_field}",
                        operateur=my_operator,
                        quantile=quantile,
                        field=masked_field,
                        func="get_representative_values",
                    )
                    raise ValueError from excpt
            else:
                LOGGER.error(
                    f"Aggregation method {aggregation['method']} unknown",
                    func="get_representative_values",
                )
                raise ValueError("Case need to be included")
            if original_unit is not None:
                rep_value = convert_dataarray(rep_value, original_unit)

        return rep_value

    def compute_plain_and_mountain(self):
        """
        Calcul le risk et les differentes valeurs representatives/extreme que
        l'on soit en plaine ou en plaine + montagne.
        """
        mountain = self.event_conf.generic_get("mountain")
        if mountain is not None:
            mountainThreshold = self.event_conf.generic_get("mountainThreshold")
            mountain = self.changing_threshold(mountain)
            altitude_masked = self.altitude_field * self.mask
            plain_mask = (altitude_masked <= mountainThreshold) * self.mask
            plain_mask = plain_mask.where(plain_mask > 0)
            mountain_mask = (altitude_masked > mountainThreshold) * self.mask
            mountain_mask = mountain_mask.where(mountain_mask > 0)
        else:
            plain_mask = self.mask
        # On va maintenant changer les thresholds
        plain = self.changing_threshold(self.event_conf.generic_get("plain"))
        # On charge le field (comme ça on le fera qu'une seule fois)
        # field_p =self.field.load() * self.mask
        # field_p.name = self.field.name
        # Calcul du champ de risque
        plain_field = self.field * plain_mask
        plain_risk = self.get_risk(
            comp_op[plain["comparisonOp"]],
            plain_field,
            plain["threshold"],
            mask=plain_mask,
        )
        risk_field = plain_risk
        if mountain is not None:
            mountain_field = self.field * mountain_mask
            mountain_risk = self.get_risk(
                comp_op[mountain["comparisonOp"]],
                mountain_field,
                mountain["threshold"],
                mask=mountain_mask,
            )
            risk_field = plain_risk + mountain_risk
        # ===================
        # Calculs des valeurs non obligatoires #
        # ====================
        # Calcul de la densité
        if "density" in self.event_conf.generic_get("compute_list"):
            density = self.compute_density(risk_field * self.mask)
            # On rajoute la densité dans le dataset
            self.event_value["density"] = density

        # Calcul de la densité résumée
        if "summary" in self.event_conf.generic_get("compute_list"):
            density = self.compute_summarized_density(risk_field * self.mask)
            self.event_value["summarized_density"] = density
        # Calcul des extremes
        original_unit = self.event_conf.generic_get("plain")["units"]
        if "extrema" in self.event_conf.generic_get("compute_list"):
            field = self.representative_field * plain_mask
            field.name = self.representative_field.name
            (mini, maxi) = self.get_extremevalues(field, original_unit)
            if mini is not None:
                self.event_value["min_plain"] = mini
                self.event_value["max_plain"] = maxi
            if mountain is not None:
                field = self.representative_field * mountain_mask
                field.name = self.representative_field.name
                (mini, maxi) = self.get_extremevalues(field, original_unit)
                if mini is not None:
                    self.event_value["min_mountain"] = mini
                    self.event_value["max_mountain"] = maxi
        # Calcul des valeurs representatives.
        if "representative" in self.event_conf.generic_get("compute_list"):
            field = self.representative_field * plain_mask
            field.name = self.representative_field.name
            rep_value = self.get_representative_values(
                field, comparisonOp=plain["comparisonOp"], original_unit=original_unit
            )
            if rep_value is not None:
                self.event_value["rep_value_plain"] = rep_value
            if mountain is not None:
                field = self.representative_field * mountain_mask
                field.name = self.representative_field.name
                rep_value = self.get_representative_values(
                    field,
                    comparisonOp=plain["comparisonOp"],
                    original_unit=original_unit,
                )
                if rep_value is not None:
                    self.event_value["rep_value_mountain"] = rep_value
        # =========================
        # Maintenant on passe au calcul du risk en lui-meme
        # Ce calcul est obligatoire
        # =========================

        aggregation = self.event_conf.generic_get("aggregation")

        if is_aggregation_before(aggregation):
            # Cas ou le risque est calcule a partir de valeur aggregee (moyenne)
            aggregation_kwargs = aggregation.get("kwargs", {})
            aggregation_algo = aggregation["method"]
            # Aggregation sur la plaine
            agg_plain = Aggregator(plain_field, aggregate_dim=self.aggregate_dim)
            plain_field_t = agg_plain.compute(aggregation_algo, **aggregation_kwargs)
            # On definit le risque sur la plaine
            risk = self.get_risk(
                comp_op[plain["comparisonOp"]],
                plain_field_t,
                plain["threshold"],
                mask=plain_mask,
            )
            # Si on a une condition sur l'altitude
            if mountain is not None:
                agg_mountain = Aggregator(
                    mountain_field, aggregate_dim=self.aggregate_dim
                )
                mountain_field_t = agg_mountain.compute(
                    aggregation_algo, **aggregation_kwargs
                )
                mountain_risk = self.get_risk(
                    comp_op[mountain["comparisonOp"]],
                    mountain_field_t,
                    mountain["threshold"],
                    mask=mountain_mask,
                )
                risk = risk + mountain_risk
        else:
            # Cas où le risque est calcule a partir de valeur non aggregee (drr)
            # Si on aggrege au niveau de l'evenement (i-e chaque evenement de
            # maniere independante)
            if self.event_conf.doing_aggregation():
                risk = risk_field * self.mask
            else:
                # Si on aggrege au niveau du risque
                risk = risk_field
        return risk

    def compute_event(self):
        """
        Permet de calculer le risque.
        Si aggrégation il y a, elle peut être faite avant (moyenne/quantile/...)
        ou après la comparaison à l'opérateur (ddr/density)
        Cette fonction peut prendre en compte les conditions d'altitude
        (si elles ont été bien renseignées auparavant)

        Returns:
            dataarray : Le dataarray de risque pour cet évènement "unitaire".
        """
        agg_after = False
        if self.event_conf.doing_aggregation():
            aggregation = self.event_conf.generic_get("aggregation")
            if is_aggregation_after(aggregation):
                agg_after = True

        risk = self.compute_plain_and_mountain()

        if agg_after:
            agg_handler = Aggregator(risk, aggregate_dim=self.aggregate_dim)
            aggregation_kwargs = aggregation.get("kwargs", {})
            aggregation_algo = aggregation["method"]
            risk = agg_handler.compute(aggregation_algo, **aggregation_kwargs)

        if self.event_conf.doing_aggregation():
            # Si on est dans le cas de l'aggregation AMONT on sait si le risque est
            # activee ou non
            self.event_value["occurrence_event"] = risk
        LOGGER.debug(
            f"FieldName is {self.field.name}.",
            param=self.field.name,
            func="compute_event",
        )
        self.event_value["weatherVarName"] = ("tmp", [self.field.name])
        if hasattr(self.event_value, "min_plain"):
            self.event_value["units"] = ("tmp", [self.event_value.min_plain.units])
        return risk


class BertrandEvent(UnitaryEvent):
    """Classe permettant de gerer les evenements de type cumul de pluie/neige.
    Classe héritant de tous les attributs des évènements unitaires.
    """

    def __init__(self, event_conf, **kwargs):
        """Initialisation

        Args:
        cum_period : period de cumul

        On rajoute:
         - un champ correspondant au RR1 (ou autre RR non cumulé selon le fichhier)
         - On définit le champ representatif comme etant celui de RR1 cumulé sur
            n steps (mais dans le passé)
         - On définit le n comme étant (La période de cumul) /(la taille du pas
            temporel)

        On masque ici le champ RR1 (car celui est utilisé pour calculer le champ
            representatif)
        """
        LOGGER.debug("Initializing betrand risk", func="__init__")
        super().__init__(event_conf, **kwargs)
        self.RR1 = (self.loading_field("field_1") * self.mask).load()
        if self.field.GRIB_stepUnits != self.RR1.GRIB_stepUnits:
            raise ValueError(
                "Both cumul field not have the same stepUnits. "
                f"Simple field is {self.field.GRIB_stepUnits} and "
                f"cumulField is {self.RR1.GRIB_stepUnits}."
            )
        # On divise le temps sur lequel on prend des cumuls par la taille de la step
        # On obtient ainsi un risque qui est correctement défini.
        self.n = int(
            self.event_conf.generic_get("cum_period") / self.RR1.GRIB_stepUnits
        )
        self.time_dim = event_conf.generic_get("time_dimension")

        if self.n > self.field[self.time_dim].size:
            max_p = self.field[self.time_dim].size
        else:
            max_p = self.n
        self.representative_field = self.field.rolling(
            {self.time_dim: max_p}, min_periods=1
        ).max()
        self.representative_field.attrs.update(self.field.attrs)
        self.representative_field.name = self.field.name

    def get_risk(self, comp_op, field, threshold, mask, **kwargs):
        """Specify the risk for cumulative event such Rainfall

        Args:
            comp_op ([type]): L'opérateur de comparaison
            field ([type]): Le champ (RRn dans le futur)
            threshold ([type]): Le threshold pour l'evenement de type "betrand"
            mask (dataArray) : Le masque utilisé
        """
        # On masque en utilisant le masque fournit par le champ.
        LOGGER.debug("Entering get_risk for Bertrand", func="get_risk")
        RR1_masked = self.RR1 * mask
        aggregation = self.event_conf.generic_get("aggregation")

        if is_aggregation_before(
            aggregation
        ):  # is not None and aggregation["method"]  not in AGGREGATION_AFTER.keys():
            agg_handler = Aggregator(RR1_masked, aggregate_dim=self.aggregate_dim)
            aggregation_kwargs = aggregation.get("kwargs", {})
            aggregation_algo = aggregation["method"]
            RR1_masked = agg_handler.compute(aggregation_algo, **aggregation_kwargs)
        # LOGGER.warning("Field type %s, shape %s",type(field),field.shape)
        # On va remplacer self.n par la taille maxi du dataset. Sinon on ne peut
        # pas prendre de max
        if self.n > field[self.time_dim].size:
            max_p = field[self.time_dim].size
            LOGGER.debug(f"Going from {self.n} to {max_p} for rolling operation")
        else:
            max_p = self.n
        cum_max = field.rolling({self.time_dim: max_p}, min_periods=1).max()
        start = operator.and_(
            comp_op(cum_max, threshold), comp_op(RR1_masked, threshold / self.n)
        ).load()
        keep = operator.and_(
            comp_op(cum_max, threshold), comp_op(field, threshold / self.n)
        ).load()
        # On perd l'avantage de dask ici !!
        # LOGGER.warning("Before the (very very very) long do loop")
        risk = start.copy().load()
        # LOGGER.warning("After loading info %s",type(risk))
        for step in range(len(risk[self.time_dim].values)):
            current_risk = risk.isel({self.time_dim: step})
            if step > 0:
                previous_risk = risk.isel({self.time_dim: step - 1})
                current_risk = operator.or_(
                    previous_risk * keep.isel({self.time_dim: step}),
                    start.isel({self.time_dim: step}),
                )
                risk.isel({self.time_dim: step})[:] = current_risk[:].compute()
        # Fin de la perte de dask
        risk.attrs.update(self.RR1.attrs)
        # On va remettre des atrributs au risque
        LOGGER.debug("Getting out get_risk for Bertrand", func="get_risk")
        return risk


class EventInterface:
    """ Interface d'evenement.
        Permet d'initialiser les differentes types d'evenements
    """

    def get_event(self, event_conf, **kwargs):
        """ On recupere l'evenement qui nous interesse.

        Returns:
            [event]: Evènement a retourner
        """
        process = event_conf.generic_get("process")
        if process is None:
            event_handler = UnitaryEvent(event_conf, **kwargs)
        elif process == "Bertrand":
            event_handler = BertrandEvent(event_conf, **kwargs)
        else:
            raise ValueError("Process not known")
        return event_handler


class RiskGenericEvent:
    def __init__(self, risk_conf, aggregate_dim):
        """
        """
        global LOGGER

        self.risk_conf = risk_conf
        self.aggregate_dim = aggregate_dim
        self.event_list = []
        dict_id = self.risk_conf.get_ids()
        if dict_id != {}:
            LOGGER = LOGGER.bind(**dict_id)

        for n in range(self.risk_conf.generic_get("nbElementaryEvent")):
            event_conf = self.risk_conf.get_event(n)
            unitEvent = EventInterface().get_event(
                event_conf=event_conf, aggregate_dim=aggregate_dim
            )
            self.event_list.append(unitEvent)
        self.other_values = xr.Dataset()
        self.spatial_risk = xr.Dataset()
        self.risk = self.compute_risk()

    def compute_risk(self):
        """ Fonction qui calcul un niveau de risque.
        Cette fonction combien différents évènements.
        Le dataset de sortie n'est par contre pas généré. Seul le risque est ici fait.
        """

        def recursive_list(event_list, op):
            """
            Perform logical comparison
            op[n](evt[n+1],op[n-1](evt[n],...,op[0](evt[1],evt[0
            Inputs :
            - event_list : List of binary events [n+1]
            - op : List of logical operation (and/or) [n]
            """

            if len(op) + 1 != len(event_list):
                raise ValueError(
                    "Length of operands and operators are not good. "
                    f"Operand is {len(event_list)} and Operator is {len(op)}. "
                    f"Operand should be of length {len(op)+1}."
                )
            local_event_list = event_list.copy()
            local_op = op.copy()
            res = logical_op[local_op[0]](local_event_list[1], local_event_list[0])
            local_event_list.pop(0)
            local_op.pop(0)
            if local_op == []:
                return res
            else:
                local_event_list[0] = res
                return recursive_list(local_event_list, local_op)

        events = [event.compute_event() for event in self.event_list]
        op_list = self.risk_conf.get_logicalOpList()
        if op_list is None or len(op_list) == 0:
            risk = events[0]
        else:
            risk = recursive_list(events, op_list)

        aggregation = self.risk_conf.generic_get("aggregation")

        if aggregation is not None:
            mask = self.get_mask()
            ds_risk = risk * mask
            self.spatial_risk = ds_risk
            agg_handler = Aggregator(ds_risk, aggregate_dim=self.aggregate_dim)
            # Ajout pour avoir la densite de l'evenement combine en sortie
            if "density" in self.risk_conf.generic_get("compute_list"):
                self.other_values["risk_density"] = agg_handler.compute("density")

            if "summary" in self.risk_conf.generic_get("compute_list"):
                agg_handler_time = Aggregator(
                    risk, aggregate_dim=self.risk_conf.generic_get("time_dimension")
                )
                max_risk = agg_handler_time.compute("max") * mask
                agg_handler_space = Aggregator(
                    max_risk, aggregate_dim=self.aggregate_dim
                )
                self.other_values[
                    "risk_summarized_density"
                ] = agg_handler_space.compute("density")

            # On caclul maintenant l'occurrence du risque
            aggregation_kwargs = aggregation.get("kwargs", {})
            aggregation_algo = aggregation["method"]
            risk = agg_handler.compute(aggregation_algo, **aggregation_kwargs)
        elif self.risk_conf.generic_get("aggregationType") == "downStream":
            mask = self.get_mask()
            ds_risk = risk * mask
            self.spatial_risk = ds_risk

        return risk

    def get_mask(self):
        """
        Fonction qui permet de retourner le masque comme
        l'union du masque des evenenements
        """
        mask_list = []
        for i, evt in enumerate(self.event_list):
            mask_list.append(
                self.event_list[i].mask.expand_dims("place").assign_coords(place=[i])
            )
        return xr.concat(mask_list, dim="place").max("place")

    def return_spatial_information(self):
        """
        Attention ce risque n'est pas toujours pertinent.
        Mettre un verificateur de "pertinence".
        """
        return self.spatial_risk

    def make_output_dataset(self):
        """
        make_output_dataset : Cree le dataset de sortie.
            On vire les dimensions ajoutees au cours du process et qui ne servent
            a rien  en sortie (afin d'avoir un dataset propre)

        Returns:
            dataset -- Dataset contenant de nombreuses variables representatives
                du risque.
        """
        output_dataset = xr.Dataset()
        for i, event in enumerate(self.event_list):
            exp_dataset = event.event_value.expand_dims(dim="evt").assign_coords(
                evt=[i]
            )
            output_dataset = xr.merge([output_dataset, exp_dataset])

        # Ajout pour avoir la densité combinées en sortie.
        for var in self.other_values.data_vars:
            #        if hasattr(self, "density"):
            output_dataset[var] = self.other_values[var]  # density
        # =======================
        # On va ajouter les noms de zone et les type de zones.
        # =======================
        # On recupere le masque du dernier evenement pour savoir comment on a
        # cropper
        # Improvement :
        #  Mettre une condition si on n'a pas de masque (ou pas aggrégé)
        ds_mask = self.event_list[0].mask
        ds_mask = self.get_mask()
        output_dataset["occurrence"] = self.risk
        # On check que les variables sont bien presentes.
        if hasattr(ds_mask, "label"):
            output_dataset["areaName"] = ds_mask["label"]
        if hasattr(ds_mask, "areaName"):
            output_dataset["areaName"] = ds_mask["areaName"]
        else:
            output_dataset["areaName"] = (
                "id",
                ["unknown" for i in range(ds_mask.id.size)],
            )
        if hasattr(ds_mask, "type"):
            output_dataset["areaType"] = ds_mask["type"]
        if hasattr(ds_mask, "areaType"):
            output_dataset["areaType"] = ds_mask["areaType"]
        else:
            output_dataset["areaType"] = (
                "id",
                ["unknown" for i in range(ds_mask.id.size)],
            )
        return output_dataset.squeeze("tmp")
