""" formatter.py : String formatter util module
Allows you to insert dynamic values inside strings
"""
# Built-in imports
from copy import deepcopy
from datetime import datetime, timedelta

# Third parties imports
import mflog

# Own package imports
from ftap_autom.promethee.src.conf.app_configuration import (
    MFLOG_FILE,
    MFLOG_CONF,
    MFLOG_LEVEL,
    TEXT_FORMAT,
    TAGS_FORMAT,
    DATETIME_FMT,
)

# Logging
mflog.set_config(
    json_file=MFLOG_FILE, json_minimal_level=MFLOG_LEVEL, override_files=MFLOG_CONF
)
LOGGER = mflog.getLogger("formatter.mod")
LOGGER = LOGGER.bind(module="formatter")


def format_datestr(datestr, dt, fmt=DATETIME_FMT):
    """format_datestr : Replace date tags '[date]' or '[date-1]' in a
    str by a given formatted datetime.

    Args:
        datestr (str): String to format
        dt (datetime.datetime): Datetime to apply
        fmt (str, optional): Datetime format. Defaults to DATETIME_FMT.

    Returns:
        str : Formatted string
    """
    if not isinstance(datestr, str):
        return datestr

    try:
        new_datestr = datestr.replace("[date]", dt.strftime(fmt))
        new_datestr = new_datestr.replace(
            "[date-1]", (dt - timedelta(days=1)).strftime(fmt)
        )
    except AttributeError as attrib_error:
        LOGGER.error(attrib_error)
        new_datestr = datestr
    return new_datestr


class Formatter:
    """ Formatter : Class for formatting sentences with date tags. It will
    replace tags like '[J]' by the reference date weekday name, and tags like
    '[J+n]' by the reference_date + n days weekday name, and month name if
    necessary
    """

    def __init__(self, text, language="french", text_format_conf=TEXT_FORMAT):
        """__init__

        Args:
            text (str): Sentence to format
            language (str, optional): Language. Defaults to "french".
            text_format_conf (dict, optional): Rules of text formatting.
                Defaults to TEXT_FORMAT.
        """
        self.original_text = text
        self.language = language

        # Dates formatting features
        self.weekdays = text_format_conf[language]["weekdays"]
        self.months = text_format_conf[language]["months"]

    def format_dates(self, reference_datetime):
        """format_dates : replaces tags like '[J]' by the reference date weekday name,
        and tags like '[J+n]' by the reference_date + n days weekday name,
        and month name if necessary

        Args:
            reference_datetime (datetime.datetime): Reference datetime

        Returns:
            str: Formatted text
        """
        text = deepcopy(self.original_text)
        weekday_number = reference_datetime.weekday()
        text = text.replace("[J]", self.weekdays[weekday_number])
        for i in range(1, 14):
            day_name = self.weekdays[(weekday_number + i) % 7]
            if i > 5:
                datetime_i = reference_datetime + timedelta(days=i)
                monthday = datetime_i.day
                month = self.months[datetime_i.month - 1]
                day_name += " {} {}".format(monthday, month)

            text = text.replace("[J+{}]".format(i), day_name)
        return text


class TagFormatter:
    """ TagFormatter: Format a str containing tags of type '[key:func]'.
    Its like the vortex standard.
    """

    def __init__(self, tags_format_conf=TAGS_FORMAT):
        """__init__

        Args:
            tags_format_conf (dict, optional): Rules  of text formatting.
                Defaults to TAGS_FORMAT.
        """
        self.time_format = tags_format_conf["time"]
        self.datetime_format = tags_format_conf["datetime"]

    def format_tags(self, original_text, tags=dict()):
        """format_tags : Format a str containing tags of type '[key:func]'.
        Its like the vortex standard.

        Args:
            original_text (str): Str to format
            tags (dict, optional): Dictionnary (key, value) containing the
            values to insert in the str. The values formatting is done
            according to the tags_format_conf. Defaults to dict().

        Returns:
            str: formatted str
        """
        text = deepcopy(original_text)

        for key, value in tags.items():
            # Raw key formatting
            text = text.replace("[{}]".format(key), str(value))

            # Time formatting
            if isinstance(value, int):
                for func, fmt in self.time_format.items():
                    text = text.replace("[{}:{}]".format(key, func), fmt.format(value))

            # Datetime formatting
            try:
                value_dt = datetime.strptime(value, DATETIME_FMT)
                for func, fmt in self.datetime_format.items():
                    text = text.replace(
                        "[{}:{}]".format(key, func), value_dt.strftime(fmt)
                    )
            except TypeError:
                pass
            except ValueError:
                pass

            # Geometry formatting
            text = text.replace("[{}:area]".format(key), str(value).upper())

        return text
