"""Ce module regroupe quelques fonctions manipulant des arrays et utiles dans
le projet PROMETHEE."""
import os
import xarray as xr
import numpy as np
import mflog
import datetime as dt
import operator
from fastcore.dispatch import TypeDispatch
from ftap_autom.promethee.src.conf.app_configuration import (
    ALL_GRID_NC,
    MFLOG,
    DATETIME_FMT,
)


class InputError(ValueError):
    """InputError : gere les erreur dûes aux entrees
    """

    pass


xr.set_options(keep_attrs=True)
LOGGER = mflog.getLogger("utils")
mflog.set_config(**MFLOG)
LOGGER = LOGGER.bind(module="xr_utils")


def rounding(da):
    """Perform the rounding of latitude/longitude
    Args:
        da (dataArray/dataSet):  Le dataset auquel on peut changer les coordonnées

    Returns:
        dataArray/dataSet: Le dataset avec les coordonnées changées
    """
    da["latitude"] = da.latitude.round(5)
    da["longitude"] = da.longitude.round(5)
    return da


def mask_set_up(mask: xr.DataArray, da: xr.DataArray):
    """
    This function ensure that there is no 'holes' in the mask setting.
    To do so, we compare it to the dataarray.
    It make the hypothesis that both dataarray leave in the same space.
    """

    min_lat = np.max([mask.latitude.values.min(), da.latitude.values.min()])
    max_lat = np.min([mask.latitude.values.max(), da.latitude.values.max()])
    index_lat_field = operator.and_(da.latitude >= min_lat, da.latitude <= max_lat)
    min_lon = np.max([mask.longitude.values.min(), da.longitude.values.min()])
    max_lon = np.min([mask.longitude.values.max(), da.longitude.values.max()])
    index_lon_field = operator.and_(da.longitude >= min_lon, da.longitude <= max_lon)
    dempty = xr.Dataset()
    dempty["longitude"] = da.longitude.isel(longitude=index_lon_field).values
    dempty["latitude"] = da.latitude.isel(latitude=index_lat_field).values
    return mask.broadcast_like(dempty)


def loading_and_set_up(geo: dict, da: xr.DataArray) -> xr.DataArray:
    """Permet de charger le masque et de s'assurer qu'il n'y a pas de discontinuitées
    dans les latitudes/longitudes

    Args:
        geo (dict): Le dictionnaire décrivant ce qu'il y a a charger.
            Doit contenir au minimun "grid_name", "file".
            Peut aussi contenir "chunks","select_key"
        da (xr.DataArray): Le champ avec lequel on va se comparer pour boucher les
        discontinuités

    Returns:
        [xr.DataArray]: Le masque bien mis en forme.
    """
    mask = loading_mask(geo)
    return mask_set_up(mask, da)


def loading_mask(geo: dict):
    """[summary]

    Args:
        geo (dict): [description]
        grid_name ([type], optional): [description]. Defaults to None.
        select_key (str, optional): [description]. Defaults to "mask_id".

    Raises:
        ValueError: [description]
        ValueError: [description]
    """

    chunks = geo.get("chunks")
    grid_name = geo.get("grid_name")
    if grid_name is not None:
        mask = xr.open_dataset(geo["file"], chunks=chunks)[grid_name]
    else:
        raise ValueError("Grid name should be provided")
    # On renomme les dimension pour avoir latitude et longitude (au lieu de
    # latitude_globe05)
    dict_dims = {}
    for x in mask.dims:
        dict_dims[x] = x.split("_")[0]
    mask = mask.rename(dict_dims)

    # On selectionne les masques qui nous interessent au sein du lot de
    # masque.
    select_key = geo.get("select_key")
    if select_key is not None:
        mask_id = geo.get(select_key)
        if mask_id is not None:
            if isinstance(mask_id, str):
                ids = set([mask_id])
            else:
                ids = set(mask_id)
            ds_id = set(mask.id.values)
            intersection = ids.intersection(ds_id)
            if intersection != ids:
                LOGGER.warning(
                    "Toutes les id n'ont pas ete trouves dans la liste {ids} "
                )
            mask = mask.sel(id=list(intersection))
        else:
            raise (
                ValueError(
                    "A list of id should be provided in the configuration dict "
                    f"as select_key '{select_key}' is not None."
                )
            )
    else:
        LOGGER.debug("No selection on mask is performed")
    mask = rounding(mask)
    # On va rajouter les types et les noms s'ils existent
    # On doit donc reouvrir le dataset pour aller chercher ces variables.
    # Elles seront ajoutees en tant que dimension
    dmask = xr.open_dataset(geo["file"], chunks=chunks)

    if ("areaName" in dmask) and ("areaName" not in mask.coords):
        mask = xr.merge([mask, dmask["areaName"].load().sel(id=mask.id)])
        # Permet de mettre AreaName en tant que cooordonnées
        # et non en tant que variable
        # On retourne donc bien un dataArray
        mask = mask.swap_dims({"id": "areaName"}).swap_dims({"areaName": "id"})[
            grid_name
        ]

    if "areaType" in dmask and "areaType" not in mask.coords:
        mask = xr.merge([mask, dmask["areaType"].load().sel(id=mask.id)])
        # Permet de mettre AreaName en tant que cooordonnées
        # et non en tant que variable
        # On retourne donc bien un dataArray
        mask = mask.swap_dims({"id": "areaType"}).swap_dims({"areaType": "id"})[
            grid_name
        ]
    return mask


# Ouverture d'un fichier avec selection
def identity_da(da: xr.DataArray):
    """
    identity_da : Return a copy of the dataArray

    Args:
        da (xr.DataArray): the input dataArray

    """
    return da.copy(deep=True)


def open_dataarray(field: dict):
    """
    This function open a data and "select" infomration for some dimension.
    Field : Dictionnary

    This dictionnary contains :
        - "file" : The filename

    It can contains :
        - "sel" :  Dictionnary of selection
            (variable : list of value of the dimension)
        - "slice" : Dictionnary of selection
            (variable : start and end value (of the dimension))
        - "td_slice": Dictionnary of selection
            (variable : start and end value (of the dimension)).
            Use a timedelta converter.
        - "isel" : Dictionnary of selection
            (variable : list of position of the dimension)
        - "islice": Dictionnary of selection
            (variable : start and end position value (of the dimension))
        - "reader" : Information for the backend when opening file.
    """
    try:
        da = xr.open_dataarray(field["file"], **field.get("reader", {}))
    except ValueError as e:
        # On va essayer de l'ouvrir en tant que dataset et d'en extraire l'unique
        # variable ayant des dimensions
        LOGGER.warning(
            f"We get an error {repr(e)} when opening. We will try a backup solution"
        )
        ds = xr.open_dataset(field["file"], **field.get("reader", {}))
        variables = []
        for x in ds.data_vars:
            if len(ds[x].dims) >= 1:
                variables.append(x)
        if len(variables) == 1:
            da = ds[variables[0]]
            LOGGER.warning(
                "Back up solution which consist to select the only "
                "variable with coordinates works."
            )
        else:
            LOGGER.error(
                "Trying to see if there is only one variable without dimension"
            )
            raise (e)

    isel_dict = {}
    sel_dict = {}
    key_isel = set()
    key_sel = set()
    key_islice = set()
    key_slice = set()

    if "isel" in field:
        isel_dict.update(field["isel"])
        key_isel = set(field["isel"].keys())

    if "sel" in field:
        sel_dict.update(field["sel"])
        key_sel = set(field["sel"].keys())

    if "islice" in field:
        for var in field["islice"]:
            isel_dict[var] = slice(field["islice"][var][0], field["islice"][var][1])
        key_islice = set(field["islice"].keys())

    if "slice" in field:
        for var in field["slice"]:
            sel_dict[var] = slice(field["slice"][var][0], field["slice"][var][1])
        key_slice = set(field["slice"].keys())

    if "td_slice" in field:
        for var in field["td_slice"]:
            # On reconverti les dates en np.datetime64
            start = dt.datetime.strptime(field["td_slice"][var][0], DATETIME_FMT)
            stop = dt.datetime.strptime(field["td_slice"][var][1], DATETIME_FMT)
            if var in sel_dict:
                LOGGER.info(
                    f"On réecrit la selection pour {var}. "
                    f"La valeur etait {sel_dict} et devient {slice(start,stop)}."
                )
            sel_dict[var] = slice(start, stop)
        key_slice = set(field["td_slice"].keys())

    union_set = key_slice.union(key_sel, key_islice, key_isel)
    nb_key = len(key_sel) + len(key_isel) + len(key_slice) + len(key_islice)

    if len(union_set) < nb_key:
        LOGGER.info(
            "Some keys are present twice. "
            f"key_sel{list(key_sel)}, key_isel {list(key_isel)}, "
            f"key_slice {list(key_slice)} "
            f"key_islice {list(key_islice)}. \n"
            f"Number of unique key is {len(union_set)} "
            f"while total key numbers is {nb_key}"
        )
        LOGGER.info(f"Corresponding dictionnary are {sel_dict} and {isel_dict}")

    dout = da.isel(isel_dict).sel(sel_dict)
    LOGGER.debug(
        "Variable shape after openneing",
        shape=dout.shape,
        dim=dout.dims,
        process=os.getpid(),
    )
    return rounding(dout)


# On pourra l'etendre vers d'autres types de fonction en fonction du type d'entree
ARRAY_LOADER = TypeDispatch([open_dataarray, identity_da, None])
MASK_LOADER = TypeDispatch(
    [loading_mask, loading_and_set_up, identity_da, mask_set_up, None]
)

#
# Fonction permettant le changement de grilles.
#


def from_0_360(ds_grid):
    """Passage d'une grille [0:360] à [-180:180]

    Args:
        ds_grid (dataarray): Le datarray à transformer

    Returns:
        [dataarray]: Le dataarray transformé
    """
    longitude_name = [x for x in ds_grid.dims if "longitude" in x][0]
    new_longitude = np.where(
        ds_grid[longitude_name] < 180,
        ds_grid[longitude_name],
        ds_grid[longitude_name] - 360,
    )
    ds_grid[longitude_name] = new_longitude
    ds_grid = ds_grid.sortby(longitude_name)
    return ds_grid


def from_center_to_0_360(ds_grid):
    """Passage d'une grille [-180:180] à [0:360]

    Args:
        ds_grid (dataarray): Le datarray à transformer

    Returns:
        [dataarray]: Le dataarray transformé
    """
    longitude_name = [x for x in ds_grid.dims if "longitude" in x][0]
    new_longitude = np.where(
        ds_grid[longitude_name] >= 0,
        ds_grid[longitude_name],
        360 + ds_grid[longitude_name],
    )
    ds_grid[longitude_name] = new_longitude
    ds_grid = ds_grid.sortby(longitude_name)
    return ds_grid


def change_grid(da, grid_out):
    """change_grid : Fonction qui permet d'effectuer un changement de grille au
    sein de PROMETHEE Ce Changement de grille est fait par interpolation
    bilinéaire. Les résultats sont donc à prendre avec des pincettes si quelque
    chose de précis est nécessaire.

    Le nom de la grille d'entree est pour l'instant non utilisee.
    Servira dans le futur pour gerer les différents cas un peu etrange.

    Args:
        da (xr.Dataarray): Le dataset d'entrée que l'on doit interpoler.
            Ses dimensions comprennent latitude, longitude.

        grid_out (str): Le nom de la grille de sortie.
            Doit être connue dans ALL_GRID_NC
    """
    # Ouverture du fichier contenant les grilles
    dgrid = xr.open_dataset(ALL_GRID_NC, cache=False)
    LOGGER.info("Grille d'arrivee %s", grid_out)
    dgrid_out = dgrid[grid_out].rename(
        {"latitude_%s" % grid_out: "latitude", "longitude_%s" % grid_out: "longitude"}
    )
    # On change les coordonnées de la grille d'entree pour que ça corresponde
    # à la grille de sortie
    if dgrid_out.longitude.max().values > 180 and da.longitude.min().values < 0:
        da_b = from_center_to_0_360(da)
    elif dgrid_out.longitude.min().values < 0 and da.longitude.max().values > 180:
        da_b = from_0_360(da)
    else:
        da_b = da

    step_lat_new = (dgrid_out.latitude[1] - dgrid_out.latitude[0]).values
    min_lat = da_b.latitude.min().values - abs(step_lat_new)
    max_lat = da_b.latitude.max().values + abs(step_lat_new)
    if step_lat_new < 0:
        slice_lat = slice(max_lat, min_lat)
    else:
        slice_lat = slice(min_lat, max_lat)

    step_lon_new = (dgrid_out.longitude[1] - dgrid_out.longitude[0]).values
    min_lon = da_b.longitude.min().values - abs(step_lon_new)
    max_lon = da_b.longitude.max().values + abs(step_lon_new)
    if step_lon_new < 0:
        slice_lon = slice(max_lon, min_lon)
    else:
        slice_lon = slice(min_lon, max_lon)

    # Cas simple : les grilles sont centrées de la meme maniere.
    dout = da_b.interp_like(dgrid_out.sel(latitude=slice_lat).sel(longitude=slice_lon))
    dgrid.close()
    return dout


# Fonctions gérant les cumuls
def compute_stepsize(ds, var="step"):
    """
    compute_stepsize : Cette fonction retourne le pas de temps (en chaque point).
    Ce pas de temps peu varier d'une step à une autre.
    On ne fait donc pas confiance aux méta-données du fichier pour nous renseigner
    la dessus.

    Args:
        ds (dataset ou dataarray): L'array sur laquelle on veut calculer
            le pas de temps
        var (str): la coordonnée servant à calculer la stepsize.
            Cette coordonnée doit être un timedelta64s ou un datetime64
    Exemple :
       da.step = 1,2,5,10

    Returns : A dataset with stepsize as variable and "var" as coordinate.
              Stepsize is express in hour (but as int).
    """

    # On va perdre la dernière occurrence de stepsize via diff
    stepsize = (ds[var].diff(var).dt.seconds / 3600).astype(np.int)
    stepsize.name = "stepsize"
    stepsize.attrs["units"] = "hours"
    # On rajoute la dernière occurrence en remetteant stepsize sur la grille
    # temporelle de la variable d'entree
    # On suppose que la dernière occurrence est identique à l'avant dernière.

    step_comp = (
        stepsize.broadcast_like(ds[var])
        .shift({var: -1})
        .isel({var: 0})
        .expand_dims(var)
    )
    stepout = xr.merge([stepsize, step_comp]).astype(np.int)
    return stepout["stepsize"]


def desagregate_sum_values(ds, da_step, stepout=1, var="step"):
    """interpolate_values [summary]

    Args:
        ds (Dataset):  Le dataset pour lequel il faut reinterpoler
        stepvar ([type]): Un dataarray donnant la valeur du stepSize
            à chaque pas de temps de l'array initale.
        stepout (int, optional): La valeur (en heure) du pas de temps en sortie.
               Doit être supérieur ou égal au plus petit pas de temps.

    Hypothèse : On suppose que lorsqu'il y a besoin d'interpoler,
    la valeur est celle qui est après.
    On suppose de plus que la série est triée selon la variable
    d'interpolation (et que celle ci dépend du temps)
    """

    if not isinstance(da_step, xr.core.dataarray.DataArray):
        raise (InputError("da_step should be a dataarray. Get %s" % type(da_step)))

    ds_fill = (ds / da_step * stepout).resample({var: str(stepout) + "H"}).bfill()
    ds_fill.attrs["stepUnits"] = stepout
    ds_fill.attrs["GRIB_stepUnits"] = stepout
    ds_fill.attrs["history"] = (
        "Desagregation perform. The hypothesis that we need to fill by the forward "
        "value was made. The value has been desagregated by taking the mean."
    )
    return ds_fill


def compute_sum_futur(da, stepout=6, var="step", da_step=None):
    """
    compute_sum_futur
    Permet de calculer une somme dans le futur.
    Ainsi à l'instant t, on a l'information sommée entre t et t+stepout.
    On echantillone en fonction du plus petit StepSize.
    Ainsi si une série comprend des stepSize entre 1H et 6H,
    l'espacement entre les steps sera de 1H.
    Args:
        da (xr.DataArray): Dataarray
        stepout (int, optional): Nombre d'heure sur lesquelles on somme.
            Defaults to 6.
        var (str, optional): Variable sur laquelle on somme.
            Defaults to "step".

    Returns:
        xr.DataArray: Nouveau DataArray
    """
    if not isinstance(da, xr.core.dataarray.DataArray):
        raise InputError(f"da should be a dataarray. Get {type(da_step)}")

    # On calcul l'ecartement entre les pas de temps successifs.
    # Cet ecartement est propage aussi pour la premiere valeure.
    stepsize = compute_stepsize(da, var=var)
    if da_step is not None:
        # On va regarder si ce qu'on nous donne en entree est coherent avec les
        # espacements ...
        if (stepsize != da_step).any():
            LOGGER.warning(
                "Attention les pas de temps ne correspondent pas aux espacements.\n"
                f"Du coté des pas de temps on a {da_step.values} tandis que les "
                f"espacements entre deux temps successifs sont {stepsize.values} "
            )
            if (stepsize > da_step).any():
                raise InputError(
                    "Au moins un espacement est plus important que la 'stepSize' "
                    "correspondante. Cas pas encore implementé ... "
                    "et sans doute jamais..."
                )
        else:
            LOGGER.info("Les steps et l'ecartement correspondent")
        stepsize = da_step
    else:
        LOGGER.info(
            "Les pas de temps n'ont pas ete fournis. "
            "On se base sur les calculs interne."
        )
    stepmin = stepsize.min().values
    LOGGER.debug(
        "Stepsizes : ", stepmin=stepmin, stepout=stepout, da_step=da_step, var=var
    )

    if stepmin != stepsize.max().values or stepmin > stepout:
        if stepmin > stepout:
            RR = desagregate_sum_values(da, stepsize, stepout=stepout, var=var)
            LOGGER.debug(
                "Comparing max values", rr_max=RR.max().values, da_max=da.max().values
            )
        else:
            # Dans ce cas, on va avoir besoin de desagreger avant de faire la somme
            RR = desagregate_sum_values(da, stepsize, stepout=stepmin, var=var)
    else:
        RR = da.copy()

    # Si on doit cumuler
    if stepout >= stepmin:
        n = int(stepout / stepmin)
    else:
        # On va selectionner seulement par pas de stepmin
        LOGGER.debug(f"'{var}' avant selection", var=var, values=RR[var].values)
        RR = RR.sel({var: da[var]})
        LOGGER.debug(f"'{var}' après selection", var=var, values=RR[var].values)
        LOGGER.debug(
            "Impossible de stepout, entrées espacées de stepmin, on garde stepmin",
            stepout=stepout,
            stepmin=stepmin,
            var=var,
        )
        # Si on desagrege on conserve toutes les valeurs.
        n = 1

    LOGGER.debug("valeurs finales", stepout=stepout, n=n, stepmin=stepmin, var=var)
    # On va maintenant calculer la somme sur les "n" prochaines heures.
    nb_step = RR[var].size
    # Le debut de la serie
    RR6_beg = (
        RR.rolling({var: n}, min_periods=1)
        .sum()
        .shift({var: -(n) + 1})
        .isel({var: range(nb_step - n + 1)})
    )
    # La fin de la serie
    RR6_end = (
        RR.shift({var: -n + 1})
        .rolling({var: n}, min_periods=1)
        .sum()
        .isel({var: range(nb_step - n + 1, nb_step)})
    )
    # On remet le nom correctement. Indispensable pour le merge.
    RR6_beg.name = da.name
    RR6_end.name = da.name
    # On redonne aux attributs la valeur qu'ils avaient en entrant.
    # Attention, c'est certainement abusif et certains attributs devraient
    # certainement être filtré (typ Grib_StepUnits si présent)
    RR6_beg.attrs = RR.attrs
    RR6_end.attrs = RR.attrs
    # On merge
    dout = xr.merge([RR6_beg, RR6_end])
    # Maintenant on rempli des infos dans les attributs.
    dout[da.name].attrs["accum_hour"] = stepout
    dout[da.name].attrs["accum_type"] = "futur"
    if "GRIB_startStep" in dout[da.name].attrs:
        dout[da.name].attrs["GRIB_endStep"] = (
            dout[da.name].attrs["GRIB_startStep"] + stepout
        )
        dout[da.name].attrs["GRIB_stepRange"] = "{}-{}".format(
            dout[da.name].attrs["GRIB_startStep"], dout[da.name].attrs["GRIB_endStep"]
        )
    return dout


def slice_dataarray(
    da: xr.core.dataarray.DataArray,
    start: object = None,
    stop: object = None,
    step: object = None,
    dim: str = "valid_time",
) -> xr.core.dataarray.DataArray:
    """slice_dataarray : slice a given dataarray along a given dimension

    Args:
        da (xr.core.dataarray.DataArray): DataArray to slice.
        start (object, optional): Start of the slice. If None, the minimum value
            along the given dimension is taken. Defaults to None.
        stop (object, optional): Stop of the slice. If None, the maximum value
            along the given dimension is taken. Defaults to None.
        step (object, optional): Step of the slice. Defaults to None.
        dim (str, optional): Dimension along which to apply slicing.
            Defaults to "valid_time".

    Returns:
        xr.core.dataarray.DataArray: Sliced DataArray.
    """
    vals = da[dim].values
    real_start = start if start is not None else vals.min()
    real_stop = stop if stop is not None else vals.max()
    if all(vals[i] > vals[i + 1] for i in range(len(vals) - 1)):
        # Changing start/stop order if dim's values are reversed
        real_start, real_stop = real_stop, real_start
    return da.sel({dim: slice(real_start, real_stop, step)})


def extend_dataarray(
    da: xr.core.dataarray.DataArray,
    start: np.datetime64 = None,
    stop: np.datetime64 = None,
    step: int = 1,
    dim: str = "valid_time",
    freq_base: str = "h",
) -> xr.core.dataarray.DataArray:
    """extend_dataarray: extend a given dataarray to a new given start and
    a new given stop and gives a new step. Beware, it doesn't fill new steps

    Args:
        da (xr.core.dataarray.DataArray):  DataArray to extend.
        start (np.datetime64, optional): New start datetime.
            Defaults to None.
        stop (np.datetime64, optional): New stop datetime.
            Defaults to None.
        step (int, optional): New step. Defaults to 1.
        dim (str, optional): Step's dimension name. Defaults to "valid_time".
        freq_base (str, optional): Frequency base to choose among all pandas's
            frequency strings available here :
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html.
            Defaults to "h".

    Returns:
        xr.core.dataarray.DataArray: new dataarray with extend steps.
    """
    delta = np.timedelta64(step, freq_base)
    real_start = np.datetime64(start) if start is not None else da[dim].values.min()
    real_stop = np.datetime64(stop) if stop is not None else da[dim].values.max()
    new_range = np.arange(real_start, real_stop + delta, delta)
    new_range_da = xr.DataArray(
        np.ones(new_range.shape), dims=dim, coords={dim: new_range},
    )
    return da.broadcast_like(new_range_da)


def fill_dataarray(
    da: xr.core.dataarray.DataArray,
    source_steps: list,
    target_step: int = 1,
    dim: str = "valid_time",
    freq_base: str = "h",
) -> xr.core.dataarray.DataArray:
    """fill_dataarray : fill values in a given dataarray along a given 'dim'
    dimension to a 'target_step' and with a filling tolerance of 'source_step'

    Args:
        da (xr.core.dataarray.DataArray): DataArray to fill
        source_steps (list): List of source files step lengths. Used for
            limiting the filling of values.
        target_step (int, optional): Target file step length. Defaults to 1.
        dim (str, optional): Step's dimension name. Defaults to "valid_time".
        freq_base (str, optional): Frequency base to choose among all pandas's
            frequency strings available here :
            https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html.
            Defaults to "h".

    Returns:
        xr.core.dataarray.DataArray: new dataarray with filled values.
    """
    if len(source_steps) != len(da[dim]):
        LOGGER.error(
            "Source steps list and given datarray must have the same length",
            source_steps_len=len(source_steps),
            da_len=len(da[dim]),
        )
        return da

    delta = np.timedelta64(1, freq_base)
    new_das = []

    # Extension de chaque step
    for i, local_stop in enumerate(da[dim].values):
        sub_da = da.sel({dim: local_stop})
        source_step = source_steps[i] - 1
        local_start = local_stop - source_step * delta
        new_das.append(
            extend_dataarray(
                da=sub_da,
                start=local_start,
                stop=local_stop,
                step=target_step,
                dim=dim,
                freq_base=freq_base,
            ).bfill(dim=dim)
        )

    # Reconcatenation du nouveau dataarray
    return xr.concat(new_das, dim=dim)
