""" core.py

Core "binary" file
Compute risks and texts and exports them to JSON
"""
# Built-in imports
import json
import argparse
from copy import deepcopy
from multiprocessing import Pool
from functools import partial

# Third parties imports
import mflog

# Own package imports
from ftap_autom.promethee.src.conf.app_configuration import (
    MFLOG_LEVEL,
    MFLOG_FILE,
    MFLOG_CONF,
    WORKING_DIR,
)
from ftap_autom.promethee.src.lib.parameter_generator import ParameterGeneratorTransit
from ftap_autom.promethee.src.utils.exception import handle_error
from ftap_autom.promethee.src.lib.interpretor import AleaInterp
from ftap_autom.promethee.src.lib.localisation.localisation_manager import Localisation
from ftap_autom.promethee.src.lib.localisation.output_localisation import (
    OutputLocalisationExporter,
    OutputLocalisationFormatter,
)

# Logging
mflog.set_config(
    json_file=MFLOG_FILE, json_minimal_level=MFLOG_LEVEL, override_files=MFLOG_CONF
)
LOGGER = mflog.getLogger("core_localisaiton.mod")
LOGGER = LOGGER.bind(module="coreLocalisation")


def get_component_configs(prod_config):
    """get_component_configs : Retrieve all components configs from the
    production configuration and add them production elements like the
    production id, the original config hash, and the mask hash.

    Args:
        prod_config (dict): Single production configuration

    Returns:
        list : List of components configurations
    """
    component_configs = []
    for component_config in prod_config["components"]:
        component_configs += [
            {
                **deepcopy(component_config),
                "prod_id": prod_config["id"],
                "config_hash": prod_config.get("config_hash"),
                "mask_hash": prod_config.get("mask_hash"),
                "prod_hash": prod_config.get("prod_hash"),
            }
        ]
    return component_configs


def compute_risks(component_config):
    """compute_risks : Compute risks and detailed comments before formatting

    Args:
        component_config (dict): "Compacted" component configuration

    Returns:
        dict : Computed component dictionnary structured this way :
            {
                'risks' : risks_dataset,
                'comments' : {component_id : text}
            }
    """
    # Computing risks
    alea_conf = AleaInterp(component_config)
    parameter_handler = ParameterGeneratorTransit(alea_conf)
    parameter_handler.compute_risks()
    parameter_handler.get_final_risk_level()
    return parameter_handler


def compute(component_config):
    """compute : Factory function which computes components and format them
    into a dictionnary following the expected "CDP Bulletin" structure.

    Args:
        component_config (dict): Component configuration dictionnary

    Returns:
        list : List of formatted components
    """
    formatted_list = []
    component_type = component_config["type"]
    if component_type in ["risk", "Risk"]:
        # LOGGER.debug("Calling compute risk")
        component = compute_risks(component_config)
        # LOGGER.debug("Risk computed")
        # Rajouter le calcul de la localisation
        ds_max = component.get_final_risk_level()
        for geo_id in component_config["geos"]["mask_id"]:
            area_max_risk = ds_max.sel(id=geo_id)
            maxi = area_max_risk.max().values
            # On va avoir besoin de localiser
            compo_name = component_config["name"]
            hazard_name = component_config["hazardName"]
            start_time = area_max_risk.valid_time.min().dt.strftime("%Y%m%dH%H").values
            end_time = area_max_risk.valid_time.max().dt.strftime("%Y%m%dH%H").values

            if maxi > 1:
                LOGGER.warning(
                    f"Doing localisation {compo_name},{hazard_name}: "
                    f"Start {start_time}, end {end_time}"
                )
                risk_period = area_max_risk.sel(valid_time=(area_max_risk == maxi))
                loc_handler = Localisation(
                    component_config,
                    risk_level=maxi,
                    geo_id=geo_id,
                    period=set(risk_period.valid_time.values),
                )
                alea_conf = AleaInterp(component_config)
                formatted_list.append(
                    OutputLocalisationFormatter(
                        alea_conf, loc_handler, risk_level=maxi, max_risk=area_max_risk
                    ).formatted_component
                )
            else:
                LOGGER.warning(
                    "No need of localisation because there is no risk "
                    f"{compo_name},{hazard_name}: Start {start_time}, End {end_time}"
                )
    return formatted_list


def main(filename, nproc, working_dir="./"):
    """ main
    Args:
        filename (str): Fichier contenant la configuration a tratier
        nproc (int):  Nombre de processus à lancer en parallèle
    To Do :
       Passer en apply_async (ou apply) à la place de map.
       Cela permettra de "délimiter" les erreurs possibles.

    """
    # Retrieve config
    with open(filename, "r") as prod_configs_f:
        prod_configs = json.load(prod_configs_f)

    module_error_handler = partial(handle_error, log=LOGGER)
    # Group the risk
    if nproc > 1:
        pool = Pool(nproc)
        res_config = []
        for config in prod_configs.keys():
            pool.apply_async(
                get_component_configs,
                args=(prod_configs[config],),
                callback=res_config.extend,
                error_callback=module_error_handler,
            )
        pool.close()
        pool.join()
        # Calcul du risque
        results = []
        pool = Pool(nproc)
        for config in res_config:
            pool.apply_async(
                compute,
                args=(config,),
                callback=results.extend,
                error_callback=module_error_handler,
            )
        pool.close()
        pool.join()
    else:
        results = []
        for elt in prod_configs.keys():
            conf = get_component_configs(prod_configs[elt])
            for component in conf:
                res = compute(component)
                results.extend(res)

    exporter = OutputLocalisationExporter(results, working_dir=working_dir)
    exporter.export_all()


if __name__ == "__main__":
    # Arguments parsing
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f", "--filename", default=WORKING_DIR["configs"]["prod"], type=str
    )
    parser.add_argument("-n", "--nproc", default=16, type=int)
    args = parser.parse_args()
    main(args.filename, args.nproc)
