""" process_config.py

Configuration proccessing "binary" file
Processes a "global" configuration file and, given a production date, infers
"version", "data", "mask" and "production" configurations
"""
import os
import json
import argparse
from datetime import datetime

from ftap_autom.promethee.src.lib.config_processor import ConfigMetronomeProcessor
from ftap_autom.promethee.src.conf.app_configuration import (
    WORKING_DIR,
    DATETIME_FMT,
    JSON_KWARGS,
    RULES_XLS,
)


if __name__ == "__main__":
    # Arguments parsing
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-w", "--workingdir", default="./", type=str, help="Working directory."
    )
    parser.add_argument(
        "-d",
        "--draftdate",
        default=datetime.now().strftime(DATETIME_FMT),
        type=str,
        help="Drafting or launching datetime",
    )
    parser.add_argument(
        "-n",
        "--nproc",
        default=16,
        type=int,
        help="Number of CPU to use for multiprocessing.",
    )
    parser.add_argument(
        "-f",
        "--fname",
        default=WORKING_DIR["configs"]["global"],
        type=str,
        help="Relative path (from the working directory) "
        "to the configuration file to process.",
    )
    parser.add_argument(
        "-r",
        "--rules",
        default=list(RULES_XLS.keys())[0],
        type=str,
        help="Name of the rules convention (for files selection). This "
        "argument must be a key of the app_configuration.RULES_XLS "
        "dictionary. To choose among {}".format(set(RULES_XLS.keys())),
    )
    parser.add_argument(
        "-a",
        "--vapp",
        default="promethee",
        type=str,
        help="Application's name (default is 'Promethee')",
    )
    parser.add_argument(
        "-c",
        "--vconf",
        default="msb",
        type=str,
        help="Application's configuration (default is 'msb')",
    )
    parser.add_argument(
        "-x",
        "--experiment",
        default="TEST",
        type=str,
        help="Application's experiment name (default is TEST)",
    )
    args = parser.parse_args()
    print(args)

    # Filenames
    global_filename = os.path.join(args.workingdir, args.fname)

    mask_filename = os.path.join(args.workingdir, WORKING_DIR["configs"]["mask"])
    prod_filename = os.path.join(args.workingdir, WORKING_DIR["configs"]["prod"])
    data_filename = os.path.join(args.workingdir, WORKING_DIR["configs"]["data"])
    version_filename = os.path.join(args.workingdir, WORKING_DIR["configs"]["version"])

    # Running the config processor
    config_processor = ConfigMetronomeProcessor(
        global_config_filename=global_filename,
        rules=args.rules,
        drafting_datetime=datetime.strptime(args.draftdate, DATETIME_FMT),
        working_dir=args.workingdir,
        vapp=args.vapp,
        vconf=args.vconf,
        experiment=args.experiment,
    )
    mask_dico, data_dico, prod_dico = config_processor.process_all()

    # Mask config
    with open(mask_filename, "w") as mask_f:
        json.dump(mask_dico, mask_f, **JSON_KWARGS)

    # Prod config
    with open(prod_filename, "w") as prod_f:
        json.dump(prod_dico, prod_f, **JSON_KWARGS)

    # Data config
    with open(data_filename, "w") as data_f:
        json.dump(data_dico, data_f, **JSON_KWARGS)

    # Version config
    version_dico = {
        "version": config_processor.hashcode,
        "drafting_datetime": config_processor.drafting_datetime.strftime(DATETIME_FMT),
        "reference_datetime": config_processor.reference_datetime.strftime(
            DATETIME_FMT
        ),
        "production_datetime": config_processor.production_datetime.strftime(
            DATETIME_FMT
        ),
        "configuration_datetime": config_processor.get_configuration_date(),
    }
    with open(version_filename, "w") as version_f:
        json.dump(version_dico, version_f, **JSON_KWARGS)
