""" process_mask.py

Mask proccessing "binary" file
"""
# Built-in imports
import os
import argparse
import json
import multiprocessing as mp
from functools import partial

# third parties imports
import mflog
import xarray as xr

# own package imports
from ftap_autom.promethee.src.conf.app_configuration import (
    MFLOG_LEVEL,
    MFLOG_FILE,
    MFLOG_CONF,
    ALTITUDES_NC,
    WORKING_DIR,
)
import ftap_autom.promethee.src.lib.mask_processor as mpr
import ftap_autom.promethee.src.utils.hash as hashmod
from ftap_autom.promethee.src.utils.exception import handle_error

# Logging
mflog.set_config(
    json_file=MFLOG_FILE, json_minimal_level=MFLOG_LEVEL, override_files=MFLOG_CONF
)
LOGGER = mflog.getLogger("process_mask.mod")
LOGGER = LOGGER.bind(module="process_mask")


def main(conf, grid_file):
    """main : main function

    Args:
        conf (dict): Single mask configuration
        grid_file (str): Path to a netcdf file defining grids
    """
    # Quelques checks à faire au prealable
    do_something = True
    #  Si la conf contient le fichier on l'utilise.
    #  Sinon on suppose qu'on cree le fichier quoiqu'il arrive
    if "file" in conf:
        LOGGER.info("File is in the conf")
        output_file = conf.get("file")
        LOGGER.info(output_file)
        LOGGER.info(os.path.exists(output_file))
        if os.path.exists(output_file):
            LOGGER.info("File already exist")
            if "mask_hash" in conf:
                current_hash = conf.get("mask_hash")
            else:
                handler = hashmod.MD5(conf["geos"])
                current_hash = handler.hash
            LOGGER.info("Current conf hash is %s", current_hash)
            ds = xr.open_dataset(output_file)
            if ds.attrs.get("md5sum", "") == current_hash:
                do_something = False
            else:
                LOGGER.info("Current hash is different from the one in the file")
            ds.close()
    if do_something:
        LOGGER.info("Launching mask creation")
        mask_handler = mpr.MaskProcessor(conf, grid_file)
        mask_handler.create_masks()
    else:
        LOGGER.info(
            "Mask already exist and have the same md5sum. Mask creation is skipped."
        )
    LOGGER.info(f"Mask {output_file} has been created")


if __name__ == "__main__":
    # Argument parsing
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c", "--config", default=WORKING_DIR["configs"]["mask"], type=str
    )
    parser.add_argument("-g", "--grid_file", default=ALTITUDES_NC, type=str)
    parser.add_argument("-n", "--nproc", default=16, type=int)
    args = parser.parse_args()

    with open(args.config, "r") as fp:
        dict_config = json.load(fp)

    module_error_handler = partial(handle_error, log=LOGGER)
    nproc = args.nproc
    if nproc > 1:
        pool = mp.Pool(nproc)
        for conf in dict_config.keys():
            pool.apply_async(
                main,
                args=(dict_config[conf],),
                kwds={"grid_file": args.grid_file},
                error_callback=module_error_handler,
            )
        pool.close()
        pool.join()
    else:
        for conf in dict_config.keys():
            print(conf)
            main(dict_config[conf], args.grid_file)
